(function(){

/////////////////////////////////////////////////////////////////////////
//                                                                     //
// server/admin-server.js                                              //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
                                                                       //
// Sandstorm - Personal Cloud Sandbox                                  //
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors
// All rights reserved.                                                //
//                                                                     //
// Licensed under the Apache License, Version 2.0 (the "License");     //
// you may not use this file except in compliance with the License.    //
// You may obtain a copy of the License at                             //
//                                                                     //
//   http://www.apache.org/licenses/LICENSE-2.0                        //
//                                                                     //
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,   //
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.                                      //
                                                                       //
var ADMIN_TOKEN_EXPIRATION_TIME = 15 * 60 * 1000;                      // 17
var publicAdminSettings = ["google", "github", "ldap", "emailToken", "splashUrl", "signupDialog", "adminAlert", "adminAlertTime", "adminAlertUrl", "termsUrl", "privacyUrl", "appMarketUrl", "appIndexUrl", "appUpdatesEnabled", "serverTitle", "returnAddress", "ldapNameField", "organizationEmail", "organizationLdap", "organizationGoogle"];
                                                                       //
var FEATURE_KEY_FIELDS_PUBLISHED_TO_ADMINS = ["customer", "expires", "features", "isElasticBilling", "isTrial", "issued", "userLimit"];
                                                                       //
var PUBLIC_FEATURE_KEY_FIELDS = ["expires", "features"];               // 30
                                                                       //
var Fs = Npm.require("fs");                                            // 34
var SANDSTORM_ADMIN_TOKEN = SANDSTORM_VARDIR + "/adminToken";          // 35
                                                                       //
var getSmtpUrl = function () {                                         // 37
  var setting = Settings.findOne({ _id: "smtpUrl" });                  // 38
  if (setting) {                                                       // 39
    return setting.value;                                              // 40
  } else {                                                             //
    return process.env.MAIL_URL;                                       // 42
  }                                                                    //
};                                                                     //
                                                                       //
var tokenIsValid = function (token) {                                  // 46
  if (token && Fs.existsSync(SANDSTORM_ADMIN_TOKEN)) {                 // 47
    var stats = Fs.statSync(SANDSTORM_ADMIN_TOKEN);                    // 48
    var expireTime = new Date(Date.now() - ADMIN_TOKEN_EXPIRATION_TIME);
    if (stats.mtime < expireTime) {                                    // 50
      return false;                                                    // 51
    } else {                                                           //
      return Fs.readFileSync(SANDSTORM_ADMIN_TOKEN, { encoding: "utf8" }) === token;
    }                                                                  //
  } else {                                                             //
    return false;                                                      // 56
  }                                                                    //
};                                                                     //
                                                                       //
var checkAuth = function (token) {                                     // 60
  check(token, Match.OneOf(undefined, null, String));                  // 61
  if (!isAdmin() && !tokenIsValid(token)) {                            // 62
    throw new Meteor.Error(403, "User must be admin or provide a valid token");
  }                                                                    //
};                                                                     //
                                                                       //
var clearAdminToken = function (token) {                               // 67
  if (tokenIsValid(token)) {                                           // 68
    Fs.unlinkSync(SANDSTORM_ADMIN_TOKEN);                              // 69
    console.log("Admin token deleted.");                               // 70
  }                                                                    //
};                                                                     //
                                                                       //
Meteor.methods({                                                       // 74
  setAccountSetting: function (token, serviceName, value) {            // 75
    checkAuth(token);                                                  // 76
    check(serviceName, String);                                        // 77
    check(value, Boolean);                                             // 78
                                                                       //
    // TODO(someday): currently this relies on the fact that an account is tied to a single
    // identity, and thus has only that entry in "services". This will need to be looked at when
    // multiple login methods/identities are allowed for a single account.
    if (!value && !tokenIsValid(token) && serviceName in Meteor.user().services) {
      throw new Meteor.Error(403, "You can not disable the login service that your account uses.");
    }                                                                  //
                                                                       //
    // Only check configurations for OAuth services.                   //
    var oauthServices = ["google", "github"];                          // 89
    if (value && oauthServices.indexOf(serviceName) != -1) {           // 90
      var ServiceConfiguration = Package["service-configuration"].ServiceConfiguration;
      var config = ServiceConfiguration.configurations.findOne({ service: serviceName });
      if (!config) {                                                   // 93
        throw new Meteor.Error(403, "You must configure the " + serviceName + " service before you can enable it. Click the \"configure\" link.");
      }                                                                //
                                                                       //
      if (!config.clientId || !config.secret) {                        // 98
        throw new Meteor.Error(403, "You must provide a non-empty clientId and secret for the " + serviceName + " service before you can enable it. Click the \"configure\" link.");
      }                                                                //
    }                                                                  //
                                                                       //
    Settings.upsert({ _id: serviceName }, { $set: { value: value } });
    if (value) {                                                       // 105
      Settings.update({ _id: serviceName }, { $unset: { automaticallyReset: 1 } });
    }                                                                  //
  },                                                                   //
                                                                       //
  setSetting: function (token, name, value) {                          // 110
    checkAuth(token);                                                  // 111
    check(name, String);                                               // 112
    check(value, Match.OneOf(null, String, Date, Boolean));            // 113
                                                                       //
    Settings.upsert({ _id: name }, { $set: { value: value } });        // 115
  },                                                                   //
                                                                       //
  submitFeatureKey: function (token, textBlock) {                      // 118
    checkAuth(token);                                                  // 119
    check(textBlock, Match.OneOf(null, String));                       // 120
                                                                       //
    var db = this.connection.sandstormDb;                              // 122
                                                                       //
    if (!textBlock) {                                                  // 124
      // Delete the feature key.                                       //
      db.collections.featureKey.remove("currentFeatureKey");           // 126
      return;                                                          // 127
    }                                                                  //
                                                                       //
    // textBlock is a base64'd string, possibly with newlines and comment lines starting with "-"
    var featureKeyBase64 = _.chain(textBlock.split("\n")).filter(function (line) {
      return line.length > 0 && line[0] !== "-";                       //
    }).value().join("");                                               //
                                                                       //
    var buf = new Buffer(featureKeyBase64, "base64");                  // 136
    if (buf.length < 64) {                                             // 137
      throw new Meteor.Error(401, "Invalid feature key");              // 138
    }                                                                  //
                                                                       //
    // loadSignedFeatureKey is provided in feature-key.js              //
    var featureKey = loadSignedFeatureKey(buf);                        // 142
    if (!featureKey) {                                                 // 143
      throw new Meteor.Error(401, "Invalid feature key");              // 144
    }                                                                  //
                                                                       //
    // Persist the feature key in the database.                        //
    db.collections.featureKey.upsert("currentFeatureKey", {            // 148
      _id: "currentFeatureKey",                                        // 151
      value: buf                                                       // 152
    });                                                                //
  },                                                                   //
                                                                       //
  getSmtpUrl: function (token) {                                       // 157
    checkAuth(token);                                                  // 158
                                                                       //
    return getSmtpUrl();                                               // 160
  },                                                                   //
                                                                       //
  adminConfigureLoginService: function (token, options) {              // 163
    checkAuth(token);                                                  // 164
    check(options, Match.ObjectIncluding({ service: String }));        // 165
                                                                       //
    var ServiceConfiguration = Package["service-configuration"].ServiceConfiguration;
                                                                       //
    ServiceConfiguration.configurations.upsert({ service: options.service }, options);
  },                                                                   //
                                                                       //
  clearResumeTokensForService: function (token, serviceName) {         // 172
    checkAuth(token);                                                  // 173
    check(serviceName, String);                                        // 174
                                                                       //
    var query = {};                                                    // 176
    query["services." + serviceName] = { $exists: true };              // 177
    Meteor.users.find(query).forEach(function (identity) {             // 178
      if (identity.services.resume && identity.services.resume.loginTokens && identity.services.resume.loginTokens.length > 0) {
        Meteor.users.update({ _id: identity._id }, { $set: { "services.resume.loginTokens": [] } });
      }                                                                //
                                                                       //
      Meteor.users.update({ "loginIdentities.id": identity._id }, { $set: { "services.resume.loginTokens": [] } });
    });                                                                //
  },                                                                   //
                                                                       //
  adminUpdateUser: function (token, userInfo) {                        // 189
    checkAuth(token);                                                  // 190
    check(userInfo, {                                                  // 191
      userId: String,                                                  // 192
      signupKey: Boolean,                                              // 193
      isAdmin: Boolean                                                 // 194
    });                                                                //
                                                                       //
    var userId = userInfo.userId;                                      // 197
    if (userId === Meteor.userId() && !userInfo.isAdmin) {             // 198
      throw new Meteor.Error(403, "User cannot remove admin permissions from itself.");
    }                                                                  //
                                                                       //
    Meteor.users.update({ _id: userId }, { $set: _.omit(userInfo, ["_id", "userId"]) });
  },                                                                   //
                                                                       //
  testSend: function (token, smtpUrl, to) {                            // 205
    checkAuth(token);                                                  // 206
    check(smtpUrl, String);                                            // 207
    check(to, String);                                                 // 208
                                                                       //
    SandstormEmail.send({                                              // 210
      to: to,                                                          // 211
      from: globalDb.getServerTitle() + " <" + globalDb.getReturnAddress() + ">",
      subject: "Testing your Sandstorm's SMTP setting",                // 213
      text: "Success! Your outgoing SMTP is working.",                 // 214
      smtpUrl: smtpUrl                                                 // 215
    });                                                                //
  },                                                                   //
                                                                       //
  createSignupKey: function (token, note, quota) {                     // 219
    checkAuth(token);                                                  // 220
    check(note, String);                                               // 221
    check(quota, Match.OneOf(undefined, null, Number));                // 222
                                                                       //
    var key = Random.id();                                             // 224
    var content = { _id: key, used: false, note: note };               // 225
    if (typeof quota === "number") content.quota = quota;              // 226
    SignupKeys.insert(content);                                        // 227
    return key;                                                        // 228
  },                                                                   //
                                                                       //
  sendInvites: function (token, origin, from, list, subject, message, quota) {
    checkAuth(token);                                                  // 232
    check([origin, from, list, subject, message], [String]);           // 233
    check(quota, Match.OneOf(undefined, null, Number));                // 234
                                                                       //
    if (!from.trim()) {                                                // 236
      throw new Meteor.Error(403, "Must enter 'from' address.");       // 237
    }                                                                  //
                                                                       //
    if (!list.trim()) {                                                // 240
      throw new Meteor.Error(403, "Must enter 'to' addresses.");       // 241
    }                                                                  //
                                                                       //
    this.unblock();                                                    // 244
                                                                       //
    list = list.split("\n");                                           // 246
    for (var i in babelHelpers.sanitizeForInObject(list)) {            // 247
      var email = list[i].trim();                                      // 248
                                                                       //
      if (email) {                                                     // 250
        var key = Random.id();                                         // 251
                                                                       //
        var content = {                                                // 253
          _id: key,                                                    // 254
          used: false,                                                 // 255
          note: "E-mail invite to " + email,                           // 256
          email: email,                                                // 257
          definitelySent: false                                        // 258
        };                                                             //
        if (typeof quota === "number") content.quota = quota;          // 260
        SignupKeys.insert(content);                                    // 261
        SandstormEmail.send({                                          // 262
          to: email,                                                   // 263
          from: from,                                                  // 264
          subject: subject,                                            // 265
          text: message.replace(/\$KEY/g, origin + Router.routes.signup.path({ key: key }))
        });                                                            //
        SignupKeys.update(key, { $set: { definitelySent: true } });    // 268
      }                                                                //
    }                                                                  //
                                                                       //
    return { sent: true };                                             // 272
  },                                                                   //
                                                                       //
  offerIpNetwork: function (token) {                                   // 275
    checkAuth(token);                                                  // 276
    if (!isAdmin()) {                                                  // 277
      throw new Meteor.Error(403, "Offering IpNetwork is only allowed for logged in users " + "(a token is not sufficient). Please sign in with an admin account");
    }                                                                  //
                                                                       //
    var requirements = [{ userIsAdmin: Meteor.userId() }];             // 282
    var sturdyRef = waitPromise(saveFrontendRef({ ipNetwork: true }, { webkey: null }, requirements)).sturdyRef;
    return ROOT_URL.protocol + "//" + globalDb.makeApiHost(sturdyRef) + "#" + sturdyRef;
  },                                                                   //
                                                                       //
  offerIpInterface: function (token) {                                 // 291
    checkAuth(token);                                                  // 292
    if (!isAdmin()) {                                                  // 293
      throw new Meteor.Error(403, "Offering IpInterface is only allowed for logged in users " + "(a token is not sufficient). Please sign in with an admin account");
    }                                                                  //
                                                                       //
    var requirements = [{ userIsAdmin: Meteor.userId() }];             // 298
    var sturdyRef = waitPromise(saveFrontendRef({ ipInterface: true }, { webkey: null }, requirements)).sturdyRef;
    return ROOT_URL.protocol + "//" + globalDb.makeApiHost(sturdyRef) + "#" + sturdyRef;
  },                                                                   //
                                                                       //
  adminToggleDisableCap: function (token, capId, value) {              // 307
    checkAuth(token);                                                  // 308
    check(capId, String);                                              // 309
    check(value, Boolean);                                             // 310
                                                                       //
    if (value) {                                                       // 312
      ApiTokens.update({ _id: capId }, { $set: { revoked: true } });   // 313
    } else {                                                           //
      ApiTokens.update({ _id: capId }, { $set: { revoked: false } });  // 315
    }                                                                  //
  },                                                                   //
                                                                       //
  updateQuotas: function (token, list, quota) {                        // 319
    checkAuth(token);                                                  // 320
    check(list, String);                                               // 321
    check(quota, Match.OneOf(undefined, null, Number));                // 322
                                                                       //
    if (!list.trim()) {                                                // 324
      throw new Meteor.Error(400, "Must enter addresses.");            // 325
    }                                                                  //
                                                                       //
    var items = list.split("\n");                                      // 328
    var invalid = [];                                                  // 329
    for (var i in babelHelpers.sanitizeForInObject(items)) {           // 330
      var modifier = typeof quota === "number" ? { $set: { quota: quota } } : { $unset: { quota: "" } };
      var n = SignupKeys.update({ email: items[i] }, modifier, { multi: true });
      n += Meteor.users.update({ signupEmail: items[i] }, modifier, { multi: true });
                                                                       //
      if (n < 1) invalid.push(items[i]);                               // 336
    }                                                                  //
                                                                       //
    if (invalid.length > 0) {                                          // 339
      throw new Meteor.Error(404, "These addresses did not map to any user nor invite: " + invalid.join(", "));
    }                                                                  //
  },                                                                   //
                                                                       //
  dismissAdminStatsNotifications: function (token) {                   // 345
    checkAuth(token);                                                  // 346
    globalDb.collections.notifications.remove({ "admin.type": "reportStats" });
  },                                                                   //
                                                                       //
  signUpAsAdmin: function (token) {                                    // 350
    check(token, String);                                              // 351
    checkAuth(token);                                                  // 352
    if (!this.userId) {                                                // 353
      throw new Meteor.Error(403, "Must be logged in to sign up as admin.");
    }                                                                  //
                                                                       //
    if (!Meteor.user().loginIdentities) {                              // 357
      throw new Meteor.Error(403, "Must be logged into an account to sign up as admin.");
    }                                                                  //
                                                                       //
    Meteor.users.update({ _id: this.userId }, { $set: { isAdmin: true, signupKey: "admin" } });
    clearAdminToken(token);                                            // 362
  }                                                                    //
});                                                                    //
                                                                       //
var authorizedAsAdmin = function (token, userId) {                     // 366
  return Match.test(token, Match.OneOf(undefined, null, String)) && (userId && isAdminById(userId) || tokenIsValid(token));
};                                                                     //
                                                                       //
Meteor.publish("admin", function (token) {                             // 371
  if (!authorizedAsAdmin(token, this.userId)) return [];               // 372
  return Settings.find();                                              // 373
});                                                                    //
                                                                       //
Meteor.publish("adminServiceConfiguration", function (token) {         // 376
  if (!authorizedAsAdmin(token, this.userId)) return [];               // 377
  return Package["service-configuration"].ServiceConfiguration.configurations.find();
});                                                                    //
                                                                       //
Meteor.publish("publicAdminSettings", function () {                    // 381
  return Settings.find({ _id: { $in: publicAdminSettings } });         // 382
});                                                                    //
                                                                       //
Meteor.publish("adminToken", function (token) {                        // 385
  check(token, String);                                                // 386
  this.added("adminToken", "adminToken", { tokenIsValid: tokenIsValid(token) });
  this.ready();                                                        // 388
});                                                                    //
                                                                       //
Meteor.publish("allUsers", function (token) {                          // 391
  if (!authorizedAsAdmin(token, this.userId)) return [];               // 392
  return Meteor.users.find();                                          // 393
});                                                                    //
                                                                       //
Meteor.publish("activityStats", function (token) {                     // 396
  if (!authorizedAsAdmin(token, this.userId)) return [];               // 397
  return ActivityStats.find();                                         // 398
});                                                                    //
                                                                       //
Meteor.publish("statsTokens", function (token) {                       // 401
  if (!authorizedAsAdmin(token, this.userId)) return [];               // 402
  return StatsTokens.find();                                           // 403
});                                                                    //
                                                                       //
Meteor.publish("allPackages", function (token) {                       // 406
  if (!authorizedAsAdmin(token, this.userId)) return [];               // 407
  return Packages.find({ manifest: { $exists: true } }, { fields: { appId: 1, "manifest.appVersion": 1,
      "manifest.actions": 1, "manifest.appTitle": 1 } });              // 410
});                                                                    //
                                                                       //
Meteor.publish("realTimeStats", function (token) {                     // 413
  if (!authorizedAsAdmin(token, this.userId)) return [];               // 414
                                                                       //
  // Last five minutes.                                                //
  this.added("realTimeStats", "now", computeStats(new Date(Date.now() - 5 * 60 * 1000)));
                                                                       //
  // Since last sample.                                                //
  var lastSample = ActivityStats.findOne({}, { sort: { timestamp: -1 } });
  var lastSampleTime = lastSample ? lastSample.timestamp : new Date(0);
  this.added("realTimeStats", "today", computeStats(lastSampleTime));  // 422
                                                                       //
  // TODO(someday): Update every few minutes?                          //
                                                                       //
  this.ready();                                                        // 426
});                                                                    //
                                                                       //
Meteor.publish("adminLog", function (token) {                          // 429
  if (!authorizedAsAdmin(token, this.userId)) return [];               // 430
                                                                       //
  var logfile = SANDSTORM_LOGDIR + "/sandstorm.log";                   // 432
                                                                       //
  var fd = Fs.openSync(logfile, "r");                                  // 434
  var startSize = Fs.fstatSync(fd).size;                               // 435
                                                                       //
  // Start tailing at EOF - 8k.                                        //
  var offset = Math.max(0, startSize - 8192);                          // 438
                                                                       //
  var _this = this;                                                    // 440
  function doTail() {                                                  // 441
    for (;;) {                                                         // 442
      var buf = new Buffer(Math.max(1024, startSize - offset));        // 443
      var n = Fs.readSync(fd, buf, 0, buf.length, offset);             // 444
      if (n <= 0) break;                                               // 445
      _this.added("adminLog", offset, { text: buf.toString("utf8", 0, n) });
      offset += n;                                                     // 447
    }                                                                  //
  }                                                                    //
                                                                       //
  // Watch the file for changes.                                       //
  var watcher = Fs.watch(logfile, { persistent: false }, Meteor.bindEnvironment(doTail));
                                                                       //
  // When the subscription stops, stop watching the file.              //
  this.onStop(function () {                                            // 455
    watcher.close();                                                   // 456
    Fs.closeSync(fd);                                                  // 457
  });                                                                  //
                                                                       //
  // Read initial 8k tail data immediately.                            //
  doTail();                                                            // 461
                                                                       //
  // Notify ready.                                                     //
  this.ready();                                                        // 464
});                                                                    //
                                                                       //
Meteor.publish("adminApiTokens", function (token) {                    // 467
  if (!authorizedAsAdmin(token, this.userId)) return [];               // 468
  return ApiTokens.find({                                              // 469
    $or: [{ "frontendRef.ipNetwork": { $exists: true } }, { "frontendRef.ipInterface": { $exists: true } }]
  }, {                                                                 //
    fields: {                                                          // 475
      frontendRef: 1,                                                  // 476
      created: 1,                                                      // 477
      requirements: 1,                                                 // 478
      revoked: 1                                                       // 479
    }                                                                  //
  });                                                                  //
});                                                                    //
                                                                       //
Meteor.publish("featureKey", function (forAdmin, token) {              // 484
  var _this2 = this;                                                   //
                                                                       //
  if (forAdmin && !authorizedAsAdmin(token, this.userId)) return [];   // 485
                                                                       //
  var fields = forAdmin ? FEATURE_KEY_FIELDS_PUBLISHED_TO_ADMINS : PUBLIC_FEATURE_KEY_FIELDS;
                                                                       //
  var db = this.connection.sandstormDb;                                // 490
  var featureKeyQuery = db.collections.featureKey.find({ _id: "currentFeatureKey" });
  var observeHandle = featureKeyQuery.observe({                        // 492
    added: function (doc) {                                            // 493
      // Load and verify the signed feature key.                       //
      var buf = new Buffer(doc.value);                                 // 495
      var featureKey = loadSignedFeatureKey(buf);                      // 496
                                                                       //
      if (featureKey) {                                                // 498
        var _ref;                                                      //
                                                                       //
        // If the signature is valid, publish the feature key information.
        var filteredFeatureKey = (_ref = _).pick.apply(_ref, [featureKey].concat(fields));
        _this2.added("featureKey", doc._id, filteredFeatureKey);       // 501
      }                                                                //
    },                                                                 //
                                                                       //
    changed: function (newDoc, oldDoc) {                               // 505
      // Load and reverify the new signed feature key.                 //
      var buf = new Buffer(newDoc.value);                              // 507
      var featureKey = loadSignedFeatureKey(buf);                      // 508
                                                                       //
      if (featureKey) {                                                // 510
        var _ref2;                                                     //
                                                                       //
        // If the signature is valid, call this.changed() with the interesting fields.
        var filteredFeatureKey = (_ref2 = _).pick.apply(_ref2, [featureKey].concat(fields));
        _this2.changed("featureKey", newDoc._id, filteredFeatureKey);  // 513
      } else {                                                         //
        // Otherwise, call this.removed(), since the new feature key is invalid.
        _this2.removed("featureKey", oldDoc._id);                      // 516
      }                                                                //
    },                                                                 //
                                                                       //
    removed: function (oldDoc) {                                       // 520
      _this2.removed("featureKey", oldDoc._id);                        // 521
    }                                                                  //
  });                                                                  //
                                                                       //
  this.onStop(function () {                                            // 525
    observeHandle.stop();                                              // 526
  });                                                                  //
                                                                       //
  this.ready();                                                        // 529
});                                                                    //
                                                                       //
function observeOauthService(name) {                                   // 532
  Settings.find({ _id: name, value: true }).observe({                  // 533
    added: function () {                                               // 534
      // Tell the oauth library it should accept login attempts from this service.
      Accounts.oauth.registerService(name);                            // 536
    },                                                                 //
                                                                       //
    removed: function () {                                             // 539
      // Tell the oauth library it should deny login attempts from this service.
      Accounts.oauth.unregisterService(name);                          // 541
    }                                                                  //
  });                                                                  //
}                                                                      //
                                                                       //
observeOauthService("github");                                         // 546
observeOauthService("google");                                         // 547
/////////////////////////////////////////////////////////////////////////

}).call(this);

//# sourceMappingURL=admin-server.js.map
