(function(){

/////////////////////////////////////////////////////////////////////////
//                                                                     //
// server/grain-server.js                                              //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
                                                                       //
// Sandstorm - Personal Cloud Sandbox                                  //
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors
// All rights reserved.                                                //
//                                                                     //
// Licensed under the Apache License, Version 2.0 (the "License");     //
// you may not use this file except in compliance with the License.    //
// You may obtain a copy of the License at                             //
//                                                                     //
//   http://www.apache.org/licenses/LICENSE-2.0                        //
//                                                                     //
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,   //
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.                                      //
                                                                       //
var Crypto = Npm.require("crypto");                                    // 17
                                                                       //
var emailLinkWithInlineStyle = function (url, text) {                  // 19
  return "<a href='" + url + "' style='display:inline-block;text-decoration:none;" + "font-family:sans-serif;width:200px;min-height:30px;line-height:30px;" + "border-radius:4px;text-align:center;background:#428bca;color:white'>" + text + "</a>";
};                                                                     //
                                                                       //
Meteor.publish("grainTopBar", function (grainId) {                     // 26
  check(grainId, String);                                              // 27
                                                                       //
  var result = [Grains.find({                                          // 29
    _id: grainId,                                                      // 31
    $or: [{ userId: this.userId }, { "private": { $ne: true } }]       // 32
  }, {                                                                 //
    fields: {                                                          // 37
      title: 1,                                                        // 38
      userId: 1,                                                       // 39
      identityId: 1,                                                   // 40
      "private": 1                                                     // 41
    }                                                                  //
  })];                                                                 //
  if (this.userId) {                                                   // 45
    var myIdentityIds = SandstormDb.getUserIdentityIds(globalDb.getUser(this.userId));
    result.push(ApiTokens.find({                                       // 47
      grainId: grainId,                                                // 48
      $or: [{ "owner.user.identityId": { $in: myIdentityIds } }, { identityId: { $in: myIdentityIds } }]
    }));                                                               //
  }                                                                    //
                                                                       //
  return result;                                                       // 56
});                                                                    //
                                                                       //
// We allow users to learn package information about a grain they own.
// This is used for obtaining icon and app title information for grains
// you own, which is used in the sidebar. It is not a security/privacy
// risk since it only exposes this information for grains the user owns.
Meteor.publish("packageByGrainId", function (grainId) {                // 63
  check(grainId, String);                                              // 64
  var publishThis = [];                                                // 65
  // We need to publish the packageId so that client-side code can     //
  // find the right package.                                           //
  var thisGrainCursor = Grains.find({                                  // 68
    _id: grainId,                                                      // 69
    userId: this.userId                                                // 70
  }, {                                                                 //
    fields: { packageId: 1 }                                           // 72
  });                                                                  //
  publishThis.push(thisGrainCursor);                                   // 74
                                                                       //
  if (thisGrainCursor.count()) {                                       // 76
    var thisGrain = thisGrainCursor.fetch()[0];                        // 77
    var thisPackageCursor = Packages.find({ _id: thisGrain.packageId });
    publishThis.push(thisPackageCursor);                               // 79
  }                                                                    //
                                                                       //
  return publishThis;                                                  // 82
});                                                                    //
                                                                       //
Meteor.publish("tokenInfo", function (token) {                         // 85
  // Allows the client side to map a raw token to its entry in ApiTokens, and the additional
  // metadata that it will need to display the app icon and title.  We do not care about making
  // the metadata reactive.                                            //
  check(token, String);                                                // 89
                                                                       //
  var hashedToken = Crypto.createHash("sha256").update(token).digest("base64");
  var apiToken = ApiTokens.findOne({                                   // 92
    _id: hashedToken                                                   // 93
  }, {                                                                 //
    fields: {                                                          // 95
      grainId: 1,                                                      // 96
      owner: 1                                                         // 97
    }                                                                  //
  });                                                                  //
  if (!apiToken) {                                                     // 100
    this.added("tokenInfo", token, { invalidToken: true });            // 101
  } else {                                                             //
    var grainId = apiToken.grainId;                                    // 103
    var grain = Grains.findOne({                                       // 104
      _id: grainId                                                     // 105
    }, {                                                               //
      fields: {                                                        // 107
        packageId: 1,                                                  // 108
        appId: 1                                                       // 109
      }                                                                //
    });                                                                //
    if (!grain) {                                                      // 112
      this.added("tokenInfo", token, { invalidToken: true });          // 113
    } else {                                                           //
      if (apiToken.owner && apiToken.owner.user) {                     // 115
        var identity = globalDb.getIdentity(apiToken.owner.user.identityId);
        var metadata = apiToken.owner.user.denormalizedGrainMetadata;  // 117
        if (identity && metadata) {                                    // 118
          this.added("tokenInfo", token, { identityOwner: _.pick(identity, "_id", "profile"),
            grainId: grainId,                                          // 121
            grainMetadata: metadata });                                // 122
        } else {                                                       //
          this.added("tokenInfo", token, { invalidToken: true });      // 124
        }                                                              //
      } else if (!apiToken.owner || "webkey" in apiToken.owner) {      //
        var pkg = Packages.findOne({ _id: grain.packageId }, { fields: { manifest: 1 } });
        var appTitle = pkg && pkg.manifest && pkg.manifest.appTitle || { defaultText: "" };
        var appIcon = undefined;                                       // 129
        if (pkg && pkg.manifest && pkg.manifest.metadata && pkg.manifest.metadata.icons) {
          var icons = pkg.manifest.metadata.icons;                     // 131
          appIcon = icons.grain || icons.appGrid;                      // 132
        }                                                              //
                                                                       //
        var denormalizedGrainMetadata = {                              // 135
          appTitle: appTitle,                                          // 136
          icon: appIcon,                                               // 137
          appId: appIcon ? undefined : grain.appId                     // 138
        };                                                             //
        this.added("tokenInfo", token, { webkey: true,                 // 140
          grainId: grainId,                                            // 142
          grainMetadata: denormalizedGrainMetadata });                 // 143
      } else {                                                         //
        this.added("tokenInfo", token, { invalidToken: true });        // 145
      }                                                                //
    }                                                                  //
  }                                                                    //
                                                                       //
  this.ready();                                                        // 150
  return;                                                              // 151
});                                                                    //
                                                                       //
Meteor.publish("requestingAccess", function (grainId) {                // 154
  check(grainId, String);                                              // 155
                                                                       //
  if (!this.userId) {                                                  // 157
    throw new Meteor.Error(403, "Must be logged in to request access.");
  }                                                                    //
                                                                       //
  var identityIds = SandstormDb.getUserIdentityIds(Meteor.users.findOne({ _id: this.userId }));
                                                                       //
  var grain = globalDb.getGrain(grainId);                              // 163
  if (!grain) {                                                        // 164
    throw new Meteor.Error(404, "Grain not found.");                   // 165
  }                                                                    //
                                                                       //
  if (grain.userId === this.userId) {                                  // 168
    this.added("grantedAccessRequests", Random.id(), { grainId: grainId, identityId: grain.identityId });
  }                                                                    //
                                                                       //
  var _this = this;                                                    // 173
  var query = ApiTokens.find({ grainId: grainId, accountId: grain.userId,
    parentToken: { $exists: false },                                   // 175
    "owner.user.identityId": { $in: identityIds },                     // 176
    revoked: { $ne: true } });                                         // 177
  var handle = query.observe({                                         // 178
    added: function (apiToken) {                                       // 179
      _this.added("grantedAccessRequests", Random.id(), { grainId: grainId, identityId: apiToken.owner.user.identityId });
    }                                                                  //
  });                                                                  //
                                                                       //
  this.onStop(function () {                                            // 185
    return handle.stop();                                              //
  });                                                                  //
});                                                                    //
                                                                       //
Meteor.publish("grainSize", function (grainId) {                       // 188
  // Publish pseudo-collection containing the size of the grain opened in the given session.
  check(grainId, String);                                              // 190
                                                                       //
  var grain = Grains.findOne(grainId);                                 // 192
  if (!grain || grain.userId !== this.userId) {                        // 193
    return [];                                                         // 194
  }                                                                    //
                                                                       //
  var supervisor = globalBackend.cap().getGrain(this.userId, grainId).supervisor;
                                                                       //
  var _this = this;                                                    // 199
  var stopped = false;                                                 // 200
  var promise = getGrainSize(supervisor);                              // 201
                                                                       //
  function getNext(oldSize) {                                          // 203
    promise = getGrainSize(supervisor, oldSize);                       // 204
    promise.then(function (size) {                                     // 205
      if (!stopped) {                                                  // 206
        if (size !== oldSize) {                                        // 207
          // sometimes there are false alarms                          //
          _this.changed("grainSizes", grainId, { size: size });        // 208
        }                                                              //
                                                                       //
        getNext(size);                                                 // 211
      }                                                                //
    }, function (err) {                                                //
                                                                       //
      if (!stopped) {                                                  // 215
        if (err.kjType === "disconnected") {                           // 216
          _this.stop();                                                // 217
        } else {                                                       //
          _this.error(err);                                            // 219
        }                                                              //
      }                                                                //
    });                                                                //
  }                                                                    //
                                                                       //
  promise.then(function (size) {                                       // 225
    if (!stopped) {                                                    // 226
      _this.added("grainSizes", grainId, { size: size });              // 227
      _this.ready();                                                   // 228
      getNext(size);                                                   // 229
    }                                                                  //
  }, function (err) {                                                  //
                                                                       //
    if (!stopped) {                                                    // 233
      if (err.kjType === "disconnected") {                             // 234
        _this.stop();                                                  // 235
      } else {                                                         //
        _this.error(err);                                              // 237
      }                                                                //
    }                                                                  //
  });                                                                  //
                                                                       //
  _this.onStop(function () {                                           // 242
    stopped = true;                                                    // 243
    promise.cancel();                                                  // 244
  });                                                                  //
});                                                                    //
                                                                       //
Meteor.methods({                                                       // 248
  deleteGrain: function (grainId) {                                    // 249
    check(grainId, String);                                            // 250
                                                                       //
    if (this.userId) {                                                 // 252
      var grain = Grains.findOne({ _id: grainId, userId: this.userId });
      if (grain) {                                                     // 254
        Grains.remove(grainId);                                        // 255
        globalDb.removeApiTokens({                                     // 256
          grainId: grainId,                                            // 257
          $or: [{ owner: { $exists: false } }, { owner: { webkey: null } }]
        });                                                            //
        if (grain.lastUsed) {                                          // 263
          DeleteStats.insert({ type: isDemoUser() ? "demoGrain" : "grain",
            lastActive: grain.lastUsed, appId: grain.appId });         // 265
        }                                                              //
                                                                       //
        if (!this.isSimulation) {                                      // 268
          waitPromise(globalBackend.deleteGrain(grainId, this.userId));
          Meteor.call("deleteUnusedPackages", grain.appId);            // 270
        }                                                              //
      }                                                                //
    }                                                                  //
  },                                                                   //
                                                                       //
  forgetGrain: function (grainId, identityId) {                        // 276
    // TODO(cleanup): For now we are ignoring `identityId`, but maybe we should expose finer-grained
    //  forgetting.                                                    //
                                                                       //
    check(grainId, String);                                            // 280
    check(identityId, String);                                         // 281
                                                                       //
    if (!this.userId) {                                                // 283
      throw new Meteor.Error(403, "Must be logged in to forget a grain.");
    }                                                                  //
                                                                       //
    SandstormDb.getUserIdentityIds(Meteor.user()).forEach(function (identityId) {
      globalDb.removeApiTokens({ grainId: grainId, "owner.user.identityId": identityId });
    });                                                                //
  },                                                                   //
                                                                       //
  updateGrainPreferredIdentity: function (grainId, identityId) {       // 292
    check(grainId, String);                                            // 293
    check(identityId, String);                                         // 294
    if (!this.userId) {                                                // 295
      throw new Meteor.Error(403, "Must be logged in.");               // 296
    }                                                                  //
                                                                       //
    var grain = globalDb.getGrain(grainId) || {};                      // 299
    if (!grain.userId === this.userId) {                               // 300
      throw new Meteor.Error(403, "Grain not owned by current user.");
    }                                                                  //
                                                                       //
    Grains.update({ _id: grainId }, { $set: { identityId: identityId } });
  },                                                                   //
                                                                       //
  updateGrainTitle: function (grainId, newTitle, identityId) {         // 307
    check(grainId, String);                                            // 308
    check(newTitle, String);                                           // 309
    check(identityId, String);                                         // 310
    if (this.userId) {                                                 // 311
      var grain = Grains.findOne(grainId);                             // 312
      if (grain) {                                                     // 313
        if (grain.userId === this.userId) {                            // 314
          Grains.update({ _id: grainId, userId: this.userId }, { $set: { title: newTitle } });
        } else {                                                       //
          if (!globalDb.userHasIdentity(this.userId, identityId)) {    // 317
            throw new Meteor.Error(403, "Current user does not have identity " + identityId);
          }                                                            //
                                                                       //
          var token = ApiTokens.findOne({                              // 321
            grainId: grainId,                                          // 322
            objectId: { $exists: false },                              // 323
            "owner.user.identityId": identityId                        // 324
          }, {                                                         //
            sort: { created: 1 }                                       // 326
          });                                                          //
          if (token) {                                                 // 328
            ApiTokens.update(token._id, { $set: { "owner.user.title": newTitle } });
          }                                                            //
        }                                                              //
      }                                                                //
    }                                                                  //
  },                                                                   //
                                                                       //
  privatizeGrain: function (grainId) {                                 // 336
    check(grainId, String);                                            // 337
    if (this.userId) {                                                 // 338
      Grains.update({ _id: grainId, userId: this.userId }, { $set: { "private": true } });
    }                                                                  //
  },                                                                   //
                                                                       //
  inviteUsersToGrain: function (origin, identityId, grainId, title, roleAssignment, contacts, message) {
    var _this2 = this;                                                 //
                                                                       //
    if (!this.isSimulation) {                                          // 345
      var _ret = (function () {                                        //
        check(origin, String);                                         // 346
        check(identityId, String);                                     // 347
        check(grainId, String);                                        // 348
        check(title, String);                                          // 349
        check(roleAssignment, roleAssignmentPattern);                  // 350
        check(contacts, [{                                             // 351
          _id: String,                                                 // 353
          isDefault: Match.Optional(Boolean),                          // 354
          profile: Match.ObjectIncluding({                             // 355
            service: String,                                           // 356
            name: String,                                              // 357
            intrinsicName: String                                      // 358
          })                                                           //
        }]);                                                           //
        check(message, { text: String, html: String });                // 362
        if (!_this2.userId) {                                          // 363
          throw new Meteor.Error(403, "Must be logged in to share by email.");
        }                                                              //
                                                                       //
        if (!globalDb.userHasIdentity(_this2.userId, identityId)) {    // 367
          throw new Meteor.Error(403, "Not an identity of the current user: " + identityId);
        }                                                              //
                                                                       //
        if (contacts.length === 0) {                                   // 371
          throw new Meteor.Error(400, "No contacts were provided.");   // 372
        }                                                              //
                                                                       //
        var accountId = _this2.userId;                                 // 375
        var identity = globalDb.getIdentity(identityId);               // 376
        var sharerDisplayName = identity.profile.name;                 // 377
        var outerResult = { successes: [], failures: [] };             // 378
        contacts.forEach(function (contact) {                          // 379
          if (contact.isDefault && contact.profile.service === "email") {
            var emailAddress = contact.profile.intrinsicName;          // 381
            var result = SandstormPermissions.createNewApiToken(globalDb, { identityId: identityId, accountId: accountId }, grainId, "email invitation for " + emailAddress, roleAssignment, { webkey: { forSharing: true } });
            var url = origin + "/shared/" + result.token;              // 386
            var html = message.html + "<br><br>" + emailLinkWithInlineStyle(url, "Open Shared Grain") + "<div style='font-size:8pt;font-style:italic;color:gray'>" + "Note: If you forward this email to other people, they will be able to access " + "the share as well. To prevent this, remove the button before forwarding.</div>";
            try {                                                      // 392
              SandstormEmail.send({                                    // 393
                to: emailAddress,                                      // 394
                from: "Sandstorm server <no-reply@" + HOSTNAME + ">",  // 395
                subject: sharerDisplayName + " has invited you to join a grain: " + title,
                text: message.text + "\n\nFollow this link to open the shared grain:\n\n" + url + "\n\nNote: If you forward this email to other people, they will be able to " + "access the share as well. To prevent this, remove the link before forwarding.",
                html: html                                             // 400
              });                                                      //
            } catch (e) {                                              //
              outerResult.failures.push({ contact: contact, error: e.toString() });
            }                                                          //
          } else {                                                     //
            var result = SandstormPermissions.createNewApiToken(globalDb, { identityId: identityId, accountId: accountId }, grainId, "direct invitation to " + contact.profile.intrinsicName, roleAssignment, { user: { identityId: contact._id, title: title } });
            var url = origin + "/shared/" + result.token;              // 410
            try {                                                      // 411
              var _identity = Meteor.users.findOne({ _id: contact._id });
              var email = _.findWhere(SandstormDb.getVerifiedEmails(_identity), { primary: true });
              if (email) {                                             // 415
                var intrinsicName = contact.profile.intrinsicName;     // 416
                var loginNote = undefined;                             // 417
                if (contact.profile.service === "google") {            // 418
                  loginNote = "Google account with address " + email.email;
                } else if (contact.profile.service === "github") {     //
                  loginNote = "Github account with username " + intrinsicName;
                } else if (contact.profile.service === "email") {      //
                  loginNote = "email address " + intrinsicName;        // 423
                } else {                                               //
                  throw new Meteor.Error(500, "Unknown service to email share link.");
                }                                                      //
                                                                       //
                var html = message.html + "<br><br>" + emailLinkWithInlineStyle(url, "Open Shared Grain") + "<div style='font-size:8pt;font-style:italic;color:gray'>" + "Note: You will need to log in with your " + loginNote + " to access this grain.";
                SandstormEmail.send({                                  // 433
                  to: email.email,                                     // 434
                  from: "Sandstorm server <no-reply@" + HOSTNAME + ">",
                  subject: sharerDisplayName + " has invited you to join a grain: " + title,
                  text: message.text + "\n\nFollow this link to open the shared grain:\n\n" + url + "\n\nNote: You will need to log in with your " + loginNote + " to access this grain.",
                  html: html                                           // 440
                });                                                    //
              } else {                                                 //
                outerResult.failures.push({ contact: contact, warning: "User does not have a " + "verified email, so notification of this share was not sent to them. Please " + "manually share " + url + " with them." });
              }                                                        //
            } catch (e) {                                              //
              outerResult.failures.push({ contact: contact, error: e.toString(),
                warning: "Share succeeded, but there was an error emailing the user. Please " + "manually share " + url + " with them." });
            }                                                          //
          }                                                            //
        });                                                            //
                                                                       //
        return {                                                       // 455
          v: outerResult                                               //
        };                                                             //
      })();                                                            //
                                                                       //
      if (typeof _ret === "object") return _ret.v;                     //
    }                                                                  //
  },                                                                   //
                                                                       //
  requestAccess: function (origin, grainId, identityId) {              // 459
    check(origin, String);                                             // 460
    check(grainId, String);                                            // 461
    check(identityId, String);                                         // 462
    if (!this.isSimulation) {                                          // 463
      if (!this.userId) {                                              // 464
        throw new Meteor.Error(403, "Must be logged in to request access.");
      }                                                                //
                                                                       //
      if (!globalDb.userHasIdentity(this.userId, identityId)) {        // 468
        throw new Meteor.Error(403, "Not an identity of the current user: " + identityId);
      }                                                                //
                                                                       //
      var grain = Grains.findOne(grainId);                             // 472
      if (!grain) {                                                    // 473
        throw new Meteor.Error(404, "No such grain");                  // 474
      }                                                                //
                                                                       //
      var grainOwner = globalDb.getUser(grain.userId);                 // 477
      var email = _.findWhere(SandstormDb.getUserEmails(grainOwner), { primary: true });
      if (!email) {                                                    // 479
        throw new Meteor.Error("no email", "Grain owner has no email address.");
      }                                                                //
                                                                       //
      var emailAddress = email.email;                                  // 483
                                                                       //
      var identity = globalDb.getIdentity(identityId);                 // 485
      globalDb.addContact(grainOwner._id, identityId);                 // 486
                                                                       //
      var envelopeFrom = globalDb.getReturnAddress();                  // 488
      var fromEmail = globalDb.getServerTitle() + " <" + globalDb.getReturnAddress() + ">";
      var senderEmails = SandstormDb.getVerifiedEmails(identity);      // 490
      var senderPrimaryEmail = _.findWhere(senderEmails, { primary: true });
      var accountPrimaryEmailAddress = Meteor.user().primaryEmail;     // 492
      if (_.findWhere(senderEmails, { email: accountPrimaryEmailAddress })) {
        fromEmail = accountPrimaryEmailAddress;                        // 494
      } else if (senderPrimaryEmail) {                                 //
        fromEmail = senderPrimaryEmail.email;                          // 496
      }                                                                //
                                                                       //
      // TODO(soon): In the HTML version, we should display an identity card.
      var identityNote = "";                                           // 500
      if (identity.profile.service === "google") {                     // 501
        identityNote = " (" + identity.privateIntrinsicName + ")";     // 502
      } else if (identity.profile.service === "github") {              //
        identityNote = " (" + identity.profile.intrinsicName + " on GitHub)";
      } else if (identity.profile.service === "email") {               //
        identityNote = " (" + identity.profile.intrinsicName + ")";    // 506
      }                                                                //
                                                                       //
      var message = identity.profile.name + identityNote + " is requesting access to your grain: " + grain.title + ".";
                                                                       //
      var url = origin + "/share/" + grainId + "/" + identityId;       // 512
                                                                       //
      var html = message + "<br><br>" + emailLinkWithInlineStyle(url, "Open Sharing Menu");
                                                                       //
      var user = Meteor.user();                                        // 517
      var ACCESS_REQUEST_LIMIT = 10;                                   // 518
      var resetCount = true;                                           // 519
      if (user.accessRequests) {                                       // 520
        if (user.accessRequests.resetOn < new Date()) {                // 521
          Meteor.users.update({ _id: user._id }, { $unset: { accessRequests: 1 } });
        } else if (user.accessRequests.count >= ACCESS_REQUEST_LIMIT) {
          throw new Meteor.Error(403, "For spam control reasons, you are not allowed to make " + "more than " + ACCESS_REQUEST_LIMIT + " access requests per day.");
        } else {                                                       //
          resetCount = false;                                          // 528
        }                                                              //
      }                                                                //
                                                                       //
      var modifier = { $inc: { "accessRequests.count": 1 } };          // 532
      if (resetCount) {                                                // 533
        var tomorrow = new Date(Date.now() + 86400000);                // 534
        modifier.$set = { "accessRequests.resetOn": tomorrow };        // 535
      }                                                                //
                                                                       //
      Meteor.users.update({ _id: user._id }, modifier);                // 538
                                                                       //
      SandstormEmail.send({                                            // 540
        to: emailAddress,                                              // 541
        envelopeFrom: envelopeFrom,                                    // 542
        from: fromEmail,                                               // 543
        subject: grain.title + " - Request for access",                // 544
        text: message + "\n\nFollow this link to share access:\n\n" + url,
        html: html                                                     // 546
      });                                                              //
    }                                                                  //
  }                                                                    //
});                                                                    //
/////////////////////////////////////////////////////////////////////////

}).call(this);

//# sourceMappingURL=grain-server.js.map
