(function(){

/////////////////////////////////////////////////////////////////////////
//                                                                     //
// server/install-server.js                                            //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
                                                                       //
// Sandstorm - Personal Cloud Sandbox                                  //
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors
// All rights reserved.                                                //
//                                                                     //
// Licensed under the Apache License, Version 2.0 (the "License");     //
// you may not use this file except in compliance with the License.    //
// You may obtain a copy of the License at                             //
//                                                                     //
//   http://www.apache.org/licenses/LICENSE-2.0                        //
//                                                                     //
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,   //
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.                                      //
                                                                       //
var localizedTextPattern = {                                           // 17
  defaultText: String,                                                 // 18
  localizations: Match.Optional([{ locale: String, text: String }])    // 19
};                                                                     //
                                                                       //
var uploadTokens = {};                                                 // 22
                                                                       //
UserActions.allow({                                                    // 24
  insert: function (userId, action) {                                  // 25
    // TODO(cleanup): This check keeps breaking. Use a method instead that takes the package
    //   ID as an argument.                                            //
    check(action, {                                                    // 28
      userId: String,                                                  // 29
      packageId: String,                                               // 30
      appId: String,                                                   // 31
      appTitle: Match.Optional(localizedTextPattern),                  // 32
      appMarketingVersion: Match.Optional(Object),                     // 33
      appVersion: Match.Integer,                                       // 34
      title: localizedTextPattern,                                     // 35
      nounPhrase: Match.Optional(localizedTextPattern),                // 36
      command: {                                                       // 37
        executablePath: Match.Optional(String),                        // 38
        deprecatedExecutablePath: Match.Optional(String),              // 39
        args: Match.Optional([String]),                                // 40
        argv: Match.Optional([String]),                                // 41
        environ: Match.Optional([{ key: String, value: String }])      // 42
      }                                                                //
    });                                                                //
    return userId && isSignedUpOrDemo() && action.userId === userId;   // 45
  },                                                                   //
                                                                       //
  remove: function (userId, action) {                                  // 48
    return userId && action.userId === userId;                         // 49
  }                                                                    //
});                                                                    //
                                                                       //
// Not all users are allowed to upload apps. We need to manually implement authorization
// because Meteor.userId() is not available in server-side routes.     //
                                                                       //
Meteor.methods({                                                       // 56
  cancelDownload: function (packageId) {                               // 57
    check(packageId, String);                                          // 58
                                                                       //
    // TODO(security):  Only let user cancel download if they initiated it.
    cancelDownload(packageId);                                         // 61
  },                                                                   //
                                                                       //
  newUploadToken: function () {                                        // 64
    if (!isSignedUp()) {                                               // 65
      throw new Meteor.Error(403, "Unauthorized", "Only invited users can upload apps.");
    }                                                                  //
                                                                       //
    if (globalDb.isUninvitedFreeUser()) {                              // 69
      throw new Meteor.Error(403, "Unauthorized", "Only paid users can upload apps.");
    }                                                                  //
                                                                       //
    var token = Random.id();                                           // 73
    uploadTokens[token] = setTimeout(function () {                     // 74
      delete uploadTokens[token];                                      // 75
    }, 20 * 60 * 1000);                                                //
                                                                       //
    return token;                                                      // 78
  },                                                                   //
                                                                       //
  upgradeGrains: function (appId, version, packageId) {                // 81
    this.connection.sandstormDb.upgradeGrains(appId, version, packageId, globalBackend);
  },                                                                   //
                                                                       //
  ensureInstalled: function (packageId, url, isRetry) {                // 85
    check(packageId, String);                                          // 86
    check(url, Match.OneOf(String, undefined, null));                  // 87
    check(isRetry, Boolean);                                           // 88
                                                                       //
    if (!packageId.match(/^[a-zA-Z0-9]*$/)) {                          // 90
      throw new Meteor.Error(400, "The package ID contains illegal characters.");
    }                                                                  //
                                                                       //
    if (!this.userId) {                                                // 94
      if (allowDemo && isSafeDemoAppUrl(url)) {                        // 95
        // continue on                                                 //
      } else {                                                         //
          // jscs:ignore disallowEmptyBlocks                           //
          throw new Meteor.Error(403, "You must be logged in to install packages.");
        }                                                              //
    } else if (!isSignedUp() && !isDemoUser()) {                       //
      throw new Meteor.Error(403, "This Sandstorm server requires you to get an invite before installing apps.");
    } else if (isUserOverQuota(Meteor.user())) {                       //
      throw new Meteor.Error(402, "You are out of storage space. Please delete some things and try again.");
    }                                                                  //
                                                                       //
    var pkg = Packages.findOne(packageId);                             // 108
                                                                       //
    if (!pkg || pkg.status !== "ready") {                              // 110
      if (!this.userId || isDemoUser() || globalDb.isUninvitedFreeUser()) {
        if (!isSafeDemoAppUrl(url)) {                                  // 112
          // TODO(someday): Billing prompt on client side.             //
          throw new Meteor.Error(403, "Sorry, demo and free users cannot upload custom apps; " + "they may only install apps from apps.sandstorm.io.");
        }                                                              //
      }                                                                //
    }                                                                  //
                                                                       //
    if (!pkg || isRetry) {                                             // 120
      globalDb.startInstall(packageId, url, isRetry);                  // 121
    }                                                                  //
  }                                                                    //
});                                                                    //
                                                                       //
Router.map(function () {                                               // 126
  this.route("upload", {                                               // 127
    where: "server",                                                   // 128
    path: "/upload/:token",                                            // 129
                                                                       //
    action: function () {                                              // 131
      if (!this.params.token || !uploadTokens[this.params.token]) {    // 132
        this.response.writeHead(403, {                                 // 133
          "Content-Type": "text/plain"                                 // 134
        });                                                            //
        this.response.write("Invalid upload token.");                  // 136
        this.response.end();                                           // 137
      } else if (this.request.method === "POST") {                     //
        try {                                                          // 139
          var packageId = promiseToFuture(doClientUpload(this.request)).wait();
          this.response.writeHead(200, {                               // 141
            "Content-Length": packageId.length,                        // 142
            "Content-Type": "text/plain"                               // 143
          });                                                          //
          this.response.write(packageId);                              // 145
          this.response.end();                                         // 146
          clearTimeout(uploadTokens[this.params.token]);               // 147
          delete uploadTokens[this.params.token];                      // 148
        } catch (error) {                                              //
          console.error(error.stack);                                  // 150
          this.response.writeHead(500, {                               // 151
            "Content-Type": "text/plain"                               // 152
          });                                                          //
          this.response.write("Unpacking SPK failed; is it valid?");   // 154
          this.response.end();                                         // 155
        };                                                             //
      } else {                                                         //
        this.response.writeHead(405, {                                 // 158
          "Content-Type": "text/plain"                                 // 159
        });                                                            //
        this.response.write("You can only POST here.");                // 161
        this.response.end();                                           // 162
      }                                                                //
    }                                                                  //
  });                                                                  //
});                                                                    //
/////////////////////////////////////////////////////////////////////////

}).call(this);

//# sourceMappingURL=install-server.js.map
