(function(){

/////////////////////////////////////////////////////////////////////////
//                                                                     //
// server/proxy.js                                                     //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
                                                                       //
// Sandstorm - Personal Cloud Sandbox                                  //
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors
// All rights reserved.                                                //
//                                                                     //
// Licensed under the Apache License, Version 2.0 (the "License");     //
// you may not use this file except in compliance with the License.    //
// You may obtain a copy of the License at                             //
//                                                                     //
//   http://www.apache.org/licenses/LICENSE-2.0                        //
//                                                                     //
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,   //
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.                                      //
                                                                       //
var Crypto = Npm.require("crypto");                                    // 17
var ChildProcess = Npm.require("child_process");                       // 18
var Fs = Npm.require("fs");                                            // 19
var Path = Npm.require("path");                                        // 20
var Future = Npm.require("fibers/future");                             // 21
var Http = Npm.require("http");                                        // 22
var Url = Npm.require("url");                                          // 23
var Promise = Npm.require("es6-promise").Promise;                      // 24
var Capnp = Npm.require("capnp");                                      // 25
var Net = Npm.require("net");                                          // 26
                                                                       //
var ByteStream = Capnp.importSystem("sandstorm/util.capnp").ByteStream;
var ApiSession = Capnp.importSystem("sandstorm/api-session.capnp").ApiSession;
var WebSession = Capnp.importSystem("sandstorm/web-session.capnp").WebSession;
var HackSession = Capnp.importSystem("sandstorm/hack-session.capnp");  // 31
var Supervisor = Capnp.importSystem("sandstorm/supervisor.capnp").Supervisor;
var Backend = Capnp.importSystem("sandstorm/backend.capnp").Backend;   // 33
                                                                       //
var storeReferralProgramInfoApiTokenCreated = function (db, accountId, identityId, apiTokenAccountId) {
  // From the Referral program's perspective, if Bob's Account has no referredByComplete, then we
  // update Bob's Identity to say it's referredBy Alice's Account (which is apiTokenAccountId).
  check(accountId, String);                                            // 38
  check(identityId, String);                                           // 39
  check(apiTokenAccountId, String);                                    // 40
                                                                       //
  // Bail out early if quota enforcement is disabled.                  //
  if (!Meteor.settings["public"].quotaEnabled) {                       // 43
    return;                                                            // 44
  }                                                                    //
                                                                       //
  var aliceAccountId = apiTokenAccountId;                              // 47
  var bobAccountId = accountId;                                        // 48
  var bobIdentityId = identityId;                                      // 49
                                                                       //
  if (Meteor.users.find({                                              // 51
    _id: bobAccountId,                                                 // 52
    referredByComplete: { $exists: true }                              // 53
  }).count() > 0) {                                                    //
    return;                                                            // 55
  }                                                                    //
                                                                       //
  // Only actually update Bob's Identity ID if there is no referredBy.
  Meteor.users.update({ _id: bobIdentityId, referredBy: { $exists: false } }, { $set: { referredBy: aliceAccountId } });
};                                                                     //
                                                                       //
function referralProgramLogSharingTokenUse(db, bobAccountId) {         // 64
  // Hooray! The sharing token is valid! Someone (let's call them Charlie) is going to get a UiView
  // to this grain!  This means that the user who created this apiToken knows how to use the 'share
  // access' interface. Let's call them Bob.                           //
  //                                                                   //
  // If Bob's Account.referredByComplete is not yet set, then look at Bob's Identities and take the
  // first referredBy we find -- let's call that Alice.                //
  //                                                                   //
  // We copy Alice's account ID to Bob's Account.referredByComplete, and then update Alice's
  // referredIdentityIds to point at Bob's Identity, and then remove the referredBy from Bob's
  // Identity since it has become redundant.                           //
  //                                                                   //
  // Implementation note: this does mean that Alice can get referral credit for Bob by sharing a
  // link with Bob, even if Bob already had an account.                //
                                                                       //
  // Bail out early if quota support is not enabled.                   //
  if (!Meteor.settings["public"].quotaEnabled) {                       // 80
    return;                                                            // 81
  }                                                                    //
                                                                       //
  // Bail out if Bob has a referredByComplete.                         //
  if (Meteor.users.find({ _id: bobAccountId, referredByComplete: { $exists: true } }).count() > 0) {
    return;                                                            // 86
  }                                                                    //
                                                                       //
  // Look for a referredBy on any of Bob's identities.                 //
  var bobIdentityIds = SandstormDb.getUserIdentityIds(Meteor.users.findOne({ _id: bobAccountId }));
  var bobIdentityWithReferredBy = Meteor.users.findOne({               // 91
    _id: { $in: bobIdentityIds },                                      // 92
    referredBy: { $exists: true }                                      // 93
  });                                                                  //
                                                                       //
  if (!bobIdentityWithReferredBy) {                                    // 96
    return;                                                            // 97
  }                                                                    //
                                                                       //
  var aliceAccountId = bobIdentityWithReferredBy.referredBy;           // 100
                                                                       //
  // Store Bob's Account.referralCompletedBy.                          //
  var now = new Date();                                                // 103
  Meteor.users.update({                                                // 104
    _id: bobAccountId,                                                 // 105
    referredByComplete: { $exists: false }                             // 106
  }, {                                                                 //
    $set: {                                                            // 108
      referredByComplete: bobIdentityWithReferredBy.referredBy,        // 109
      referredCompleteDate: now                                        // 110
    }                                                                  //
  });                                                                  //
                                                                       //
  // Update Alice's Account.referredIdentityIds.                       //
  Meteor.users.update({ _id: aliceAccountId }, {                       // 115
    $push: { referredIdentityIds: bobIdentityWithReferredBy._id }      // 116
  });                                                                  //
                                                                       //
  // Remove now-redundant Bob identity referredBy.                     //
  Meteor.users.update({ _id: bobIdentityWithReferredBy._id }, {        // 120
    $unset: { referredBy: true }                                       // 121
  });                                                                  //
}                                                                      //
                                                                       //
// User-agent strings that should be allowed to use http basic authentication.
// These are regex matches, so ensure they are escaped properly with double
// backslashes. For security reasons, we MUST NOT whitelist any user-agents
// that may render html and execute embedded scripts.                  //
BASIC_AUTH_USER_AGENTS = ["git\\/", "GitHub-Hookshot\\/", "mirall\\/", "Mozilla\\/5\\.0 \\([^\\\\]*\\) mirall\\/", "Mozilla\\/5\\.0 \\(iOS\\) ownCloud-iOS\\/", "Mozilla\\/5\\.0 \\(Android\\) ownCloud-android\\/", "litmus\\/"];
BASIC_AUTH_USER_AGENTS_REGEX = new RegExp("^(" + BASIC_AUTH_USER_AGENTS.join("|") + ")", "");
                                                                       //
var SESSION_PROXY_TIMEOUT = 60000;                                     // 140
                                                                       //
var sandstormCoreFactory = makeSandstormCoreFactory();                 // 142
var backendAddress = "unix:" + (SANDSTORM_ALTHOME || "") + Backend.socketPath;
var sandstormBackendConnection = Capnp.connect(backendAddress, sandstormCoreFactory);
var sandstormBackend = sandstormBackendConnection.restore(null, Backend);
                                                                       //
// TODO(cleanup): This initilization belongs with the rest of our package initialization in
//   db-deprecated.js. We can't put it there now because we need to contruct sandstormCoreFactory first.
globalBackend = new SandstormBackend(globalDb, sandstormBackend);      // 149
Meteor.onConnection(function (connection) {                            // 150
  connection.sandstormBackend = globalBackend;                         // 151
});                                                                    //
                                                                       //
// We've observed a problem in production where occasionally the front-end stops talking to the
// back-end. It happens very rarely -- like once a month -- and we've been unable to reproduce it
// in testing, making it very hard to debug. The problem appears both on Sandstorm and Blackrock.
// Restarting the node process (and only the node process) always fixes the problem.
//                                                                     //
// Here, I've added some code that attempts to detect the problem by doing a health check
// periodically and then remaking the connection if it seems broken. We'll see if this helps!
var backendHealthy = true;                                             // 161
var disconnectCount = 0;                                               // 162
Meteor.setInterval(function () {                                       // 163
  if (!backendHealthy) {                                               // 164
    if (disconnectCount++ > 2) process.abort();                        // 165
    console.error("error: Backend hasn't responded in 30 seconds! Reconnecting.");
    if (Capnp.enableVerboseDebugLogging) Capnp.enableVerboseDebugLogging(true);
    sandstormBackendConnection.close();                                // 168
    sandstormBackendConnection = Capnp.connect(backendAddress, sandstormCoreFactory);
    sandstormBackend = sandstormBackendConnection.restore(null, Backend);
    globalBackend._backendCap = sandstormBackend;                      // 171
  }                                                                    //
                                                                       //
  var debugLog = !backendHealthy;                                      // 174
  backendHealthy = false;                                              // 175
  var promise = sandstormBackend.ping().then(function () {             // 176
    backendHealthy = true;                                             // 177
    if (Capnp.enableVerboseDebugLogging) Capnp.enableVerboseDebugLogging(false);
  }, function (err) {                                                  //
    console.error("error: Backend ping threw error!", err.stack);      // 180
    // The connection will be remade on the next interval. Note that we do NOT normally observe
    // exceptions being thrown for this problem; we see the connection simply stop responding.
    // So we don't expect this branch to execute in any case.          //
  });                                                                  //
                                                                       //
  if (debugLog) {                                                      // 186
    console.log("capnp.js: outer promise:", promise);                  // 187
  }                                                                    //
}, 30000);                                                             //
                                                                       //
// =======================================================================================
// Meteor context <-> Async Node.js context adapters                   //
// TODO(cleanup):  Move to a different file.                           //
                                                                       //
var inMeteorInternal = Meteor.bindEnvironment(function (callback) {    // 195
  callback();                                                          // 196
});                                                                    //
                                                                       //
inMeteor = function (callback) {                                       // 199
  // Calls the callback in a Meteor context.  Returns a Promise for its result.
  return new Promise(function (resolve, reject) {                      // 201
    inMeteorInternal(function () {                                     // 202
      try {                                                            // 203
        resolve(callback());                                           // 204
      } catch (err) {                                                  //
        reject(err);                                                   // 206
      }                                                                //
    });                                                                //
  });                                                                  //
};                                                                     //
                                                                       //
promiseToFuture = function (promise) {                                 // 212
  var result = new Future();                                           // 213
  promise.then(result["return"].bind(result), result["throw"].bind(result));
  return result;                                                       // 215
};                                                                     //
                                                                       //
waitPromise = function (promise) {                                     // 218
  return promiseToFuture(promise).wait();                              // 219
};                                                                     //
                                                                       //
// =======================================================================================
// API for creating / starting grains from Meteor methods.             //
                                                                       //
var proxiesByHostId = {};                                              // 225
                                                                       //
Meteor.methods({                                                       // 227
  newGrain: function (packageId, command, title, identityId) {         // 228
    // Create and start a new grain.                                   //
                                                                       //
    check(packageId, String);                                          // 231
    check(command, Object); // Manifest.Command from package.capnp.    // 232
    check(title, String);                                              // 233
    check(identityId, String);                                         // 234
                                                                       //
    if (!this.userId) {                                                // 236
      throw new Meteor.Error(403, "Unauthorized", "Must be logged in to create grains.");
    }                                                                  //
                                                                       //
    if (!globalDb.userHasIdentity(this.userId, identityId)) {          // 240
      throw new Meteor.Error(403, "Current user does not own the identity: " + identityId);
    }                                                                  //
                                                                       //
    if (!isSignedUpOrDemo()) {                                         // 244
      throw new Meteor.Error(403, "Unauthorized", "Only invited users or demo users can create grains.");
    }                                                                  //
                                                                       //
    if (isUserOverQuota(Meteor.user())) {                              // 249
      throw new Meteor.Error(402, "You are out of storage space. Please delete some things and try again.");
    }                                                                  //
                                                                       //
    var pkg = Packages.findOne(packageId);                             // 254
    var isDev = false;                                                 // 255
    if (!pkg) {                                                        // 256
      // Maybe they wanted a dev package.  Check there too.            //
      pkg = DevPackages.findOne(packageId);                            // 258
      isDev = true;                                                    // 259
    }                                                                  //
                                                                       //
    if (!pkg) {                                                        // 262
      throw new Meteor.Error(404, "Not Found", "No such package is installed.");
    }                                                                  //
                                                                       //
    var appId = pkg.appId;                                             // 266
    var manifest = pkg.manifest;                                       // 267
    var grainId = Random.id(22); // 128 bits of entropy                // 268
    Grains.insert({                                                    // 269
      _id: grainId,                                                    // 270
      packageId: packageId,                                            // 271
      appId: appId,                                                    // 272
      appVersion: manifest.appVersion,                                 // 273
      userId: this.userId,                                             // 274
      identityId: identityId,                                          // 275
      title: title,                                                    // 276
      "private": true                                                  // 277
    });                                                                //
                                                                       //
    globalBackend.startGrainInternal(packageId, grainId, this.userId, command, true, isDev);
    globalBackend.updateLastActive(grainId, this.userId, identityId);  // 281
    return grainId;                                                    // 282
  },                                                                   //
                                                                       //
  openSession: function (grainId, identityId, cachedSalt) {            // 285
    // Open a new UI session on an existing grain.  Starts the grain if it is not already
    // running.                                                        //
                                                                       //
    check(grainId, String);                                            // 289
    check(identityId, Match.OneOf(undefined, null, String));           // 290
    check(cachedSalt, Match.OneOf(undefined, null, String));           // 291
                                                                       //
    if (this.userId && identityId && !globalDb.userHasIdentity(this.userId, identityId)) {
      throw new Meteor.Error(403, "Current user does not own the identity: " + identityId);
    }                                                                  //
                                                                       //
    if (!Grains.findOne({ _id: grainId })) {                           // 297
      throw new Meteor.Error(404, "Grain not found", "Grain ID: " + grainId);
    }                                                                  //
                                                                       //
    var db = this.connection.sandstormDb;                              // 301
    check(cachedSalt, Match.OneOf(undefined, null, String));           // 302
    if (!SandstormPermissions.mayOpenGrain(db, { grain: { _id: grainId, identityId: identityId } })) {
      throw new Meteor.Error(403, "Unauthorized", "User is not authorized to open this grain.");
    }                                                                  //
                                                                       //
    var opened = globalBackend.openSessionInternal(grainId, this.userId, identityId, null, null, cachedSalt);
    var result = opened.methodResult;                                  // 310
    var proxy = new Proxy(grainId, this.userId, result.sessionId, result.hostId, identityId, false, opened.supervisor);
    proxiesByHostId[result.hostId] = proxy;                            // 313
    return result;                                                     // 314
  },                                                                   //
                                                                       //
  openSessionFromApiToken: function (params, identityId, cachedSalt) {
    // Given an API token, either opens a new WebSession to the underlying grain or returns a
    // path to which the client should redirect in order to open such a session.
                                                                       //
    check(params, {                                                    // 321
      token: String,                                                   // 322
      incognito: Boolean                                               // 323
    });                                                                //
    check(identityId, Match.OneOf(undefined, null, String));           // 325
    check(cachedSalt, Match.OneOf(undefined, null, String));           // 326
                                                                       //
    if (this.userId && identityId && !globalDb.userHasIdentity(this.userId, identityId)) {
      throw new Meteor.Error(403, "Current user does not own the identity: " + identityId);
    }                                                                  //
                                                                       //
    var token = params.token;                                          // 332
    var incognito = params.incognito;                                  // 333
    var hashedToken = Crypto.createHash("sha256").update(token).digest("base64");
    var apiToken = ApiTokens.findOne(hashedToken);                     // 335
    validateWebkey(apiToken);                                          // 336
    var grain = Grains.findOne({ _id: apiToken.grainId });             // 337
    if (!grain) {                                                      // 338
      throw new Meteor.Error(404, "Grain not found", "Grain ID: " + apiToken.grainId);
    }                                                                  //
                                                                       //
    if (apiToken.accountId) {                                          // 342
      referralProgramLogSharingTokenUse(globalDb, apiToken.accountId);
    }                                                                  //
                                                                       //
    var title = undefined;                                             // 346
    if (grain.userId === apiToken.accountId) {                         // 347
      title = grain.title;                                             // 348
    } else {                                                           //
      var sharerToken = apiToken.identityId && ApiTokens.findOne({     // 350
        grainId: apiToken.grainId,                                     // 352
        "owner.user.identityId": apiToken.identityId                   // 353
      }, {                                                             //
        sort: {                                                        // 355
          "owner.user.lastUsed": -1                                    // 356
        }                                                              //
      });                                                              //
      if (sharerToken) {                                               // 359
        title = sharerToken.owner.user.title;                          // 360
      } else {                                                         //
        title = "shared grain";                                        // 362
      }                                                                //
    }                                                                  //
                                                                       //
    if (this.userId && !incognito) {                                   // 366
      if (identityId != apiToken.identityId && identityId != grain.identityId && !ApiTokens.findOne({ "owner.user.identityId": identityId, parentToken: hashedToken })) {
        var owner = { user: { identityId: identityId, title: title } };
                                                                       //
        // Create a new API token for the identity redeeming this token.
        var result = SandstormPermissions.createNewApiToken(globalDb, { rawParentToken: token }, apiToken.grainId, apiToken.petname, { allAccess: null }, owner);
        globalDb.addContact(apiToken.accountId, identityId);           // 374
                                                                       //
        // If the parent API token is forSharing and it has an accountId, then the logged-in user (call
        // them Bob) is about to access a grain owned by someone (call them Alice) and save a reference
        // to it as a new ApiToken. (For share-by-link, this occurs when viewing the grain. For
        // share-by-identity, this happens immediately.)               //
        if (result.parentApiToken) {                                   // 380
          var parentApiToken = result.parentApiToken;                  // 381
          if (parentApiToken.forSharing && parentApiToken.accountId) {
            storeReferralProgramInfoApiTokenCreated(globalDb, this.userId, owner.user.identityId, parentApiToken.accountId);
          }                                                            //
        }                                                              //
      }                                                                //
                                                                       //
      return { redirectToGrain: apiToken.grainId };                    // 389
    } else {                                                           //
      if (!SandstormPermissions.mayOpenGrain(globalDb, { token: apiToken })) {
        throw new Meteor.Error(403, "Unauthorized", "User is not authorized to open this grain.");
      }                                                                //
                                                                       //
      var opened = globalBackend.openSessionInternal(apiToken.grainId, null, null, title, apiToken, cachedSalt);
                                                                       //
      var result = opened.methodResult;                                // 399
      var proxy = new Proxy(apiToken.grainId, grain.userId, result.sessionId, result.hostId, identityId, false);
      proxy.apiToken = apiToken;                                       // 402
      proxiesByHostId[result.hostId] = proxy;                          // 403
      return result;                                                   // 404
    }                                                                  //
  },                                                                   //
                                                                       //
  keepSessionAlive: function (sessionId) {                             // 408
    // TODO(security):  Prevent draining someone else's quota by holding open several grains shared
    //   by them.                                                      //
    check(sessionId, String);                                          // 411
                                                                       //
    var session = Sessions.findAndModify({                             // 413
      query: { _id: sessionId },                                       // 414
      update: { $set: { timestamp: new Date().getTime() } },           // 415
      fields: { grainId: 1, identityId: 1, hostId: 1 }                 // 416
    });                                                                //
                                                                       //
    if (session) {                                                     // 419
      // Session still present in database, so send keep-alive to backend.
      try {                                                            // 421
        var grainId = session.grainId;                                 // 422
        var hostId = session.hostId;                                   // 423
        var supervisor = proxiesByHostId[hostId] && proxiesByHostId[hostId].supervisor;
        if (!supervisor) {                                             // 425
          supervisor = globalBackend.continueGrain(grainId).supervisor;
        }                                                              //
                                                                       //
        waitPromise(supervisor.keepAlive());                           // 429
        globalBackend.updateLastActive(grainId, this.userId, session.identityId);
      } catch (err) {                                                  //
        // Ignore disconnects, which imply that the grain shut down already. It'll start back up on
        // the next request, so whatever.                              //
        if (err.kjType !== "disconnected") {                           // 434
          throw err;                                                   // 435
        }                                                              //
      }                                                                //
                                                                       //
      return true;                                                     // 439
    } else {                                                           //
      return false;                                                    // 441
    }                                                                  //
  },                                                                   //
                                                                       //
  shutdownGrain: function (grainId) {                                  // 445
    check(grainId, String);                                            // 446
    var grain = Grains.findOne(grainId);                               // 447
    if (!grain || !this.userId || grain.userId !== this.userId) {      // 448
      throw new Meteor.Error(403, "Unauthorized", "User is not the owner of this grain");
    }                                                                  //
                                                                       //
    waitPromise(globalBackend.shutdownGrain(grainId, grain.userId, true));
  }                                                                    //
});                                                                    //
                                                                       //
var validateWebkey = function (apiToken, refreshedExpiration) {        // 456
  // Validates that `apiToken` is a valid UiView webkey, throwing an exception if it is not. If
  // `refreshedExpiration` is set and if the token has an `expiresIfUnused` field, then the
  // `expiresIfUnused` field is reset to `refreshedExpiration`.        //
                                                                       //
  if (!apiToken) {                                                     // 461
    throw new Meteor.Error(403, "Invalid authorization token");        // 462
  }                                                                    //
                                                                       //
  if (apiToken.revoked) {                                              // 465
    throw new Meteor.Error(403, "Authorization token has been revoked");
  }                                                                    //
                                                                       //
  if (apiToken.owner && !("webkey" in apiToken.owner)) {               // 469
    throw new Meteor.Error(403, "Unauthorized to open non-webkey token.");
  }                                                                    //
                                                                       //
  if (apiToken.expires && apiToken.expires.getTime() <= Date.now()) {  // 473
    throw new Meteor.Error(403, "Authorization token expired");        // 474
  }                                                                    //
                                                                       //
  if (apiToken.expiresIfUnused) {                                      // 477
    if (apiToken.expiresIfUnused.getTime() <= Date.now()) {            // 478
      throw new Meteor.Error(403, "Authorization token expired");      // 479
    } else if (refreshedExpiration) {                                  //
      ApiTokens.update(apiToken._id, { $set: { expiresIfUnused: refreshedExpiration } });
    } else {                                                           //
      // It's getting used now, so clear the expiresIfUnused field.    //
      ApiTokens.update(apiToken._id, { $set: { expiresIfUnused: null } });
    }                                                                  //
  }                                                                    //
                                                                       //
  if (apiToken.objectId || apiToken.frontendRef) {                     // 488
    throw new Meteor.Error(403, "ApiToken refers to a non-webview Capability.");
  }                                                                    //
};                                                                     //
                                                                       //
// Used by shared/grain.js (which sounds like a broken dependency, since this code is only available
// on the server)                                                      //
getGrainSize = function (supervisor, oldSize) {                        // 495
  var promise = undefined;                                             // 496
  if (oldSize === undefined) {                                         // 497
    promise = supervisor.getGrainSize();                               // 498
  } else {                                                             //
    promise = supervisor.getGrainSizeWhenDifferent(oldSize);           // 500
  }                                                                    //
                                                                       //
  var promise2 = promise.then(function (result) {                      // 503
    return parseInt(result.size);                                      // 503
  });                                                                  //
  promise2.cancel = function () {                                      // 504
    promise.cancel();                                                  // 504
  };                                                                   //
                                                                       //
  return promise2;                                                     // 506
};                                                                     //
                                                                       //
Meteor.startup(function () {                                           // 509
  var shutdownApp = function (appId) {                                 // 510
    Grains.find({ appId: appId }).forEach(function (grain) {           // 511
      waitPromise(globalBackend.shutdownGrain(grain._id, grain.userId));
    });                                                                //
  };                                                                   //
                                                                       //
  DevPackages.find().observe({                                         // 516
    removed: function (devPackage) {                                   // 517
      shutdownApp(devPackage.appId);                                   // 517
    },                                                                 //
                                                                       //
    changed: function (oldDevPackage, newDevPackage) {                 // 519
      shutdownApp(oldDevPackage.appId);                                // 520
      if (oldDevPackage.appId !== newDevPackage.appId) {               // 521
        shutdownApp(newDevPackage.appId);                              // 522
      }                                                                //
    },                                                                 //
                                                                       //
    added: function (devPackage) {                                     // 526
      shutdownApp(devPackage.appId);                                   // 526
    }                                                                  //
  });                                                                  //
                                                                       //
  Sessions.find().observe({                                            // 529
    removed: function (session) {                                      // 530
      var proxy = proxiesByHostId[session.hostId];                     // 531
      delete proxiesByHostId[session.hostId];                          // 532
      if (proxy) {                                                     // 533
        proxy.close();                                                 // 534
      }                                                                //
    }                                                                  //
  });                                                                  //
});                                                                    //
                                                                       //
// Kill off sessions idle for >~3 minutes.                             //
var TIMEOUT_MS = 180000;                                               // 541
var gcSessions = function () {                                         // 542
  var now = new Date().getTime();                                      // 543
  Sessions.remove({ timestamp: { $lt: now - TIMEOUT_MS } });           // 544
};                                                                     //
                                                                       //
SandstormDb.periodicCleanup(TIMEOUT_MS, gcSessions);                   // 547
                                                                       //
var getProxyForHostId = function (hostId, isAlreadyOpened) {           // 549
  // Get the Proxy corresponding to the given grain session host, possibly (re)creating it if it
  // doesn't already exist. The first request on the session host will always create a new proxy.
  // Later requests may create a proxy if they go to a different front-end replica or if the
  // front-end was restarted.                                          //
  check(hostId, String);                                               // 554
                                                                       //
  return Promise.resolve(undefined).then(function () {                 // 556
    var proxy = proxiesByHostId[hostId];                               // 557
    if (proxy) {                                                       // 558
      return proxy;                                                    // 559
    } else {                                                           //
      // Set table entry to null for now so that we can detect if it is concurrently deleted.
      proxiesByHostId[hostId] = null;                                  // 562
                                                                       //
      return inMeteor(function () {                                    // 564
        var session = Sessions.findOne({ hostId: hostId });            // 565
        if (!session) {                                                // 566
          if (isAlreadyOpened) {                                       // 567
            return new Promise(function (resolve, reject) {            // 568
              var observer = undefined;                                // 569
              var task = Meteor.setTimeout(function () {               // 570
                observer.stop();                                       // 571
                reject(new Meteor.Error(504, "Requested session that no longer exists, and " + "timed out waiting for client to restore it. This can happen if you have " + "opened an app's content in a new window and then closed it in the " + "UI. If you see this error *inside* the Sandstorm UI, please report a " + "bug and describe the circumstances of the error."));
              }, SESSION_PROXY_TIMEOUT);                               //
              observer = Sessions.find({ hostId: hostId }).observe({   // 578
                added: function () {                                   // 579
                  observer.stop();                                     // 580
                  Meteor.clearTimeout(task);                           // 581
                  resolve(getProxyForHostId(hostId, false));           // 582
                }                                                      //
              });                                                      //
            });                                                        //
          } else {                                                     //
            // Does not appear to be a valid session host.             //
            return undefined;                                          // 588
          }                                                            //
        }                                                              //
                                                                       //
        var apiToken = undefined;                                      // 592
        if (session.hashedToken) {                                     // 593
          apiToken = ApiTokens.findOne({ _id: session.hashedToken });  // 594
          // We don't have to fully validate the API token here because if it changed the session
          // would have been deleted.                                  //
          if (!apiToken) {                                             // 597
            throw new Meteor.Error(410, "ApiToken has been deleted");  // 598
          }                                                            //
        }                                                              //
                                                                       //
        var grain = Grains.findOne(session.grainId);                   // 602
        if (!grain) {                                                  // 603
          // Grain was deleted, I guess.                               //
          throw new Meteor.Error(410, "Resource has been deleted");    // 605
        }                                                              //
                                                                       //
        // Note that we don't need to call mayOpenGrain() because the existence of a session
        // implies this check was already performed.                   //
                                                                       //
        var proxy = new Proxy(grain._id, grain.userId, session._id, hostId, session.identityId, false);
        if (apiToken) proxy.apiToken = apiToken;                       // 612
                                                                       //
        // Only add the proxy to the table if it was not concurrently deleted (which could happen
        // e.g. if the user's access was revoked).                     //
        if (hostId in proxiesByHostId) {                               // 616
          proxiesByHostId[hostId] = proxy;                             // 617
        } else {                                                       //
          throw new Meteor.Error(403, "Session was concurrently closed.");
        }                                                              //
                                                                       //
        return proxy;                                                  // 622
      });                                                              //
    }                                                                  //
  });                                                                  //
};                                                                     //
                                                                       //
// =======================================================================================
// API tokens                                                          //
                                                                       //
var ApiSessionProxies = (function () {                                 //
  // Class that caches proxies for requests that come in through HTTP API endpoints. Such
  // requests are not associated with any entry in the `Sessions` collection, so we need
  // some other means of detecting unused proxies and closing them. This class works by keeping
  // the proxies in two buckets. When a proxy is used, it is put in the new bucket. Periodically,
  // the old bucket is emptied and the contents of new bucket are moved to the old bucket.
  //                                                                   //
  // Each bucket is a two-level map, keyed by hashes of the tokens and their `ApiSession.Params`
  // structs.                                                          //
                                                                       //
  function ApiSessionProxies(intervalMillis) {                         // 641
    var _this2 = this;                                                 //
                                                                       //
    babelHelpers.classCallCheck(this, ApiSessionProxies);              //
                                                                       //
    check(intervalMillis, Number);                                     // 642
    this.newBucket = {};                                               // 643
    this.oldBucket = {};                                               // 644
    this.interval = Meteor.setInterval(function () {                   // 645
      for (var oldHashedToken in babelHelpers.sanitizeForInObject(_this2.oldBucket)) {
        for (var oldHashedParams in babelHelpers.sanitizeForInObject(_this2.oldBucket[oldHashedToken])) {
          var newProxy = _this2.newBucket[oldHashedToken] && _this2.newBucket[oldHashedToken][oldHashedParams];
          var oldProxy = _this2.oldBucket[oldHashedToken][oldHashedParams];
          if (oldProxy && !newProxy) {                                 // 651
            if (Object.keys(oldProxy.websockets).length > 0) {         // 652
              // A client has an open websocket. Keep the proxy around.
              _this2.put(oldHashedToken, oldHashedParams, oldProxy);   // 654
            } else {                                                   //
              // We can close this proxy and forget about it.          //
              oldProxy.close();                                        // 657
            }                                                          //
          }                                                            //
        }                                                              //
      }                                                                //
                                                                       //
      _this2.oldBucket = _this2.newBucket;                             // 663
      _this2.newBucket = {};                                           // 664
    }, intervalMillis);                                                //
  }                                                                    //
                                                                       //
  ApiSessionProxies.prototype.get = (function () {                     // 631
    function get(hashedToken, hashedParams) {                          // 668
      if (this.newBucket[hashedToken] && this.newBucket[hashedToken][hashedParams]) {
        return this.newBucket[hashedToken][hashedParams];              // 670
      } else if (this.oldBucket[hashedToken] && this.oldBucket[hashedToken][hashedParams]) {
        var proxy = this.oldBucket[hashedToken][hashedParams];         // 672
        delete this.oldBucket[hashedToken][hashedParams];              // 673
        this.put(hashedToken, hashedParams, proxy);                    // 674
        return proxy;                                                  // 675
      } else {                                                         //
        return null;                                                   // 677
      }                                                                //
    }                                                                  //
                                                                       //
    return get;                                                        //
  })();                                                                //
                                                                       //
  ApiSessionProxies.prototype.put = (function () {                     // 631
    function put(hashedToken, hashedParams, proxy) {                   // 681
      if (!this.newBucket[hashedToken]) {                              // 682
        this.newBucket[hashedToken] = {};                              // 683
      }                                                                //
                                                                       //
      this.newBucket[hashedToken][hashedParams] = proxy;               // 686
    }                                                                  //
                                                                       //
    return put;                                                        //
  })();                                                                //
                                                                       //
  ApiSessionProxies.prototype.removeProxiesOfToken = (function () {    // 631
    function removeProxiesOfToken(hashedToken) {                       // 689
      if (hashedToken in this.oldBucket) {                             // 690
        for (var hashedParams in babelHelpers.sanitizeForInObject(this.oldBucket[hashedToken])) {
          this.oldBucket[hashedToken][hashedParams].close();           // 692
        }                                                              //
                                                                       //
        delete this.oldBucket[hashedToken];                            // 695
      }                                                                //
                                                                       //
      if (hashedToken in this.newBucket) {                             // 698
        for (var hashedParams in babelHelpers.sanitizeForInObject(this.newBucket[hashedToken])) {
          this.newBucket[hashedToken][hashedParams].close();           // 700
        }                                                              //
                                                                       //
        delete this.newBucket[hashedToken];                            // 703
      }                                                                //
    }                                                                  //
                                                                       //
    return removeProxiesOfToken;                                       //
  })();                                                                //
                                                                       //
  return ApiSessionProxies;                                            //
})();                                                                  //
                                                                       //
var apiSessionProxies = new ApiSessionProxies(3 * 60 * 1000);          // 708
                                                                       //
Meteor.startup(function () {                                           // 710
  var clearSessionsAndProxies = function (token) {                     // 711
    // Clears all sessions and API proxies associated with `token` or any token that is downstream
    // in the sharing graph.                                           //
    // TODO(soon): Only clear sessions and proxies for which the permissions have changed.
    var downstream = SandstormPermissions.downstreamTokens(globalDb, { token: token });
    downstream.push(token);                                            // 716
    var identityIds = [];                                              // 717
    var tokenIds = [];                                                 // 718
                                                                       //
    downstream.forEach(function (token) {                              // 720
      apiSessionProxies.removeProxiesOfToken(token._id);               // 721
                                                                       //
      tokenIds.push(token._id);                                        // 723
      if (token.owner && token.owner.user) {                           // 724
        identityIds.push(token.owner.user.identityId);                 // 725
      }                                                                //
    });                                                                //
                                                                       //
    Sessions.find({                                                    // 729
      grainId: token.grainId,                                          // 730
      $or: [{ identityId: { $in: identityIds } }, { hashedToken: { $in: tokenIds } }]
    }, {                                                               //
      fields: { hostId: 1 }                                            // 735
    }).forEach(function (session) {                                    //
      var proxy = proxiesByHostId[session.hostId];                     // 737
      if (proxy) {                                                     // 738
        proxy.close();                                                 // 739
      }                                                                //
                                                                       //
      delete proxiesByHostId[session.hostId];                          // 742
    });                                                                //
                                                                       //
    Sessions.remove({                                                  // 745
      grainId: token.grainId,                                          // 746
      $or: [{ identityId: { $in: identityIds } }, { hashedToken: { $in: tokenIds } }]
    });                                                                //
  };                                                                   //
                                                                       //
  Grains.find().observe({                                              // 754
    changed: function (newGrain, oldGrain) {                           // 755
      if (oldGrain["private"] != newGrain["private"]) {                // 756
        Sessions.remove({ grainId: oldGrain._id, identityId: { $ne: oldGrain.identityId } });
        ApiTokens.find({ grainId: oldGrain._id }).forEach(function (apiToken) {
          apiSessionProxies.removeProxiesOfToken(apiToken._id);        // 759
        });                                                            //
      }                                                                //
    }                                                                  //
  });                                                                  //
                                                                       //
  ApiTokens.find({ grainId: { $exists: true }, objectId: { $exists: false } }).observe({
    added: function (newApiToken) {                                    // 766
      // TODO(soon): Unfortunately, added() gets called for all existing role assignments when the
      //   front-end restarts, meaning clearing sessions here will cause people's views to refresh
      //   on server upgrade, which is not a nice user experience. It's also sad to force-refresh
      //   people when they gained new permissions since they might be in the middle of something,
      //   and it's not strictly necessary for security. OTOH, it's sad to be non-reactive. Maybe
      //   we should notify people that they have new permissions and let them click a thing to
      //   refresh?                                                    //
      //clearSessions(roleAssignment.grainId, roleAssignment.recipient);
      //clearApiProxies(roleAssignment.grainId);                       //
    },                                                                 //
                                                                       //
    changed: function (newApiToken, oldApiToken) {                     // 778
      if (!_.isEqual(newApiToken.roleAssignment, oldApiToken.roleAssignment) || !_.isEqual(newApiToken.revoked, oldApiToken.revoked)) {
        clearSessionsAndProxies(newApiToken);                          // 781
      }                                                                //
    },                                                                 //
                                                                       //
    removed: function (oldApiToken) {                                  // 785
      clearSessionsAndProxies(oldApiToken);                            // 786
    }                                                                  //
  });                                                                  //
});                                                                    //
                                                                       //
function getApiSessionParams(request) {                                // 791
  var params = {};                                                     // 792
  if ("x-sandstorm-passthrough" in request.headers) {                  // 793
    var optIns = request.headers["x-sandstorm-passthrough"].split(",").map(function (s) {
      return s.trim();                                                 // 796
    });                                                                //
    // The only currently supported passthrough value is 'address', but others could be useful in
    // the future.                                                     //
                                                                       //
    if (optIns.indexOf("address") !== -1) {                            // 800
      // Sadly, we can't use request.socket.remoteFamily because it's not available in the
      // (rather-old) version of node that comes in the Meteor bundle we're using. Hence this
      // hackery.                                                      //
      var addressToPass = request.socket.remoteAddress;                // 804
      if (isRfc1918OrLocal(addressToPass) && "x-real-ip" in request.headers) {
        // Allow overriding the socket's remote address with X-Real-IP header if the request comes
        // from either localhost or an RFC1918 address. These are not useful for geolocation anyway.
        addressToPass = request.headers["x-real-ip"];                  // 808
      }                                                                //
                                                                       //
      if (Net.isIPv4(addressToPass)) {                                 // 811
        // Map IPv4 addresses in IPv6.                                 //
        // This conveniently comes out to a 48-bit number, which is precisely representable in a
        // double (which has 53 mantissa bits). Thus we can avoid using Bignum/strings, which we
        // might otherwise need to precisely represent 64-bit fields.  //
        var v4Int = 0xFFFF00000000 + addressToPass.split(".").map(function (x) {
          return parseInt(x, 10);                                      // 817
        }).reduce(function (a, b) {                                    //
          return 256 * a + b;                                          // 818
        });                                                            //
        params.remoteAddress = {                                       // 819
          lower64: v4Int,                                              // 820
          upper64: 0                                                   // 821
        };                                                             //
      } else if (Net.isIPv6(addressToPass)) {                          //
        // TODO(test): Unit test this                                  //
        // Parse a valid v6 address.                                   //
        // Split into groups, then insert an appropriate number of 0's if :: was used.
        var groups = addressToPass.split(":");                         // 827
                                                                       //
        // Strip extra empty group in the case of a leading or trailing '::'.
        if (groups[0] === "") {                                        // 830
          groups.shift();                                              // 831
        }                                                              //
                                                                       //
        if (groups[groups.length - 1] === "") {                        // 834
          groups.pop();                                                // 835
        }                                                              //
                                                                       //
        var lastGroup = groups[groups.length - 1];                     // 838
        // Handle IPv4-mapped IPv6 addresses.  These end in a dotted-quad IPv4 address, which we
        // should expand into two groups of 4-character hex strings, like the rest of the address.
        if (Net.isIPv4(lastGroup)) {                                   // 841
          groups.pop();                                                // 842
          var _quad = lastGroup.split(".").map(function (x) {          // 843
            return parseInt(x, 10);                                    // 843
          });                                                          //
          groups.push((_quad[0] * 256 + _quad[1]).toString(16));       // 844
          groups.push((_quad[2] * 256 + _quad[3]).toString(16));       // 845
        }                                                              //
                                                                       //
        var groupsToAdd = 8 - groups.length;                           // 848
        var emptyGroupIndex = groups.indexOf("");                      // 849
        var cleanGroups = undefined;                                   // 850
        if (emptyGroupIndex !== -1) {                                  // 851
          var head = groups.slice(0, emptyGroupIndex);                 // 852
          // groupsToAdd + 1 because we sliced out the empty element   //
          var mid = Array(groupsToAdd + 1);                            // 854
          for (var i = 0; i < groupsToAdd + 1; i++) {                  // 855
            mid[i] = "0";                                              // 856
          }                                                            //
                                                                       //
          var tail = groups.slice(emptyGroupIndex + 1, groups.length);
          cleanGroups = [].concat(head, mid, tail);                    // 860
        } else {                                                       //
          cleanGroups = groups;                                        // 862
        }                                                              //
                                                                       //
        var ints = cleanGroups.map(function (x) {                      // 865
          return parseInt(x, 16);                                      // 865
        });                                                            //
        // We use strings because we'd lose data from loss of precision casting the 64-bit uints
        // into 53-bit-mantissa doubles.                               //
        params.remoteAddress = {                                       // 868
          upper64: quadToIntString(ints.slice(0, 4)),                  // 869
          lower64: quadToIntString(ints.slice(4, 8))                   // 870
        };                                                             //
      }                                                                //
    }                                                                  //
  }                                                                    //
                                                                       //
  return Capnp.serialize(ApiSession.Params, params);                   // 876
}                                                                      //
                                                                       //
// Used by server/drivers/external-ui/view.js                          //
getProxyForApiToken = function (token, request) {                      // 880
  check(token, String);                                                // 881
  var hashedToken = Crypto.createHash("sha256").update(token).digest("base64");
  var serializedParams = getApiSessionParams(request);                 // 883
  var hashedParams = Crypto.createHash("sha256").update(serializedParams).digest("base64");
  return Promise.resolve(undefined).then(function () {                 // 885
    var proxy = apiSessionProxies.get(hashedToken, hashedParams);      // 886
    if (proxy) {                                                       // 887
      if (proxy.expires && proxy.expires.getTime() <= Date.now()) {    // 888
        throw new Meteor.Error(403, "Authorization token expired");    // 889
      }                                                                //
                                                                       //
      return proxy;                                                    // 892
    } else {                                                           //
      return inMeteor(function () {                                    // 894
        var tokenInfo = ApiTokens.findOne(hashedToken);                // 895
        validateWebkey(tokenInfo);                                     // 896
                                                                       //
        var grain = Grains.findOne(tokenInfo.grainId);                 // 898
        if (!grain) {                                                  // 899
          // Grain was deleted, I guess.                               //
          throw new Meteor.Error(410, "Resource has been deleted");    // 901
        }                                                              //
                                                                       //
        var proxy = undefined;                                         // 904
        if (tokenInfo.userInfo) {                                      // 905
          throw new Error("API tokens created with arbitrary userInfo no longer supported");
        } else {                                                       //
          var identityId = null;                                       // 908
          if (tokenInfo.identityId && !tokenInfo.forSharing) {         // 909
            identityId = tokenInfo.identityId;                         // 910
          }                                                            //
                                                                       //
          proxy = new Proxy(tokenInfo.grainId, grain.userId, null, null, identityId, true);
          proxy.apiToken = tokenInfo;                                  // 914
          proxy.apiSessionParams = serializedParams;                   // 915
        }                                                              //
                                                                       //
        if (!SandstormPermissions.mayOpenGrain(globalDb, { token: tokenInfo })) {
          // Note that only public grains may be opened without a user ID.
          throw new Meteor.Error(403, "Unauthorized.");                // 920
        }                                                              //
                                                                       //
        if (tokenInfo.expires) {                                       // 923
          proxy.expires = tokenInfo.expires;                           // 924
        }                                                              //
                                                                       //
        apiSessionProxies.put(hashedToken, hashedParams, proxy);       // 927
                                                                       //
        return proxy;                                                  // 929
      });                                                              //
    }                                                                  //
  });                                                                  //
};                                                                     //
                                                                       //
var apiUseBasicAuth = function (req, hostId) {                         // 935
  // If the request was to a token-specific host *and* the request has no Origin header (meaning
  // it is not an XHR from a web site), then we permit the use of HTTP basic auth. The reason we
  // prohibit XHRs is because we want to discourage people from using basic auth in a browser,
  // because the browser will cache the credentials, which exposes the hostname to XSRF attack
  // (which is mostly, but not completely, mitigated by the unguessable hostname). The reason we
  // want to allow basic auth at all is because a lot of existing client apps support only basic
  // auth. New clients and web-based clients should use `Authorization: bearer <token>` instead.
  if (globalDb.isTokenSpecificHostId(hostId) && !req.headers.origin) {
    return true;                                                       // 944
  }                                                                    //
                                                                       //
  // Historically, all API tokens were served from the same host rather than have token-specific
  // hosts. In this model, basic auth in browsers was far more dangerous since the hostname was
  // not a secret -- in fact, even benign attempts to use two APIs on the same server from the same
  // browser could interfere if using basic auth. Hence, basic auth was prohibited except when
  // coming from certain whitelisted clients known not to be browsers, e.g. Mirall, the ownCloud
  // client app.                                                       //
  //                                                                   //
  // Since many clients in the wild have already been configured to use the shared API host, we
  // must continue to support them, so this logic remains.             //
  var agent = req.headers["user-agent"];                               // 956
  return agent.match(BASIC_AUTH_USER_AGENTS_REGEX);                    // 957
};                                                                     //
                                                                       //
var apiTokenForRequest = function (req, hostId) {                      // 960
  // Extract the API token from the request.                           //
                                                                       //
  var auth = req.headers.authorization;                                // 963
  var token = undefined;                                               // 964
  if (auth && auth.slice(0, 7).toLowerCase() === "bearer ") {          // 965
    token = auth.slice(7).trim();                                      // 966
  } else if (auth && auth.slice(0, 6).toLowerCase() === "basic " && apiUseBasicAuth(req, hostId)) {
    token = new Buffer(auth.slice(6).trim(), "base64").toString().split(":")[1];
  } else {                                                             //
    token = undefined;                                                 // 971
  }                                                                    //
                                                                       //
  if (token && hostId !== "api") {                                     // 974
    // Verify that the token matches the specific host.                //
    if (hostId !== globalDb.apiHostIdForToken(token)) {                // 976
      token = undefined;                                               // 977
    }                                                                  //
  }                                                                    //
                                                                       //
  return token;                                                        // 981
};                                                                     //
                                                                       //
// =======================================================================================
// Routing to proxies.                                                 //
//                                                                     //
                                                                       //
// pre-meteor.js calls this                                            //
tryProxyUpgrade = function (hostId, req, socket, head) {               // 989
  // Attempt to handle a WebSocket upgrade by dispatching it to a grain. Returns a promise that
  // resolves true if an appropriate grain is found, false if there was no match (but the caller
  // should consider other host types, like static web publishing), or throws an error if the
  // request is definitely invalid.                                    //
                                                                       //
  if (globalDb.isApiHostId(hostId)) {                                  // 995
    var token = apiTokenForRequest(req, hostId);                       // 996
    if (token) {                                                       // 997
      return getProxyForApiToken(token).then(function (proxy) {        // 998
        // Meteor sets the timeout to five seconds. Change that back to two
        // minutes, which is the default value.                        //
        socket.setTimeout(120000);                                     // 1001
                                                                       //
        proxy.upgradeHandler(req, socket, head);                       // 1003
        return true;                                                   // 1004
      });                                                              //
    } else {                                                           //
      return Promise.resolve(false);                                   // 1007
    }                                                                  //
  } else {                                                             //
    var isAlreadyOpened = req.headers.cookie && req.headers.cookie.indexOf("sandstorm-sid=") !== -1;
    return getProxyForHostId(hostId, isAlreadyOpened).then(function (proxy) {
      if (proxy) {                                                     // 1012
        // Cross-origin requests are not allowed on UI session hosts.  //
        var origin = req.headers.origin;                               // 1014
        if (origin !== PROTOCOL + "//" + req.headers.host) {           // 1015
          throw new Meteor.Error(403, "Detected illegal cross-origin WebSocket from: " + origin);
        }                                                              //
                                                                       //
        // Meteor sets the timeout to five seconds. Change that back to two
        // minutes, which is the default value.                        //
        socket.setTimeout(120000);                                     // 1021
                                                                       //
        proxy.upgradeHandler(req, socket, head);                       // 1023
        return true;                                                   // 1024
      } else {                                                         //
        return false;                                                  // 1026
      }                                                                //
    });                                                                //
  }                                                                    //
};                                                                     //
                                                                       //
// pre-meteor.js calls this                                            //
tryProxyRequest = function (hostId, req, res) {                        // 1033
  // Attempt to handle an HTTP request by dispatching it to a grain. Returns a promise that
  // resolves true if an appropriate grain is found, false if there was no match (but the caller
  // should consider other host types, like static web publishing), or throws an error if the
  // request is definitely invalid.                                    //
                                                                       //
  var hostIdHash = globalDb.isApiHostId(hostId);                       // 1039
  if (hostIdHash) {                                                    // 1040
    var _ret = (function () {                                          //
      // This is a request for the API host.                           //
                                                                       //
      res.setHeader("Access-Control-Allow-Origin", "*");               // 1043
                                                                       //
      if (req.method === "OPTIONS") {                                  // 1045
        // Reply to CORS preflight request.                            //
                                                                       //
        // All we want to do is permit APIs to be accessed from arbitrary origins. Since clients
        // must send a valid Authorization header, and since cookies are not used for
        // authorization, this is perfectly safe. In a sane world, we would only need to send back
        // 'Access-Control-Allow-Origin: *' and be done with it.       //
        //                                                             //
        // However, CORS demands that we explicitly whitelist individual methods and headers for
        // use cross-origin, as if this is somehow useful for implementing any practical security
        // policy (it isn't). To make matters worse, we are REQUIRED to enumerate each one
        // individually. We cannot just write '*' for these lists. WTF, CORS?
        //                                                             //
        // Luckily, the request tells us exactly what method and headers are being requested, so we
        // only need to copy those over, rather than create an exhaustive list. But this is still
        // overly complicated.                                         //
                                                                       //
        var accessControlHeaders = {                                   // 1062
          "Access-Control-Allow-Methods": "GET, HEAD, POST, PUT, PATCH, DELETE",
          "Access-Control-Max-Age": "3600"                             // 1064
        };                                                             //
                                                                       //
        // Copy all requested headers to the allowed headers list.     //
        var requestedHeaders = req.headers["access-control-request-headers"];
        if (requestedHeaders) {                                        // 1069
          accessControlHeaders["Access-Control-Allow-Headers"] = requestedHeaders;
        }                                                              //
                                                                       //
        // Add the requested method to the allowed methods list, if it's not there already.
        var requestedMethod = req.headers["access-control-request-method"];
        if (requestedMethod && !_.contains(["GET", "HEAD", "POST", "PUT", "PATCH", "DELETE"], requestedMethod)) {
          accessControlHeaders["Access-Control-Allow-Methods"] += ", " + requestedMethod;
        }                                                              //
                                                                       //
        for (header in babelHelpers.sanitizeForInObject(accessControlHeaders)) {
          res.setHeader(header, accessControlHeaders[header]);         // 1081
        }                                                              //
      }                                                                //
                                                                       //
      var errorHandler = function (err) {                              // 1085
        if (err instanceof Meteor.Error) {                             // 1086
          console.log("error: " + err);                                // 1087
          res.writeHead(err.error, err.reason, { "Content-Type": "text/plain" });
        } else {                                                       //
          res.writeHead(500, "Internal Server Error", { "Content-Type": "text/plain" });
        }                                                              //
                                                                       //
        res.end(err.stack);                                            // 1093
      };                                                               //
                                                                       //
      var token = apiTokenForRequest(req, hostId);                     // 1096
      if (token && req.headers["x-sandstorm-token-keepalive"]) {       // 1097
        inMeteor(function () {                                         // 1098
          var keepaliveDuration = parseInt(req.headers["x-sandstorm-token-keepalive"]);
          check(keepaliveDuration, Match.Integer);                     // 1100
          var hashedToken = Crypto.createHash("sha256").update(token).digest("base64");
          validateWebkey(ApiTokens.findOne(hashedToken), new Date(Date.now() + keepaliveDuration));
        }).then(function () {                                          //
          res.writeHead(200, {});                                      // 1104
          res.end();                                                   // 1105
        }, errorHandler);                                              //
      } else if (token) {                                              //
        getProxyForApiToken(token, req).then(function (proxy) {        // 1108
          proxy.requestHandler(req, res);                              // 1109
        }, errorHandler);                                              //
      } else {                                                         //
        // No token. Look up static API host info.                     //
        inMeteor(function () {                                         // 1113
          var apiHost = globalDb.collections.apiHosts.findOne(hostIdHash);
                                                                       //
          if (req.method === "OPTIONS") {                              // 1116
            // OPTIONS request with no authorization token.            //
                                                                       //
            var dav = ((apiHost || {}).options || {}).dav || [];       // 1119
            if (dav.length > 0) {                                      // 1120
              res.setHeader("DAV", dav.join(", "));                    // 1121
              res.setHeader("Access-Control-Expose-Headers", "DAV");   // 1122
            }                                                          //
                                                                       //
            res.writeHead(200, {});                                    // 1125
            res.end();                                                 // 1126
          } else {                                                     //
            var resources = (apiHost || {}).resources || {};           // 1128
            var path = SandstormDb.escapeMongoKey(req.url.split("?")[0]);
            if (path in resources) {                                   // 1130
              // Serve a static resource.                              //
              var resource = resources[path];                          // 1132
              console.log(path, resources, resource);                  // 1133
              if (resource.language) res.setHeader("Content-Language", resource.language);
              if (resource.encoding) res.setHeader("Content-Encoding", resource.encoding);
              res.writeHead(200, {                                     // 1136
                "Content-Type": resource.type                          // 1137
              });                                                      //
              res.end(resource.body);                                  // 1139
            } else {                                                   //
              if (apiUseBasicAuth(req, hostId)) {                      // 1141
                res.writeHead(401, {                                   // 1142
                  "Content-Type": "text/plain",                        // 1143
                  "WWW-Authenticate": "Basic realm='Sandstorm API'"    // 1144
                });                                                    //
              } else {                                                 //
                // TODO(someday): Display some sort of nifty API browser.
                res.writeHead(403, { "Content-Type": "text/plain" });  // 1148
              }                                                        //
                                                                       //
              res.end("Missing or invalid authorization header.\n\n" + "This address serves APIs, which allow external apps (such as a phone app) to\n" + "access data on your Sandstorm server. This address is not meant to be opened\n" + "in a regular browser.");
            }                                                          //
          }                                                            //
        }).then(function () {}, errorHandler);                         //
      }                                                                //
                                                                       //
      return {                                                         // 1160
        v: Promise.resolve(true)                                       //
      };                                                               //
    })();                                                              //
                                                                       //
    if (typeof _ret === "object") return _ret.v;                       //
  } else {                                                             //
    var isAlreadyOpened = req.headers.cookie && req.headers.cookie.indexOf("sandstorm-sid=") !== -1;
    return getProxyForHostId(hostId, isAlreadyOpened).then(function (proxy) {
      if (proxy) {                                                     // 1164
        proxy.requestHandler(req, res);                                // 1165
        return true;                                                   // 1166
      } else {                                                         //
        return false;                                                  // 1168
      }                                                                //
    });                                                                //
  }                                                                    //
};                                                                     //
                                                                       //
// =======================================================================================
// Proxy class                                                         //
//                                                                     //
// Connects to a grain and exports it on a wildcard host.              //
//                                                                     //
                                                                       //
var Proxy = (function () {                                             //
  function Proxy(grainId, ownerId, sessionId, hostId, identityId, isApi, supervisor) {
    var _this3 = this;                                                 //
                                                                       //
    babelHelpers.classCallCheck(this, Proxy);                          //
                                                                       //
    this.grainId = grainId;                                            // 1182
    this.ownerId = ownerId;                                            // 1183
    this.identityId = identityId;                                      // 1184
    this.supervisor = supervisor; // note: optional parameter; we can reconnect
    this.sessionId = sessionId;                                        // 1186
    this.isApi = isApi;                                                // 1187
    this.hasLoaded = false;                                            // 1188
    this.websockets = {};                                              // 1189
    this.websocketCounter = 0; // Used for generating unique socket IDs.
    if (sessionId) {                                                   // 1191
      if (!hostId) throw new Error("sessionId must come with hostId");
      if (isApi) throw new Error("API proxy shouldn't have sessionId");
      this.hostId = hostId;                                            // 1194
    } else {                                                           //
      if (!isApi) throw new Error("non-API proxy requires sessionId");
      if (hostId) throw new Error("API proxy sholudn't have hostId");  // 1197
    }                                                                  //
                                                                       //
    if (this.identityId) {                                             // 1200
      var identity = globalDb.getIdentity(this.identityId);            // 1201
      if (!identity) {                                                 // 1202
        throw new Error("identity not found: " + this.identityId);     // 1203
      }                                                                //
                                                                       //
      this.userInfo = {                                                // 1206
        displayName: { defaultText: identity.profile.name },           // 1207
        preferredHandle: identity.profile.handle,                      // 1208
        identityId: new Buffer(identity._id, "hex")                    // 1209
      };                                                               //
      if (identity.profile.pictureUrl) this.userInfo.pictureUrl = identity.profile.pictureUrl;
      if (identity.profile.pronoun) this.userInfo.pronouns = identity.profile.pronoun;
    } else {                                                           //
      this.userInfo = {                                                // 1214
        displayName: { defaultText: "Anonymous User" },                // 1215
        preferredHandle: "anonymous"                                   // 1216
      };                                                               //
    }                                                                  //
                                                                       //
    var _this = this;                                                  // 1220
                                                                       //
    this.requestHandler = function (request, response) {               // 1222
      if (_this3.sessionId) {                                          // 1223
        // Implement /_sandstorm-init for setting the session cookie.  //
        var url = Url.parse(request.url, true);                        // 1225
        if (url.pathname === "/_sandstorm-init" && url.query.sessionid === _this.sessionId) {
          _this.doSessionInit(request, response, url.query.path);      // 1227
          return;                                                      // 1228
        }                                                              //
      }                                                                //
                                                                       //
      Promise.resolve(undefined).then(function () {                    // 1232
        var contentLength = request.headers["content-length"];         // 1233
        if ((request.method === "POST" || request.method === "PUT") && (contentLength === undefined || contentLength > 1024 * 1024)) {
          // The input is either very long, or we don't know how long it is, so use streaming mode.
          return _this.handleRequestStreaming(request, response, contentLength, 0);
        } else {                                                       //
          return readAll(request).then(function (data) {               // 1239
            return _this.handleRequest(request, data, response, 0);    // 1240
          });                                                          //
        }                                                              //
      })["catch"](function (err) {                                     //
        _this.setHasLoaded();                                          // 1244
                                                                       //
        var body = err.stack;                                          // 1246
        if (err.cppFile) {                                             // 1247
          body += "\nC++ location:" + err.cppFile + ":" + (err.line || "??");
        }                                                              //
                                                                       //
        if (err.kjType) {                                              // 1251
          body += "\ntype: " + err.kjType;                             // 1252
        }                                                              //
                                                                       //
        if (response.headersSent) {                                    // 1255
          // Unfortunately, it's too late to tell the client what happened.
          console.error("HTTP request failed after response already sent:", body);
          response.end();                                              // 1258
        } else {                                                       //
          if (err instanceof Meteor.Error) {                           // 1260
            response.writeHead(err.error, err.reason, { "Content-Type": "text/plain" });
          } else {                                                     //
            response.writeHead(500, "Internal Server Error", { "Content-Type": "text/plain" });
          }                                                            //
                                                                       //
          response.end(body);                                          // 1266
        }                                                              //
      });                                                              //
    };                                                                 //
                                                                       //
    this.upgradeHandler = function (request, socket, head) {           // 1271
      _this.handleWebSocket(request, socket, head, 0)["catch"](function (err) {
        console.error("WebSocket setup failed:", err.stack);           // 1273
        // TODO(cleanup):  Manually send back a 500 response?          //
        socket.destroy();                                              // 1275
      });                                                              //
    };                                                                 //
  }                                                                    //
                                                                       //
  Proxy.prototype.close = (function () {                               // 1180
    function close() {                                                 // 1280
      for (var socketIdx in babelHelpers.sanitizeForInObject(this.websockets)) {
        this.websockets[socketIdx].destroy();                          // 1282
      };                                                               //
                                                                       //
      this.websockets = {};                                            // 1285
                                                                       //
      if (this.session) {                                              // 1287
        this.session.close();                                          // 1288
        delete this.session;                                           // 1289
      }                                                                //
                                                                       //
      if (this.uiView) {                                               // 1292
        this.uiView.close();                                           // 1293
        delete this.uiView;                                            // 1294
      }                                                                //
                                                                       //
      if (this.supervisor) {                                           // 1297
        this.supervisor.close();                                       // 1298
        delete this.supervisor;                                        // 1299
      }                                                                //
    }                                                                  //
                                                                       //
    return close;                                                      //
  })();                                                                //
                                                                       //
  Proxy.prototype.getConnection = (function () {                       // 1180
    function getConnection() {                                         // 1303
      if (!this.supervisor) {                                          // 1304
        this.supervisor = globalBackend.cap().getGrain(this.ownerId, this.grainId).supervisor;
        this.uiView = null;                                            // 1306
      }                                                                //
                                                                       //
      if (!this.uiView) {                                              // 1309
        this.uiView = this.supervisor.getMainView().view;              // 1310
      }                                                                //
    }                                                                  //
                                                                       //
    return getConnection;                                              //
  })();                                                                //
                                                                       //
  Proxy.prototype._callNewWebSession = (function () {                  // 1180
    function _callNewWebSession(request, userInfo) {                   // 1314
      var params = Capnp.serialize(WebSession.Params, {                // 1315
        basePath: PROTOCOL + "//" + request.headers.host,              // 1316
        userAgent: "user-agent" in request.headers ? request.headers["user-agent"] : "UnknownAgent/0.0",
        acceptableLanguages: "accept-language" in request.headers ? request.headers["accept-language"].split(",").map(function (s) {
          return s.trim();                                             // 1321
        }) : ["en-US", "en"]                                           //
      });                                                              //
      return this.uiView.newSession(userInfo, makeHackSessionContext(this.grainId, this.sessionId, this.identityId), WebSession.typeId, params).session;
    }                                                                  //
                                                                       //
    return _callNewWebSession;                                         //
  })();                                                                //
                                                                       //
  Proxy.prototype._callNewApiSession = (function () {                  // 1180
    function _callNewApiSession(request, userInfo) {                   // 1329
      var _this4 = this;                                               //
                                                                       //
      var serializedParams = this.apiSessionParams;                    // 1330
      if (!serializedParams) {                                         // 1331
        throw new Meteor.Error(500, "Should have already computed apiSessionParams.");
      }                                                                //
                                                                       //
      // TODO(someday): We are currently falling back to WebSession if we get any kind of error upon
      // calling newSession with an ApiSession._id.                    //
      // Eventually we'll remove this logic once we're sure apps have updated.
      return this.uiView.newSession(userInfo, makeHackSessionContext(this.grainId, this.sessionId, this.identityId), ApiSession.typeId, serializedParams).then(function (session) {
        return session.session;                                        // 1342
      }, function (err) {                                              //
        return _this4._callNewWebSession(request, userInfo);           // 1344
      });                                                              //
    }                                                                  //
                                                                       //
    return _callNewApiSession;                                         //
  })();                                                                //
                                                                       //
  Proxy.prototype._callNewSession = (function () {                     // 1180
    function _callNewSession(request, viewInfo) {                      // 1349
      var userInfo = _.clone(this.userInfo);                           // 1350
      var _this = this;                                                // 1351
      var promise = inMeteor(function () {                             // 1352
        var vertex = undefined;                                        // 1353
        if (_this.apiToken) {                                          // 1354
          vertex = { token: _this.apiToken };                          // 1355
        } else {                                                       //
          // (_this.identityId might be null; this is fine)            //
          vertex = { grain: { _id: _this.grainId, identityId: _this.identityId } };
        }                                                              //
                                                                       //
        var permissions = SandstormPermissions.grainPermissions(globalDb, vertex, viewInfo);
        if (!permissions) {                                            // 1362
          throw new Meteor.Error(403, "Unauthorized", "User is not authorized to open this grain.");
        }                                                              //
                                                                       //
        Sessions.update({                                              // 1366
          _id: _this.sessionId                                         // 1367
        }, {                                                           //
          $set: {                                                      // 1369
            viewInfo: viewInfo,                                        // 1370
            permissions: permissions                                   // 1371
          }                                                            //
        });                                                            //
                                                                       //
        return permissions;                                            // 1375
      });                                                              //
                                                                       //
      return promise.then(function (permissions) {                     // 1378
        userInfo.permissions = permissions;                            // 1379
                                                                       //
        var numBytes = Math.ceil(permissions.length / 8);              // 1381
        var buf = new Buffer(numBytes);                                // 1382
        for (var ii = 0; ii < numBytes; ++ii) {                        // 1383
          buf[ii] = 0;                                                 // 1384
        }                                                              //
                                                                       //
        for (var ii = 0; ii < permissions.length; ++ii) {              // 1387
          var byteNum = Math.floor(ii / 8);                            // 1388
          var bitNum = ii % 8;                                         // 1389
          if (permissions[ii]) {                                       // 1390
            buf[byteNum] = buf[byteNum] | 1 << bitNum;                 // 1391
          }                                                            //
        }                                                              //
                                                                       //
        userInfo.deprecatedPermissionsBlob = buf;                      // 1395
                                                                       //
        if (_this.isApi) {                                             // 1397
          return _this._callNewApiSession(request, userInfo);          // 1398
        } else {                                                       //
          return _this._callNewWebSession(request, userInfo);          // 1400
        }                                                              //
      });                                                              //
    }                                                                  //
                                                                       //
    return _callNewSession;                                            //
  })();                                                                //
                                                                       //
  Proxy.prototype.getSession = (function () {                          // 1180
    function getSession(request) {                                     // 1405
      var _this5 = this;                                               //
                                                                       //
      if (!this.session) {                                             // 1406
        (function () {                                                 //
          _this5.getConnection(); // make sure we're connected         // 1407
          var _this = _this5;                                          // 1408
          var promise = _this5.uiView.getViewInfo().then(function (viewInfo) {
            return inMeteor(function () {                              // 1410
              Grains.update(_this.grainId, { $set: { cachedViewInfo: viewInfo } });
            }).then(function () {                                      //
              return _this._callNewSession(request, viewInfo);         // 1413
            });                                                        //
          }, function (error) {                                        //
            if (error.kjType === "failed" || error.kjType === "unimplemented") {
              // Method not implemented.                               //
              // TODO(apibump): Don't treat 'failed' as 'unimplemented'. Unfortunately, old apps built
              //   with old versions of Cap'n Proto don't throw 'unimplemented' exceptions, so we have
              //   to accept 'failed' here at least until the next API bump.
              return _this._callNewSession(request, {});               // 1421
            } else {                                                   //
              return Promise.reject(error);                            // 1423
            }                                                          //
          });                                                          //
          _this5.session = new Capnp.Capability(promise, WebSession);  // 1426
        })();                                                          //
      }                                                                //
                                                                       //
      return this.session;                                             // 1429
    }                                                                  //
                                                                       //
    return getSession;                                                 //
  })();                                                                //
                                                                       //
  Proxy.prototype.keepAlive = (function () {                           // 1180
    function keepAlive() {                                             // 1432
      this.getConnection();                                            // 1433
      return this.supervisor.keepAlive();                              // 1434
    }                                                                  //
                                                                       //
    return keepAlive;                                                  //
  })();                                                                //
                                                                       //
  Proxy.prototype.resetConnection = (function () {                     // 1180
    function resetConnection() {                                       // 1437
      if (this.session) {                                              // 1438
        this.session.close();                                          // 1439
        delete this.session;                                           // 1440
      }                                                                //
                                                                       //
      if (this.uiView) {                                               // 1443
        this.uiView.close();                                           // 1444
        delete this.uiView;                                            // 1445
      }                                                                //
                                                                       //
      if (this.supervisor) {                                           // 1448
        this.supervisor.close();                                       // 1449
        delete this.supervisor;                                        // 1450
      }                                                                //
    }                                                                  //
                                                                       //
    return resetConnection;                                            //
  })();                                                                //
                                                                       //
  Proxy.prototype.maybeRetryAfterError = (function () {                // 1180
    function maybeRetryAfterError(error, retryCount) {                 // 1454
      // If the error may be caused by the grain dying or a network failure, try to restart it,
      // returning a promise that resolves once restarted. Otherwise, just rethrow the error.
      // `retryCount` should be incremented for every successful retry as part of the same request;
      // we only want to retry once.                                   //
      var _this = this;                                                // 1459
      if (SandstormBackend.shouldRestartGrain(error, retryCount)) {    // 1460
        this.resetConnection();                                        // 1461
        return inMeteor(function () {                                  // 1462
          _this.supervisor = globalBackend.continueGrain(_this.grainId).supervisor;
        });                                                            //
      } else {                                                         //
        throw error;                                                   // 1466
      }                                                                //
    }                                                                  //
                                                                       //
    return maybeRetryAfterError;                                       //
  })();                                                                //
                                                                       //
  Proxy.prototype.doSessionInit = (function () {                       // 1180
    function doSessionInit(request, response, requestPath) {           // 1470
      // jscs:disable disallowQuotedKeysInObjects                      //
      var path = requestPath || "/";                                   // 1472
                                                                       //
      // Check that the path is relative (ie. starts with a /).        //
      // Also ensure that it doesn't start with 2 /, because that is interpreted as non-relative
      if (path.lastIndexOf("/", 0) !== 0 || path.lastIndexOf("//", 0) === 0) {
        response.writeHead(400, "Invalid path supplied", { "Content-Type": "text/plain" });
        response.end("Invalid path supplied.");                        // 1478
        return;                                                        // 1479
      }                                                                //
                                                                       //
      // Set the session ID.                                           //
      response.setHeader("Set-Cookie", ["sandstorm-sid=", this.sessionId, "; Max-Age=31536000; HttpOnly"].join(""));
                                                                       //
      response.setHeader("Cache-Control", "no-cache, private");        // 1485
                                                                       //
      // Redirect to the app's root URL.                               //
      // Note:  All browsers support relative locations and the next update to HTTP/1.1 will officially
      //   make them valid.  http://tools.ietf.org/html/draft-ietf-httpbis-p2-semantics-26#page-67
      response.writeHead(303, "See Other", { "Location": encodeURI(path) });
      response.end();                                                  // 1491
    }                                                                  //
                                                                       //
    return doSessionInit;                                              //
  })();                                                                //
                                                                       //
  Proxy.prototype.makeContext = (function () {                         // 1180
    function makeContext(request, response) {                          // 1494
      // Parses the cookies from the request, checks that the session ID is present and valid, then
      // returns the request context which contains the other cookies.  Throws an exception if the
      // session ID is missing or invalid.                             //
                                                                       //
      var context = {};                                                // 1499
                                                                       //
      if (this.hostId) {                                               // 1501
        var parseResult = parseCookies(request);                       // 1502
        if (!parseResult.sessionId || parseResult.sessionId !== this.sessionId) {
          throw new Meteor.Error(403, "Unauthorized");                 // 1504
        }                                                              //
                                                                       //
        if (parseResult.cookies.length > 0) {                          // 1507
          context.cookies = parseResult.cookies;                       // 1508
        }                                                              //
      } else {                                                         //
        // jscs:disable disallowEmptyBlocks                            //
        // This is an API request. Cookies are not supported.          //
      }                                                                //
                                                                       //
      context.accept = parseAcceptHeader(request);                     // 1515
                                                                       //
      context.eTagPrecondition = parsePreconditionHeader(request);     // 1517
                                                                       //
      context.additionalHeaders = [];                                  // 1519
      WebSession.Context.headerWhitelist.forEach(function (headerName) {
        if (request.headers[headerName]) {                             // 1521
          context.additionalHeaders.push({                             // 1522
            name: headerName,                                          // 1523
            value: request.headers[headerName]                         // 1524
          });                                                          //
        };                                                             //
      });                                                              //
                                                                       //
      var promise = new Promise(function (resolve, reject) {           // 1529
        response.resolveResponseStream = resolve;                      // 1530
        response.rejectResponseStream = reject;                        // 1531
      });                                                              //
                                                                       //
      context.responseStream = new Capnp.Capability(promise, ByteStream);
                                                                       //
      return context;                                                  // 1536
    }                                                                  //
                                                                       //
    return makeContext;                                                //
  })();                                                                //
                                                                       //
  Proxy.prototype.translateResponse = (function () {                   // 1180
    function translateResponse(rpcResponse, response, request) {       // 1539
      if (this.hostId) {                                               // 1540
        if (rpcResponse.setCookies && rpcResponse.setCookies.length > 0) {
          response.setHeader("Set-Cookie", rpcResponse.setCookies.map(makeSetCookieHeader));
        }                                                              //
                                                                       //
        // TODO(security): Add a Content-Security-Policy header which:
        // (1) Prevents the app from initiating HTTP requests to third parties.
        // (2) Prevents the app from navigating the parent frame.      //
        // (3) Prevents the app from opening popups.                   //
        // (4) Prohibits anyone other than the Sandstorm shell from framing the app (as a backup
        //   defense vs. clickjacking, though unguessable hostnames already mostly prevent this).
      } else {                                                         //
          // jscs:disable validateQuoteMarks                           //
          // This is an API request. Cookies are not supported.        //
                                                                       //
          // We need to make sure caches know that different bearer tokens get totally different results.
          response.setHeader('Vary', 'Authorization');                 // 1556
                                                                       //
          // APIs can be called from any origin. Because we ignore cookies, there is no security problem.
          response.setHeader('Access-Control-Allow-Origin', '*');      // 1559
                                                                       //
          // Add a Content-Security-Policy as a backup in case someone finds a way to load this resource
          // in a browser context. This policy should thoroughly neuter it.
          response.setHeader('Content-Security-Policy', "default-src 'none'; sandbox");
        }                                                              //
                                                                       //
      // On first response, update the session to have hasLoaded=true  //
      this.setHasLoaded();                                             // 1567
                                                                       //
      // TODO(security): Set X-Content-Type-Options: nosniff?          //
                                                                       //
      if ('content' in rpcResponse) {                                  // 1571
        var content = rpcResponse.content;                             // 1572
        var code = successCodes[content.statusCode];                   // 1573
        if (!code) {                                                   // 1574
          throw new Error('Unknown status code: ', content.statusCode);
        }                                                              //
                                                                       //
        if (content.mimeType) {                                        // 1578
          response.setHeader('Content-Type', content.mimeType);        // 1579
        }                                                              //
                                                                       //
        if (content.encoding) {                                        // 1582
          response.setHeader('Content-Encoding', content.encoding);    // 1583
        }                                                              //
                                                                       //
        if (content.language) {                                        // 1586
          response.setHeader('Content-Language', content.language);    // 1587
        }                                                              //
                                                                       //
        if (content.eTag) {                                            // 1590
          response.setHeader('ETag', composeETag(content.eTag));       // 1591
        }                                                              //
                                                                       //
        if ('disposition' in content && 'download' in content.disposition) {
          response.setHeader('Content-Disposition', 'attachment; filename="' + content.disposition.download.replace(/([\\"\n])/g, "\\$1") + '"');
        }                                                              //
                                                                       //
        if ('stream' in content.body) {                                // 1599
          if (request.method === 'HEAD') {                             // 1600
            content.body.stream.close();                               // 1601
            response.rejectResponseStream(new Error('HEAD request; content doesn\'t matter.'));
          } else {                                                     //
            var _ret3 = (function () {                                 //
              var streamHandle = content.body.stream;                  // 1604
              response.writeHead(code.id, code.title);                 // 1605
              var promise = new Promise(function (resolve, reject) {   // 1606
                response.resolveResponseStream(new Capnp.Capability(new ResponseStream(response, streamHandle, resolve, reject), ByteStream));
              });                                                      //
              promise.streamHandle = streamHandle;                     // 1610
              return {                                                 // 1611
                v: promise                                             //
              };                                                       //
            })();                                                      //
                                                                       //
            if (typeof _ret3 === "object") return _ret3.v;             //
          }                                                            //
        } else {                                                       //
          response.rejectResponseStream(new Error('Response content body was not a stream.'));
                                                                       //
          if ('bytes' in content.body) {                               // 1617
            response.setHeader('Content-Length', content.body.bytes.length);
          } else {                                                     //
            throw new Error('Unknown content body type.');             // 1620
          }                                                            //
        }                                                              //
                                                                       //
        response.writeHead(code.id, code.title);                       // 1624
                                                                       //
        if ('bytes' in content.body && request.method !== 'HEAD') {    // 1626
          response.write(content.body.bytes);                          // 1627
        }                                                              //
                                                                       //
        response.end();                                                // 1630
      } else if ('noContent' in rpcResponse) {                         //
        var noContent = rpcResponse.noContent;                         // 1632
        var noContentCode = noContentSuccessCodes[noContent.shouldResetForm * 1];
        response.writeHead(noContentCode.id, noContentCode.title);     // 1634
        response.end();                                                // 1635
      } else if ('preconditionFailed' in rpcResponse) {                //
        var preconditionFailed = rpcResponse.preconditionFailed;       // 1637
        if (request.method === 'GET' && 'if-none-match' in request.headers) {
          if (preconditionFailed.matchingETag) {                       // 1639
            response.setHeader('ETag', composeETag(preconditionFailed.matchingETag));
          }                                                            //
                                                                       //
          response.writeHead(304, 'Not Modified');                     // 1643
        } else {                                                       //
          response.writeHead(412, 'Precondition Failed');              // 1645
        }                                                              //
                                                                       //
        response.end();                                                // 1648
      } else if ('redirect' in rpcResponse) {                          //
        var redirect = rpcResponse.redirect;                           // 1650
        var redirectCode = redirectCodes[redirect.switchToGet * 2 + redirect.isPermanent];
        response.writeHead(redirectCode.id, redirectCode.title, {      // 1652
          'Location': redirect.location                                // 1653
        });                                                            //
        response.end();                                                // 1655
      } else if ('clientError' in rpcResponse) {                       //
        var clientError = rpcResponse.clientError;                     // 1657
        var errorCode = errorCodes[clientError.statusCode];            // 1658
        if (!errorCode) {                                              // 1659
          throw new Error('Unknown status code: ', clientError.statusCode);
        }                                                              //
                                                                       //
        response.writeHead(errorCode.id, errorCode.title, {            // 1663
          'Content-Type': 'text/html'                                  // 1664
        });                                                            //
                                                                       //
        if (request.method !== 'HEAD') {                               // 1667
          if (clientError.descriptionHtml) {                           // 1668
            response.write(clientError.descriptionHtml);               // 1669
          } else {                                                     //
            // TODO(someday):  Better default error page.              //
            response.write('<html><body><h1>' + errorCode.id + ': ' + errorCode.title + '</h1></body></html>');
          }                                                            //
        }                                                              //
                                                                       //
        response.end();                                                // 1677
      } else if ('serverError' in rpcResponse) {                       //
        response.writeHead(500, 'Internal Server Error', {             // 1679
          'Content-Type': 'text/html'                                  // 1680
        });                                                            //
                                                                       //
        if (request.method !== 'HEAD') {                               // 1683
          if (rpcResponse.serverError.descriptionHtml) {               // 1684
            response.write(rpcResponse.serverError.descriptionHtml);   // 1685
          } else {                                                     //
            // TODO(someday):  Better default error page.              //
            response.write('<html><body><h1>500: Internal Server Error</h1></body></html>');
          }                                                            //
        }                                                              //
                                                                       //
        response.end();                                                // 1692
      } else {                                                         //
        throw new Error('Unknown HTTP response type:\n' + JSON.stringify(rpcResponse));
      }                                                                //
                                                                       //
      return Promise.resolve(undefined);                               // 1697
    }                                                                  //
                                                                       //
    return translateResponse;                                          //
  })();                                                                //
                                                                       //
  Proxy.prototype.handleRequest = (function () {                       // 1180
    function handleRequest(request, data, response, retryCount) {      // 1700
      var _this = this;                                                // 1701
      return Promise.resolve(undefined).then(function () {             // 1702
        return _this.makeContext(request, response);                   // 1703
      }).then(function (context) {                                     //
        // jscs:disable requireDotNotation                             //
        // Send the RPC.                                               //
        var path = request.url.slice(1); // remove leading '/'         // 1707
        var session = _this.getSession(request);                       // 1708
                                                                       //
        var requestContent = function () {                             // 1710
          return {                                                     // 1711
            content: data,                                             // 1712
            encoding: request.headers['content-encoding'],             // 1713
            mimeType: request.headers['content-type']                  // 1714
          };                                                           //
        };                                                             //
                                                                       //
        var xmlContent = function () {                                 // 1718
          var type = request.headers['content-type'] || 'application/xml;charset=utf-8';
          var match = type.match(/[^/]*\/xml(; *charset *= *([^ ;]*))?/);
          if (!match) {                                                // 1721
            response.writeHead(415, 'Unsupported media type.', {       // 1722
              'Content-Type': 'text/plain'                             // 1723
            });                                                        //
            response.end('expected XML request body');                 // 1725
            throw new Error('expected XML request body');              // 1726
          }                                                            //
                                                                       //
          var charset = match[2] || 'ISO-8859-1';                      // 1729
                                                                       //
          var encoding = request.headers['content-encoding'];          // 1731
          if (encoding && encoding !== 'identity') {                   // 1732
            if (encoding !== 'gzip') throw new Error('unknown Content-Encoding: ' + encoding);
            data = gunzipSync(data);                                   // 1734
          }                                                            //
                                                                       //
          return data.toString(charset.toLowerCase() === 'utf-8' ? 'utf8' : 'binary');
        };                                                             //
                                                                       //
        var propfindDepth = function () {                              // 1740
          var depth = request.headers['depth'];                        // 1741
          return depth === '0' ? 'zero' : depth === '1' ? 'one' : 'infinity';
        };                                                             //
                                                                       //
        var shallow = function () {                                    // 1747
          return request.headers['depth'] === '0';                     // 1748
        };                                                             //
                                                                       //
        var noOverwrite = function () {                                // 1751
          return (request.headers['overwrite'] || '').toLowerCase() === 'f';
        };                                                             //
                                                                       //
        var destination = function () {                                // 1755
          var result = request.headers['destination'];                 // 1756
          if (!result) throw new Error('missing destination');         // 1757
          return Url.parse(result).path.slice(1); // remove leading '/'
        };                                                             //
                                                                       //
        if (request.method === 'GET' || request.method === 'HEAD') {   // 1761
          return session.get(path, context, request.method === 'HEAD');
        } else if (request.method === 'POST') {                        //
          return session.post(path, requestContent(), context);        // 1764
        } else if (request.method === 'PUT') {                         //
          return session.put(path, requestContent(), context);         // 1766
        } else if (request.method === 'PATCH') {                       //
          return session.patch(path, requestContent(), context);       // 1768
        } else if (request.method === 'DELETE') {                      //
          return session["delete"](path, context);                     // 1770
        } else if (request.method === 'PROPFIND') {                    //
          return session.propfind(path, xmlContent(), propfindDepth(), context);
        } else if (request.method === 'PROPPATCH') {                   //
          return session.proppatch(path, xmlContent(), context);       // 1774
        } else if (request.method === 'MKCOL') {                       //
          return session.mkcol(path, requestContent(), context);       // 1776
        } else if (request.method === 'COPY') {                        //
          return session.copy(path, destination(), noOverwrite(), shallow(), context);
        } else if (request.method === 'MOVE') {                        //
          return session.move(path, destination(), noOverwrite(), context);
        } else if (request.method === 'LOCK') {                        //
          return session.lock(path, xmlContent(), shallow(), context);
        } else if (request.method === 'UNLOCK') {                      //
          return session.unlock(path, request.headers['lock-token'], context);
        } else if (request.method === 'ACL') {                         //
          return session.acl(path, xmlContent(), context);             // 1786
        } else if (request.method === 'REPORT') {                      //
          return session.report(path, requestContent(), context);      // 1788
        } else if (request.method === 'OPTIONS') {                     //
          return session.options(path, context).then(function (options) {
            var dav = [];                                              // 1791
            if (options.davClass1) dav.push('1');                      // 1792
            if (options.davClass2) dav.push('2');                      // 1793
            if (options.davClass3) dav.push('3');                      // 1794
            if (options.davExtensions) {                               // 1795
              options.davExtensions.forEach(function (token) {         // 1796
                if (token.match(/^([a-zA-Z0-9!#$%&'*+.^_`|~-]+|<[\x21-\x7E]*>)$/)) {
                  dav.push(token);                                     // 1798
                }                                                      //
              });                                                      //
            }                                                          //
                                                                       //
            if (dav.length > 0) {                                      // 1803
              response.setHeader("DAV", dav.join(", "));               // 1804
              response.setHeader("Access-Control-Expose-Headers", "DAV");
            }                                                          //
                                                                       //
            response.end();                                            // 1808
            // Return no response; we already handled everything.      //
          }, function (err) {                                          //
            if (err.kjType !== 'unimplemented') throw err;             // 1811
            response.end();                                            // 1812
            // Return no response; we already handled everything.      //
          });                                                          //
        } else {                                                       //
            throw new Error('Sandstorm only supports the following methods: GET, POST, PUT, PATCH, DELETE, HEAD, PROPFIND, PROPPATCH, MKCOL, COPY, MOVE, LOCK, UNLOCK, ACL, REPORT, and OPTIONS.');
          }                                                            //
      }).then(function (rpcResponse) {                                 //
        if (rpcResponse !== undefined) {                               // 1819
          // Will be undefined for OPTIONS request.                    //
          return _this.translateResponse(rpcResponse, response, request);
        }                                                              //
      })["catch"](function (error) {                                   //
        return _this.maybeRetryAfterError(error, retryCount).then(function () {
          return _this.handleRequest(request, data, response, retryCount + 1);
        });                                                            //
      });                                                              //
    }                                                                  //
                                                                       //
    return handleRequest;                                              //
  })();                                                                //
                                                                       //
  Proxy.prototype.handleRequestStreaming = (function () {              // 1180
    function handleRequestStreaming(request, response, contentLength, retryCount) {
      var _this = this;                                                // 1830
      var context = this.makeContext(request, response);               // 1831
      var path = request.url.slice(1); // remove leading '/'           // 1832
      var session = this.getSession(request);                          // 1833
                                                                       //
      var mimeType = request.headers['content-type'] || 'application/octet-stream';
      var encoding = request.headers['content-encoding'];              // 1836
                                                                       //
      var requestStreamPromise = undefined;                            // 1838
      if (request.method === 'POST') {                                 // 1839
        requestStreamPromise = session.postStreaming(path, mimeType, context, encoding);
      } else if (request.method === 'PUT') {                           //
        requestStreamPromise = session.putStreaming(path, mimeType, context, encoding);
      } else {                                                         //
        throw new Error('Sandstorm only supports streaming POST and PUT requests.');
      }                                                                //
                                                                       //
      // TODO(perf): We ought to be pipelining the body, but we can't currently, because we have to
      //   handle the case where the app doesn't actually support streaming. We could pipeline while
      //   also buffering the data on the side in case we need it again later, but that's kind of
      //   complicated. We should fix the whole protocol to make streaming the standard.
      return requestStreamPromise.then(function (requestStreamResult) {
        var requestStream = requestStreamResult.stream;                // 1852
                                                                       //
        // Initialized when getResponse() returns, if the response is streaming.
        var downloadStreamHandle = undefined;                          // 1855
                                                                       //
        // Initialized if an upload-stream method throws.              //
        var uploadStreamError = undefined;                             // 1858
                                                                       //
        // We call `getResponse()` immediately so that the app can start streaming data down even while
        // data is still being streamed up. This theoretically allows apps to perform bidirectional
        // streaming, though probably very few actually do that.       //
        //                                                             //
        // Note that we need to be able to cancel `responsePromise` below, so it's important that it is
        // the raw Cap'n Proto promise. Hence `translateResponsePromise` is a separate variable.
        var responsePromise = requestStream.getResponse();             // 1866
                                                                       //
        var reportUploadStreamError = function (err) {                 // 1868
          // Called when an upload-stream method throws.               //
                                                                       //
          if (!uploadStreamError) {                                    // 1871
            uploadStreamError = err;                                   // 1872
                                                                       //
            // If we're still waiting on any response stuff, cancel it.
            responsePromise.cancel();                                  // 1875
            requestStream.close();                                     // 1876
            if (downloadStreamHandle) {                                // 1877
              downloadStreamHandle.close();                            // 1878
            }                                                          //
          }                                                            //
        };                                                             //
                                                                       //
        // If we have a Content-Length, pass it along to the app by calling `expectSize()`.
        if (contentLength !== undefined) {                             // 1884
          requestStream.expectSize(contentLength)["catch"](function (err) {
            // expectSize() is allowed to be unimplemented.            //
            if (err.kjType !== 'unimplemented') {                      // 1887
              reportUploadStreamError(err);                            // 1888
            }                                                          //
          });                                                          //
        }                                                              //
                                                                       //
        // Pipe the input stream to the app.                           //
        request.on('data', function (buf) {                            // 1894
          // TODO(soon): Only allow a small number of write()s to be in-flight at once,
          //   pausing the input stream if we hit that limit, so that we block the TCP socket all the
          //   way back to the source. May want to also coalesce small writes for this purpose.
          // TODO(security): The above problem may allow a DoS attack on the front-end.
          if (!uploadStreamError) requestStream.write(buf)["catch"](reportUploadStreamError);
        });                                                            //
                                                                       //
        request.on('end', function () {                                // 1902
          if (!uploadStreamError) requestStream.done()["catch"](reportUploadStreamError);
                                                                       //
          // We're all done making calls to requestStream.             //
          requestStream.close();                                       // 1906
        });                                                            //
                                                                       //
        request.on('close', function () {                              // 1909
          reportUploadStreamError(new Error('HTTP connection unexpectedly closed during request.'));
        });                                                            //
                                                                       //
        request.on('error', function (err) {                           // 1913
          reportUploadStreamError(err);                                // 1914
        });                                                            //
                                                                       //
        return responsePromise.then(function (rpcResponse) {           // 1917
          // Stop here if the upload stream has already failed.        //
          if (uploadStreamError) throw uploadStreamError;              // 1919
          var promise = _this.translateResponse(rpcResponse, response, request);
          downloadStreamHandle = promise.streamHandle;                 // 1921
          return promise;                                              // 1922
        });                                                            //
      }, function (err) {                                              //
        if (err.kjType === 'failed' && err.message.indexOf('not implemented') !== -1) {
          // Hack to work around old apps using an old version of Cap'n Proto, before the
          // 'unimplemented' exception type was introduced. :(         //
          // TODO(cleanup): When we transition to API version 2, we can move this into the
          //   compatibility layer.                                    //
          err.kjType = 'unimplemented';                                // 1930
        }                                                              //
                                                                       //
        if (SandstormBackend.shouldRestartGrain(err, 0)) {             // 1933
          // This is the kind of error that indicates we should retry. Note that we passed 0 for the
          // retry count above because we were just checking if this is a retriable error (vs. possibly
          // a method-not-implemented error); maybeRetryAfterError() will check again with the proper
          // retry count.                                              //
          return _this.maybeRetryAfterError(err, retryCount).then(function () {
            return _this.handleRequestStreaming(request, response, contentLength, retryCount + 1);
          });                                                          //
        } else if (err.kjType === 'unimplemented') {                   //
          // Streaming is not implemented. Fall back to non-streaming version.
          return readAll(request).then(function (data) {               // 1943
            return _this.handleRequest(request, data, response, 0);    // 1944
          });                                                          //
        } else {                                                       //
          throw err;                                                   // 1947
        }                                                              //
      });                                                              //
    }                                                                  //
                                                                       //
    return handleRequestStreaming;                                     //
  })();                                                                //
                                                                       //
  Proxy.prototype.handleWebSocket = (function () {                     // 1180
    function handleWebSocket(request, socket, head, retryCount) {      // 1952
      var _this = this;                                                // 1953
                                                                       //
      return Promise.resolve(undefined).then(function () {             // 1955
        return _this.makeContext(request);                             // 1956
      }).then(function (context) {                                     //
        var path = request.url.slice(1); // remove leading '/'         // 1958
        var session = _this.getSession(request);                       // 1959
                                                                       //
        if (!('sec-websocket-key' in request.headers)) {               // 1961
          throw new Error('Missing Sec-WebSocket-Accept header.');     // 1962
        }                                                              //
                                                                       //
        var magic = request.headers['sec-websocket-key'] + '258EAFA5-E914-47DA-95CA-C5AB0DC85B11';
        var acceptKey = Crypto.createHash('sha1').update(magic).digest('base64');
                                                                       //
        var protocols = [];                                            // 1968
        if ('sec-websocket-protocol' in request.headers) {             // 1969
          protocols = request.headers['sec-websocket-protocol'].split(',').map(function (s) {
            return s.trim();                                           // 1971
          });                                                          //
        }                                                              //
                                                                       //
        var receiver = new WebSocketReceiver(socket);                  // 1974
                                                                       //
        var promise = session.openWebSocket(path, context, protocols, receiver);
                                                                       //
        if (head.length > 0) {                                         // 1978
          promise.serverStream.sendBytes(head);                        // 1979
        }                                                              //
                                                                       //
        var socketIdx = _this.websocketCounter.toString();             // 1982
        _this.websockets[socketIdx] = socket;                          // 1983
        _this.websocketCounter += 1;                                   // 1984
        pumpWebSocket(socket, promise.serverStream, function () {      // 1985
          delete _this.websockets[socketIdx];                          // 1985
        });                                                            //
                                                                       //
        return promise.then(function (response) {                      // 1987
          var headers = ['HTTP/1.1 101 Switching Protocols', 'Upgrade: websocket', 'Connection: Upgrade', 'Sec-WebSocket-Accept: ' + acceptKey];
          if (response.protocol && response.protocol.length > 0) {     // 1994
            headers.push('Sec-WebSocket-Protocol: ' + response.protocol.join(', '));
          }                                                            //
                                                                       //
          headers.push('');                                            // 1998
          headers.push('');                                            // 1999
                                                                       //
          socket.write(headers.join('\r\n'));                          // 2001
          receiver.go();                                               // 2002
                                                                       //
          // Note:  At this point errors are out of our hands.         //
        });                                                            //
      })["catch"](function (error) {                                   //
        return _this.maybeRetryAfterError(error, retryCount).then(function () {
          return _this.handleWebSocket(request, socket, head, retryCount + 1);
        });                                                            //
      });                                                              //
    }                                                                  //
                                                                       //
    return handleWebSocket;                                            //
  })();                                                                //
                                                                       //
  Proxy.prototype.setHasLoaded = (function () {                        // 1180
    function setHasLoaded() {                                          // 2013
      var _this6 = this;                                               //
                                                                       //
      if (!this.hasLoaded) {                                           // 2014
        (function () {                                                 //
          _this6.hasLoaded = true;                                     // 2015
          var sessionId = _this6.sessionId;                            // 2016
          inMeteor(function () {                                       // 2017
            Sessions.update({ _id: sessionId }, { $set: { hasLoaded: true } });
          });                                                          //
        })();                                                          //
      }                                                                //
    }                                                                  //
                                                                       //
    return setHasLoaded;                                               //
  })();                                                                //
                                                                       //
  return Proxy;                                                        //
})();                                                                  //
                                                                       //
;                                                                      // 2022
                                                                       //
var PROTOCOL = Url.parse(process.env.ROOT_URL).protocol;               // 2024
                                                                       //
var isRfc1918OrLocal = function (address) {                            // 2026
  if (Net.isIPv4(address)) {                                           // 2027
    quad = address.split('.').map(function (x) {                       // 2028
      return parseInt(x, 10);                                          // 2028
    });                                                                //
    return quad[0] === 127 || quad[0] === 10 || quad[0] === 192 && quad[1] === 168 || quad[0] === 172 && quad[1] >= 16 && quad[1] < 32;
  } else if (Net.isIPv6(address)) {                                    //
    // IPv6 specifies ::1 as localhost and fd:: as reserved for private networks
    return address === '::1' || address.lastIndexOf('fd', 0) === 0;    // 2034
  } else {                                                             //
    // Ignore things that are neither IPv4 nor IPv6                    //
    return false;                                                      // 2037
  }                                                                    //
};                                                                     //
                                                                       //
var quadToIntString = function (quad) {                                // 2041
  var num = Bignum(quad[0]).shiftLeft(16).add(quad[1]).shiftLeft(16).add(quad[2]).shiftLeft(16).add(quad[3]);
  return num.toString();                                               // 2046
};                                                                     //
                                                                       //
// -----------------------------------------------------------------------------
// Session cookie management                                           //
                                                                       //
var parseCookies = function (request) {                                // 2052
  // jscs:disable requireDotNotation                                   //
  var header = request.headers['cookie'];                              // 2054
                                                                       //
  var result = { cookies: [] };                                        // 2056
  if (header) {                                                        // 2057
    var reqCookies = header.split(';');                                // 2058
    for (var i in babelHelpers.sanitizeForInObject(reqCookies)) {      // 2059
      var reqCookie = reqCookies[i];                                   // 2060
      var equalsPos = reqCookie.indexOf('=');                          // 2061
      var cookie = undefined;                                          // 2062
      if (equalsPos === -1) {                                          // 2063
        cookie = { key: reqCookie.trim(), value: '' };                 // 2064
      } else {                                                         //
        cookie = { key: reqCookie.slice(0, equalsPos).trim(), value: reqCookie.slice(equalsPos + 1) };
      }                                                                //
                                                                       //
      if (cookie.key === 'sandstorm-sid') {                            // 2069
        if (result.sessionId) {                                        // 2070
          throw new Error('Multiple sandstorm session IDs?');          // 2071
        }                                                              //
                                                                       //
        result.sessionId = cookie.value;                               // 2074
      } else {                                                         //
        result.cookies.push(cookie);                                   // 2076
      }                                                                //
    }                                                                  //
  }                                                                    //
                                                                       //
  return result;                                                       // 2081
};                                                                     //
                                                                       //
var parsePreconditionHeader = function (request) {                     // 2084
  if (request.headers['if-match']) {                                   // 2085
    if (request.headers['if-match'].trim() === '*') {                  // 2086
      return { exists: null };                                         // 2087
    }                                                                  //
                                                                       //
    var matches = parseETagList(request.headers['if-match']);          // 2090
    if (matches.length > 0) {                                          // 2091
      return { matchesOneOf: matches };                                // 2092
    }                                                                  //
  }                                                                    //
                                                                       //
  if (request.headers['if-none-match']) {                              // 2096
    if (request.headers['if-none-match'].trim() === '*') {             // 2097
      return { doesntExist: null };                                    // 2098
    }                                                                  //
                                                                       //
    var noneMatches = parseETagList(request.headers['if-none-match']);
    if (noneMatches.length > 0) {                                      // 2102
      return { matchesNoneOf: noneMatches };                           // 2103
    }                                                                  //
  }                                                                    //
                                                                       //
  return { none: null };                                               // 2107
};                                                                     //
                                                                       //
var parseETagList = function (input) {                                 // 2110
  // An ETag is a quoted, \-escaped string, possibly prefixed with W/ (outside the quotes) to
  // indicate that it is weak. We are parsing a list of comma-delimited etags.
                                                                       //
  input = input.trim();                                                // 2114
  var results = [];                                                    // 2115
                                                                       //
  while (input.length > 0) {                                           // 2117
    var match = input.match(/^\s*(W\/)?"(([^"\\]|\\.)*)"\s*($|,)/);    // 2118
    if (!match) throw new Meteor.Error(400, 'invalid etag');           // 2119
                                                                       //
    input = input.slice(match[0].length).trim();                       // 2121
    results.push({ weak: !!match[1], value: match[2].replace(/\\(.)/g, '$1') });
  }                                                                    //
                                                                       //
  return results;                                                      // 2125
};                                                                     //
                                                                       //
var composeETag = function (tag) {                                     // 2128
  var result = '"' + (tag.value || '').replace(/([\\"])/g, '\\$1') + '"';
  if (tag.weak) result = 'W/' + result;                                // 2130
  return result;                                                       // 2131
};                                                                     //
                                                                       //
var parseAcceptHeader = function (request) {                           // 2134
  // jscs:disable requireDotNotation                                   //
  var header = request.headers['accept'];                              // 2136
                                                                       //
  var result = [];                                                     // 2138
  if (header) {                                                        // 2139
    var acceptList = header.split(',');                                // 2140
    for (var i in babelHelpers.sanitizeForInObject(acceptList)) {      // 2141
      var acceptStr = acceptList[i];                                   // 2142
      var tokensList = acceptStr.split(';');                           // 2143
                                                                       //
      var temp = { mimeType: tokensList[0].trim() };                   // 2145
                                                                       //
      var tokensListRest = tokensList.slice(1);                        // 2147
      for (var j in babelHelpers.sanitizeForInObject(tokensListRest)) {
        var token = tokensListRest[j];                                 // 2149
        var equalsPos = token.indexOf('=');                            // 2150
        if (equalsPos) {                                               // 2151
          var key = token.slice(0, equalsPos).trim();                  // 2152
          var value = token.slice(equalsPos + 1).trim();               // 2153
                                                                       //
          if (key === 'q') {                                           // 2155
            temp.qValue = +value;                                      // 2156
          }                                                            //
        }                                                              //
      }                                                                //
                                                                       //
      result.push(temp);                                               // 2161
    }                                                                  //
  }                                                                    //
                                                                       //
  return result;                                                       // 2165
};                                                                     //
                                                                       //
// -----------------------------------------------------------------------------
// Regular HTTP request handling                                       //
                                                                       //
var readAll = function (stream) {                                      // 2171
  return new Promise(function (resolve, reject) {                      // 2172
    var buffers = [];                                                  // 2173
    stream.on('data', function (buf) {                                 // 2174
      buffers.push(buf);                                               // 2175
    });                                                                //
                                                                       //
    stream.on('end', function () {                                     // 2178
      resolve(Buffer.concat(buffers));                                 // 2179
    });                                                                //
                                                                       //
    stream.on('error', reject);                                        // 2182
  });                                                                  //
};                                                                     //
                                                                       //
var makeSetCookieHeader = function (cookie) {                          // 2186
  var result = [cookie.name, '=', cookie.value];                       // 2187
                                                                       //
  if ('absolute' in cookie.expires) {                                  // 2189
    result.push('; Expires=');                                         // 2190
    result.push(new Date(cookie.expires.absolute * 1000).toUTCString());
  } else if ('relative' in cookie.expires) {                           //
    result.push('; Max-Age=' + cookie.expires.relative);               // 2193
  }                                                                    //
                                                                       //
  if (cookie.path) {                                                   // 2196
    result.push('; Path=' + cookie.path);                              // 2197
  }                                                                    //
                                                                       //
  if (cookie.httpOnly) {                                               // 2200
    result.push('; HttpOnly');                                         // 2201
  }                                                                    //
                                                                       //
  return result.join('');                                              // 2204
};                                                                     //
                                                                       //
// TODO(cleanup):  Auto-generate based on annotations in web-session.capnp.
var successCodes = {                                                   // 2208
  ok: { id: 200, title: 'OK' },                                        // 2209
  created: { id: 201, title: 'Created' },                              // 2210
  accepted: { id: 202, title: 'Accepted' },                            // 2211
  multiStatus: { id: 207, title: 'Multi-Status' }                      // 2212
};                                                                     //
var noContentSuccessCodes = [                                          // 2214
// Indexed by shouldResetForm * 1                                      //
{ id: 204, title: 'No Content' }, { id: 205, title: 'Reset Content' }];
var redirectCodes = [                                                  // 2219
// Indexed by switchToGet * 2 + isPermanent                            //
{ id: 307, title: 'Temporary Redirect' }, { id: 308, title: 'Permanent Redirect' }, { id: 303, title: 'See Other' }, { id: 301, title: 'Moved Permanently' }];
var errorCodes = {                                                     // 2226
  badRequest: { id: 400, title: 'Bad Request' },                       // 2227
  forbidden: { id: 403, title: 'Forbidden' },                          // 2228
  notFound: { id: 404, title: 'Not Found' },                           // 2229
  methodNotAllowed: { id: 405, title: 'Method Not Allowed' },          // 2230
  notAcceptable: { id: 406, title: 'Not Acceptable' },                 // 2231
  conflict: { id: 409, title: 'Conflict' },                            // 2232
  gone: { id: 410, title: 'Gone' },                                    // 2233
  requestEntityTooLarge: { id: 413, title: 'Request Entity Too Large' },
  requestUriTooLong: { id: 414, title: 'Request-URI Too Long' },       // 2235
  unsupportedMediaType: { id: 415, title: 'Unsupported Media Type' },  // 2236
  imATeapot: { id: 418, title: 'I\'m a teapot' },                      // 2237
  unprocessableEntity: { id: 422, title: 'Unprocessable Entity' }      // 2238
};                                                                     //
                                                                       //
ResponseStream = (function () {                                        // 2241
  function ResponseStream(response, streamHandle, resolve, reject) {   // 2242
    babelHelpers.classCallCheck(this, ResponseStream);                 //
                                                                       //
    this.response = response;                                          // 2243
    this.streamHandle = streamHandle;                                  // 2244
    this.resolve = resolve;                                            // 2245
    this.reject = reject;                                              // 2246
    this.ended = false;                                                // 2247
  }                                                                    //
                                                                       //
  ResponseStream.prototype.write = (function () {                      // 2241
    function write(data) {                                             // 2250
      this.response.write(data);                                       // 2251
    }                                                                  //
                                                                       //
    return write;                                                      //
  })();                                                                //
                                                                       //
  ResponseStream.prototype.done = (function () {                       // 2241
    function done() {                                                  // 2254
      this.response.end();                                             // 2255
      this.streamHandle.close();                                       // 2256
      this.ended = true;                                               // 2257
    }                                                                  //
                                                                       //
    return done;                                                       //
  })();                                                                //
                                                                       //
  ResponseStream.prototype.close = (function () {                      // 2241
    function close() {                                                 // 2260
      if (this.ended) {                                                // 2261
        this.resolve();                                                // 2262
      } else {                                                         //
        this.streamHandle.close();                                     // 2264
        this.reject(new Error('done() was never called on outbound stream.'));
      }                                                                //
    }                                                                  //
                                                                       //
    return close;                                                      //
  })();                                                                //
                                                                       //
  return ResponseStream;                                               //
})();                                                                  //
                                                                       //
// TODO(cleanup): Node 0.12 has a `gunzipSync` but 0.10 (which Meteor still uses) does not.
var Zlib = Npm.require('zlib');                                        // 2271
var gunzipSync = Meteor.wrapAsync(Zlib.gunzip, Zlib);                  // 2272
                                                                       //
// -----------------------------------------------------------------------------
// WebSocket handling                                                  //
                                                                       //
WebSocketReceiver = (function () {                                     // 2277
  function WebSocketReceiver(socket) {                                 // 2278
    babelHelpers.classCallCheck(this, WebSocketReceiver);              //
                                                                       //
    this.socket = socket;                                              // 2279
    this.queue = [];                                                   // 2280
  }                                                                    //
                                                                       //
  WebSocketReceiver.prototype.go = (function () {                      // 2277
    function go() {                                                    // 2283
      for (var i in babelHelpers.sanitizeForInObject(this.queue)) {    // 2284
        this.socket.write(this.queue[i]);                              // 2285
      }                                                                //
                                                                       //
      this.queue = null;                                               // 2288
    }                                                                  //
                                                                       //
    return go;                                                         //
  })();                                                                //
                                                                       //
  WebSocketReceiver.prototype.sendBytes = (function () {               // 2277
    function sendBytes(message) {                                      // 2291
      // TODO(someday):  Flow control of some sort?                    //
      if (this.queue === null) {                                       // 2293
        this.socket.write(message);                                    // 2294
      } else {                                                         //
        this.queue.push(message);                                      // 2296
      }                                                                //
    }                                                                  //
                                                                       //
    return sendBytes;                                                  //
  })();                                                                //
                                                                       //
  WebSocketReceiver.prototype.close = (function () {                   // 2277
    function close() {                                                 // 2300
      this.socket.end();                                               // 2301
    }                                                                  //
                                                                       //
    return close;                                                      //
  })();                                                                //
                                                                       //
  return WebSocketReceiver;                                            //
})();                                                                  //
                                                                       //
pumpWebSocket = function (socket, rpcStream, destructor) {             // 2305
  socket.on('data', function (chunk) {                                 // 2306
    rpcStream.sendBytes(chunk)["catch"](function (err) {               // 2307
      if (err.kjType !== 'disconnected') {                             // 2308
        console.error('WebSocket sendBytes failed: ' + err.stack);     // 2309
      }                                                                //
                                                                       //
      socket.destroy();                                                // 2312
    });                                                                //
  });                                                                  //
                                                                       //
  socket.on('end', function (chunk) {                                  // 2316
    rpcStream.close();                                                 // 2317
  });                                                                  //
                                                                       //
  socket.on("close", function () {                                     // 2320
    destructor();                                                      // 2321
  });                                                                  //
};                                                                     //
                                                                       //
// =======================================================================================
// Debug log access                                                    //
                                                                       //
Meteor.publish('grainLog', function (grainId) {                        // 2328
  var _this7 = this;                                                   //
                                                                       //
  check(grainId, String);                                              // 2329
  var id = 0;                                                          // 2330
  var grain = Grains.findOne(grainId);                                 // 2331
  if (!grain || !this.userId || grain.userId !== this.userId) {        // 2332
    this.added('grainLog', id++, { text: 'Only the grain owner can view the debug log.' });
    this.ready();                                                      // 2334
    return;                                                            // 2335
  }                                                                    //
                                                                       //
  var connected = false;                                               // 2338
  var _this = this;                                                    // 2339
                                                                       //
  var receiver = {                                                     // 2341
    write: function (data) {                                           // 2342
      connected = true;                                                // 2343
      _this.added('grainLog', id++, { text: data.toString('utf8') });  // 2344
    },                                                                 //
                                                                       //
    close: function () {                                               // 2347
      if (connected) {                                                 // 2348
        _this.added('grainLog', id++, {                                // 2349
          text: '*** lost connection to grain (probably because it shut down) ***'
        });                                                            //
      }                                                                //
    }                                                                  //
  };                                                                   //
                                                                       //
  try {                                                                // 2356
    (function () {                                                     //
      var handle = waitPromise(globalBackend.useGrain(grainId, function (supervisor) {
        return supervisor.watchLog(8192, receiver);                    // 2358
      })).handle;                                                      //
      connected = true;                                                // 2360
      _this7.onStop(function () {                                      // 2361
        handle.close();                                                // 2362
      });                                                              //
    })();                                                              //
  } catch (err) {                                                      //
    if (!connected) {                                                  // 2365
      this.added('grainLog', id++, {                                   // 2366
        text: '*** couldn\'t connect to grain (' + err + ') ***'       // 2367
      });                                                              //
    }                                                                  //
  }                                                                    //
                                                                       //
  // Notify ready.                                                     //
  this.ready();                                                        // 2373
});                                                                    //
/////////////////////////////////////////////////////////////////////////

}).call(this);

//# sourceMappingURL=proxy.js.map
