(function(){

/////////////////////////////////////////////////////////////////////////
//                                                                     //
// server/shell-server.js                                              //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
                                                                       //
// Sandstorm - Personal Cloud Sandbox                                  //
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors
// All rights reserved.                                                //
//                                                                     //
// Licensed under the Apache License, Version 2.0 (the "License");     //
// you may not use this file except in compliance with the License.    //
// You may obtain a copy of the License at                             //
//                                                                     //
//   http://www.apache.org/licenses/LICENSE-2.0                        //
//                                                                     //
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,   //
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.                                      //
                                                                       //
// Set up browser policy.                                              //
//                                                                     //
// Note that by default (when the browser-policy package is added), Content-Security-Policy will
// already be set to same-origin-only for everything except XHR and WebSocket. Eval is disabled,
// but inline script are enabled.                                      //
//                                                                     //
// TODO(security): Consider disallowing inline scripts. Currently this forces Meteor to do an
//   extra round trip on startup to fetch server settings. That seems like something that could
//   be fixed in Meteor (e.g. embed the settings as JSON text rather than script). Startup time
//   is incredibly important, and Meteor's templating system (which we use to render all our
//   HTML) already does a good job of protecting us from script injection, so right now I think
//   we should favor avoiding the round trip.                          //
//                                                                     //
// TODO(someday): Detect when an app tries to navigate off-site using CSP's violation reporting
//   feature. Ask the user whether they want to open the link in a new tab. This is an annoying
//   prompt, but if we just open the tab directly then apps can trivially leak by opening tabs
//   with encoded URLs to an evil server. Although, this attack would be very detectable to the
//   user, so maybe it's not a big deal...                             //
                                                                       //
BrowserPolicy.framing.disallow(); // Disallow framing of the UI.       // 36
BrowserPolicy.content.allowFrameOrigin(getWildcardOrigin());           // 37
                                                                       //
// Allow anything to be loaded from the static asset host.             //
var Url = Npm.require("url");                                          // 40
var staticAssetHost = Url.parse(process.env.ROOT_URL).protocol + "//" + globalDb.makeWildcardHost("static");
BrowserPolicy.content.allowImageOrigin(staticAssetHost);               // 43
BrowserPolicy.content.allowScriptOrigin(staticAssetHost);              // 44
BrowserPolicy.content.allowFontOrigin(staticAssetHost);                // 45
BrowserPolicy.content.allowConnectOrigin(staticAssetHost);             // 46
                                                                       //
Meteor.publish("grainsMenu", function () {                             // 48
  var _this = this;                                                    //
                                                                       //
  if (this.userId) {                                                   // 49
    if (Meteor.settings["public"].quotaEnabled) {                      // 50
      (function () {                                                   //
        // Hack: Fire off an asynchronous update to the user's storage usage whenever they open the
        //   front page.                                               //
        // TODO(someday): Implement the ability to reactively subscribe to storage usage from the
        //   back-end?                                                 //
        var userId = _this.userId;                                     // 55
        globalBackend.cap().getUserStorageUsage(userId).then(function (results) {
          inMeteor(function () {                                       // 57
            Meteor.users.update(userId, { $set: { storageUsage: parseInt(results.size) } });
          });                                                          //
        })["catch"](function (err) {                                   //
          if (err.kjType !== "unimplemented") {                        // 61
            console.error(err.stack);                                  // 62
          }                                                            //
        });                                                            //
      })();                                                            //
    }                                                                  //
                                                                       //
    var identityIds = SandstormDb.getUserIdentityIds(globalDb.getUser(this.userId));
    return [UserActions.find({ userId: this.userId }), Grains.find({ userId: this.userId }), ApiTokens.find({ "owner.user.identityId": { $in: identityIds } })];
  } else {                                                             //
    return [];                                                         // 74
  }                                                                    //
});                                                                    //
                                                                       //
Meteor.publish("sessions", function (sessionId) {                      // 78
  // sessionId itself should be secret enough, but they are also not meant to be shared, so as
  // a backup we only publish the session to its owner. Note that `userId` can be null if the
  // user is not logged in or is using incognito mode.                 //
  check(sessionId, String);                                            // 82
  return Sessions.find({ _id: sessionId, $or: [{ userId: this.userId }, { userId: null }] });
});                                                                    //
                                                                       //
Meteor.publish("devPackages", function () {                            // 86
  return DevPackages.find();                                           // 87
});                                                                    //
                                                                       //
Meteor.publish("notifications", function () {                          // 90
  return Notifications.find({ userId: this.userId }, { fields: { timestamp: 1, text: 1, grainId: 1, userId: 1, isUnread: 1, appUpdates: 1,
      admin: 1, referral: 1, mailingListBonus: 1 } });                 // 93
});                                                                    //
                                                                       //
Meteor.publish("notificationGrains", function (notificationIds) {      // 96
  // Since publishes can't be reactive, we leave it to the client to subscribe to both
  // "notifications" and "notificationGrains" reactively.              //
  check(notificationIds, [String]);                                    // 99
  var notifications = Notifications.find({                             // 100
    _id: { $in: notificationIds },                                     // 101
    userId: this.userId                                                // 102
  }, {                                                                 //
    fields: { grainId: 1 }                                             // 104
  });                                                                  //
                                                                       //
  var grainIds = notifications.map(function (row) {                    // 107
    return row.grainId;                                                // 108
  });                                                                  //
                                                                       //
  return Grains.find({ _id: { $in: grainIds } }, { fields: { title: 1 } });
});                                                                    //
                                                                       //
Meteor.publish("hasUsers", function () {                               // 114
  var _this2 = this;                                                   //
                                                                       //
  // Publish pseudo-collection which tells the client if there are any users at all.
  //                                                                   //
  // TODO(cleanup):  This seems overcomplicated.  Does Meteor have a better way?
  var cursor = Meteor.users.find();                                    // 118
  if (cursor.count() > 0) {                                            // 119
    this.added("hasUsers", "hasUsers", { hasUsers: true });            // 120
  } else {                                                             //
    (function () {                                                     //
      var handle = cursor.observeChanges({                             // 122
        added: function (id) {                                         // 123
          _this2.added("hasUsers", "hasUsers", { hasUsers: true });    // 124
          handle.stop();                                               // 125
          handle = null;                                               // 126
        }                                                              //
      });                                                              //
      _this2.onStop(function () {                                      // 129
        if (handle) handle.stop();                                     // 130
      });                                                              //
    })();                                                              //
  }                                                                    //
                                                                       //
  this.ready();                                                        // 134
});                                                                    //
                                                                       //
Meteor.publish("referralInfoPseudo", function () {                     // 137
  var _this3 = this;                                                   //
                                                                       //
  // This publishes a pseudo-collection called referralInfo whose documents have the following
  // form:                                                             //
  //                                                                   //
  // - id: (String) same as the User._id of an identity this user has referred
  // - name: (String) the profile.name from that identity              //
  // - completed: (Boolean) if this referral is complete               //
                                                                       //
  //  If the user is not logged in, then we have no referralInfo.      //
  if (!this.userId) {                                                  // 146
    return [];                                                         // 147
  }                                                                    //
                                                                       //
  // Implementation note:                                              //
  //                                                                   //
  // This pseudo-collection is populated very differently for (1) the completed: false case versus
  // the (2) completed: true case.                                     //
                                                                       //
  // Case 1. Publish information about not-yet-complete referrals.     //
  var notCompletedReferralIdentitiesCursor = Meteor.users.find({       // 156
    referredBy: this.userId,                                           // 157
    "profile.name": { $exists: true }                                  // 158
  }, {                                                                 //
    fields: {                                                          // 160
      _id: 1,                                                          // 161
      referredBy: 1,                                                   // 162
      "profile.name": 1                                                // 163
    }                                                                  //
  });                                                                  //
  var notCompletedReferralIdentitiesHandle = notCompletedReferralIdentitiesCursor.observeChanges({
    // The added function gets called with the id of Bob when Alice refers Bob.
    added: function (id, fields) {                                     // 168
      _this3.added("referralInfo", id, { name: fields.profile.name, completed: false });
    },                                                                 //
    // The removed function gets called when Bob is no longer an uncompleted referral.  Note that
    // this will get more complicated once we support sending completed referrals to the client.
    removed: function (id) {                                           // 173
      _this3.removed("referralInfo", id);                              // 174
    },                                                                 //
    // The modified function gets called when Bob's profile.name changed.
    modified: function (id, fields) {                                  // 177
      _this3.modified("referralInfo", id, { name: fields.profile.name, completed: false });
    }                                                                  //
  });                                                                  //
                                                                       //
  // Case 2. Handle completed referrals.                               //
  //                                                                   //
  // - Do a query for the current list of completed identities.        //
  //                                                                   //
  // - Every time we see a new such identity, we create a query that watches that one identity in
  //   case its profile.name changes.                                  //
  //                                                                   //
  // - Also watch the first query, since the list of completed identities might change.
  var handleForProfileNameByIdentityId = {};                           // 190
  var stopWatchingAllIdentities = function () {                        // 191
    Object.keys(handleForProfileNameByIdentityId).forEach(function (identityId) {
      stopWatchingIdentity(identityId);                                // 193
    });                                                                //
  };                                                                   //
                                                                       //
  var stopWatchingIdentity = function (identityId) {                   // 197
    var handleForProfileName = handleForProfileNameByIdentityId[identityId];
    if (handleForProfileName) {                                        // 199
      _this3.removed("referralInfo", identityId);                      // 200
      handleForProfileName.stop();                                     // 201
      // delete is safe because we iterate across `Object.keys()` which returns a copy.
      delete handleForProfileNameByIdentityId[identityId];             // 203
    }                                                                  //
  };                                                                   //
                                                                       //
  var watchIdentityAndPublishReferralSuccess = function (identityId) {
    var handleForProfileName = handleForProfileNameByIdentityId[identityId];
    if (handleForProfileName) {                                        // 209
      return;                                                          // 210
    }                                                                  //
                                                                       //
    handleForProfileName = Meteor.users.find({                         // 213
      _id: identityId                                                  // 214
    }, {                                                               //
      fields: {                                                        // 216
        "profile.name": 1                                              // 217
      }                                                                //
    }).observeChanges({                                                //
      added: function (id, fields) {                                   // 220
        _this3.added("referralInfo", id, { name: fields.profile.name, completed: true });
      },                                                               //
                                                                       //
      changed: function (id, fields) {                                 // 224
        _this3.changed("referralInfo", id, { name: fields.profile.name, completed: true });
      },                                                               //
                                                                       //
      removed: function (id) {                                         // 228
        stopWatchingIdentity(id);                                      // 229
      }                                                                //
    });                                                                //
                                                                       //
    handleForProfileNameByIdentityId[identityId] = handleForProfileName;
  };                                                                   //
                                                                       //
  var completedIdentityIdsHandle = Meteor.users.find({                 // 236
    _id: this.userId,                                                  // 237
    referredIdentityIds: { $exists: true }                             // 238
  }, {                                                                 //
    fields: {                                                          // 240
      referredIdentityIds: true                                        // 241
    }                                                                  //
  }).observeChanges({                                                  //
    // `added` gets called when a user gets their first completed referral.
    added: function (id, fields) {                                     // 245
      for (var i = 0; i < fields.referredIdentityIds.length; i++) {    // 246
        // Unconditionally mark these as successful referrals and start watching.
        watchIdentityAndPublishReferralSuccess(fields.referredIdentityIds[i]);
      }                                                                //
    },                                                                 //
    // `changed` gets called when a user adds/removes referredIdentityIds, usually when a
    // referral becomes complete.                                      //
    changed: function (id, fields) {                                   // 254
      // Two major tasks.                                              //
      //                                                               //
      // 1. Look for identityIds to unsubscribe from & send removed notices to the client.
      //                                                               //
      // 2. Look for identityIds to subscribe to.                      //
                                                                       //
      // Task 1. Unsubscribe where needed.                             //
      var referredIdentityIdsAsObject = {};                            // 262
      fields.referredIdentityIds.forEach(function (i) {                // 263
        referredIdentityIdsAsObject[i] = true;                         // 263
      });                                                              //
                                                                       //
      Object.keys(handleForProfileNameByIdentityId).forEach(function (identityId) {
        // If the handle doesn't show up in the new list of referredIdentityIds, then remove
        // info from the client & stop it on the server & make it null.
        var handleForProfileName = handleForProfileNameByIdentityId[identityId];
        if (referredIdentityIdsAsObject.hasOwnProperty(identityId)) {  // 269
          stopWatchingIdentity(identityId);                            // 270
        }                                                              //
      });                                                              //
                                                                       //
      // Task 2. Subscribe where needed.                               //
      for (var i = 0; i < fields.referredIdentityIds.length; i++) {    // 275
        // The watch... function will avoid double-creating subscriptions, so this is safe.
        watchIdentityAndPublishReferralSuccess(fields.referredIdentityIds[i]);
      }                                                                //
    },                                                                 //
    // `removed` gets called when a User suddenly has no referredIdentityIds.
    removed: function () {                                             // 281
      // Remove all data from client; stop all handles.                //
      stopWatchingAllIdentities();                                     // 283
    }                                                                  //
  });                                                                  //
                                                                       //
  // With cases 1 and 2 handled, register a cleanup function, then declare victory.
  this.onStop(function () {                                            // 288
    stopWatchingAllIdentities();                                       // 289
    notCompletedReferralIdentitiesHandle.stop();                       // 290
    completedIdentityIdsHandle.stop();                                 // 291
  });                                                                  //
                                                                       //
  this.ready();                                                        // 294
});                                                                    //
                                                                       //
Meteor.publish("backers", function () {                                // 297
  var _this4 = this;                                                   //
                                                                       //
  var backers = Assets.getText("backers.txt");                         // 298
  var anonCount = 0;                                                   // 299
  var counter = 0;                                                     // 300
                                                                       //
  backers.split("\n").forEach(function (name) {                        // 302
    name = name.trim();                                                // 303
    if (name === "") {                                                 // 304
      ++anonCount;                                                     // 305
    } else {                                                           //
      _this4.added("backers", counter++, { name: name });              // 307
    }                                                                  //
  });                                                                  //
                                                                       //
  // Text file ends in \n but that shouldn't count.                    //
  --anonCount;                                                         // 312
                                                                       //
  this.added("backers", "anonymous", { count: anonCount - 1 });        // 314
                                                                       //
  this.ready();                                                        // 316
});                                                                    //
/////////////////////////////////////////////////////////////////////////

}).call(this);

//# sourceMappingURL=shell-server.js.map
