(function(){

/////////////////////////////////////////////////////////////////////////
//                                                                     //
// server/signup-server.js                                             //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
                                                                       //
// Sandstorm - Personal Cloud Sandbox                                  //
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors
// All rights reserved.                                                //
//                                                                     //
// Licensed under the Apache License, Version 2.0 (the "License");     //
// you may not use this file except in compliance with the License.    //
// You may obtain a copy of the License at                             //
//                                                                     //
//   http://www.apache.org/licenses/LICENSE-2.0                        //
//                                                                     //
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,   //
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.                                      //
                                                                       //
// This file covers creation and consumption of invite keys (i.e. to invite people to become
// users of the Sandstorm server).                                     //
                                                                       //
Meteor.publish("signupKey", function (key) {                           // 20
  check(key, String);                                                  // 21
  return SignupKeys.find(key);                                         // 22
});                                                                    //
                                                                       //
Meteor.methods({                                                       // 25
  useSignupKey: function (key) {                                       // 26
    check(key, String);                                                // 27
                                                                       //
    if (!this.userId) {                                                // 29
      throw new Meteor.Error(403, "Must be signed in.");               // 30
    }                                                                  //
                                                                       //
    var user = Meteor.user();                                          // 33
    if (user.signupKey) {                                              // 34
      // Don't waste it.                                               //
      return;                                                          // 36
    }                                                                  //
                                                                       //
    if (isDemoUser()) {                                                // 39
      throw new Meteor.Error(403, "Demo users cannot accept invite keys. Please sign in as a real user.");
    }                                                                  //
                                                                       //
    if (!user.loginIdentities) {                                       // 44
      // Don't consume signup key as identity user.                    //
      return;                                                          // 46
    }                                                                  //
                                                                       //
    var keyInfo = SignupKeys.findOne(key);                             // 49
    if (!keyInfo || keyInfo.used) {                                    // 50
      throw new Meteor.Error(403, "Invalid key or already used.");     // 51
    }                                                                  //
                                                                       //
    if (isSignedUp() && user.payments && user.payments.id) {           // 54
      // This user is already signed up with a payment account. Possibly, they signed up before
      // using their invite, and then went back and clicked on the invite. As a result they
      // probably now have two payment accounts. Mark this invite as used but also add a special
      // flag so we can find it later and cancel the dupe payment account. Record who tried to
      // use it so that we can transfer credits over if needed.        //
      SignupKeys.update(key, { $set: { used: true, rejectedBy: this.userId } });
      return;                                                          // 61
    }                                                                  //
                                                                       //
    var userFields = {                                                 // 64
      signupKey: key,                                                  // 65
      signupNote: keyInfo.note                                         // 66
    };                                                                 //
    if (keyInfo.email) {                                               // 68
      userFields.signupEmail = keyInfo.email;                          // 69
    }                                                                  //
                                                                       //
    if ("quota" in keyInfo) {                                          // 72
      userFields.quota = keyInfo.quota;                                // 73
    }                                                                  //
                                                                       //
    if (keyInfo.plan) {                                                // 76
      userFields.plan = keyInfo.plan;                                  // 77
    }                                                                  //
                                                                       //
    if (keyInfo.payments) {                                            // 80
      userFields["payments.id"] = keyInfo.payments.id;                 // 81
    }                                                                  //
                                                                       //
    Meteor.users.update(this.userId, { $set: userFields });            // 84
    SignupKeys.update(key, { $set: { used: true } });                  // 85
  }                                                                    //
});                                                                    //
/////////////////////////////////////////////////////////////////////////

}).call(this);

//# sourceMappingURL=signup-server.js.map
