(function () {

/* Imports */
var Meteor = Package.meteor.Meteor;
var ECMAScript = Package.ecmascript.ECMAScript;
var _ = Package.underscore._;
var Random = Package.random.Random;
var SandstormDb = Package['sandstorm-db'].SandstormDb;
var SandstormBackend = Package['sandstorm-backend'].SandstormBackend;
var MongoInternals = Package.mongo.MongoInternals;
var Mongo = Package.mongo.Mongo;
var Accounts = Package['accounts-base'].Accounts;
var AccountsServer = Package['accounts-base'].AccountsServer;
var check = Package.check.check;
var Match = Package.check.Match;
var babelHelpers = Package['babel-runtime'].babelHelpers;
var Symbol = Package['ecmascript-runtime'].Symbol;
var Map = Package['ecmascript-runtime'].Map;
var Set = Package['ecmascript-runtime'].Set;
var Promise = Package.promise.Promise;

(function(){

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                  //
// packages/accounts-identity/accounts-identity-server.js                                                           //
//                                                                                                                  //
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                    //
// Sandstorm - Personal Cloud Sandbox                                                                               //
// Copyright (c) 2015 Sandstorm Development Group, Inc. and contributors                                            //
// All rights reserved.                                                                                             //
//                                                                                                                  //
// Licensed under the Apache License, Version 2.0 (the "License");                                                  //
// you may not use this file except in compliance with the License.                                                 //
// You may obtain a copy of the License at                                                                          //
//                                                                                                                  //
//   http://www.apache.org/licenses/LICENSE-2.0                                                                     //
//                                                                                                                  //
// Unless required by applicable law or agreed to in writing, software                                              //
// distributed under the License is distributed on an "AS IS" BASIS,                                                //
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                         //
// See the License for the specific language governing permissions and                                              //
// limitations under the License.                                                                                   //
                                                                                                                    //
function linkIdentityToAccountInternal(db, backend, identityId, accountId) {                                        // 17
  // Links the identity to the account and grants it login access if possible. Makes the account                    //
  // durable if it is a demo account.                                                                               //
                                                                                                                    //
  check(db, SandstormDb);                                                                                           // 21
  check(backend, SandstormBackend);                                                                                 // 22
  check(identityId, String);                                                                                        // 23
  check(accountId, String);                                                                                         // 24
                                                                                                                    //
  var accountUser = Meteor.users.findOne({ _id: accountId });                                                       // 26
  if (!accountUser) {                                                                                               // 27
    throw new Meteor.Error(404, "No account found with ID " + accountId);                                           // 28
  }                                                                                                                 //
                                                                                                                    //
  if (accountUser.profile) {                                                                                        // 31
    throw new Meteor.Error(400, "Cannot link an identity to another identity.");                                    // 32
  }                                                                                                                 //
                                                                                                                    //
  if (!!_.findWhere(accountUser.loginIdentities, { id: identityId }) || !!_.findWhere(accountUser.nonloginIdentities, { id: identityId })) {
    throw new Meteor.Error("alreadyLinked", "Cannot link an identity that's alread linked to this account.");       // 37
  }                                                                                                                 //
                                                                                                                    //
  var identityUser = Meteor.users.findOne({ _id: identityId });                                                     // 41
                                                                                                                    //
  if (!identityUser) {                                                                                              // 43
    throw new Meteor.Error(404, "No identity found with ID " + identityId);                                         // 44
  }                                                                                                                 //
                                                                                                                    //
  if (!identityUser.profile) {                                                                                      // 47
    throw new Meteor.Error(400, "Cannot link an account to another account");                                       // 48
  }                                                                                                                 //
                                                                                                                    //
  db.deleteUnusedAccount(backend, identityUser._id);                                                                // 51
  if (Meteor.users.findOne({ "loginIdentities.id": identityUser._id })) {                                           // 52
    throw new Meteor.Error(403, "Cannot link an identity that can already log into another account");               // 53
  }                                                                                                                 //
                                                                                                                    //
  var alreadyLinked = !!Meteor.users.findOne({ "nonloginIdentities.id": identityUser._id });                        // 57
                                                                                                                    //
  var pushModifier = alreadyLinked ? { nonloginIdentities: { id: identityUser._id } } : { loginIdentities: { id: identityUser._id } };
                                                                                                                    //
  var modifier = undefined;                                                                                         // 63
  if (accountUser.expires) {                                                                                        // 64
    if (alreadyLinked) {                                                                                            // 65
      throw new Meteor.Error(403, "Cannot create an account for an identity that's " + "already linked to another account.");
    }                                                                                                               //
                                                                                                                    //
    modifier = { $push: pushModifier,                                                                               // 70
      $unset: { expires: 1 },                                                                                       // 71
      $set: { upgradedFromDemo: Date.now() } };                                                                     // 72
    if (Meteor.settings["public"].quotaEnabled) {                                                                   // 73
      // Demo users never got the referral notification. Send it now:                                               //
      db.sendReferralProgramNotification(accountUser._id);                                                          // 75
    }                                                                                                               //
  } else {                                                                                                          //
    modifier = { $push: pushModifier };                                                                             // 79
  }                                                                                                                 //
                                                                                                                    //
  // Make sure not to add the same identity twice.                                                                  //
  Meteor.users.update({ _id: accountUser._id,                                                                       // 83
    "nonloginIdentities.id": { $ne: identityUser._id },                                                             // 84
    "loginIdentities.id": { $ne: identityUser._id } }, modifier);                                                   // 85
                                                                                                                    //
  if (accountUser.expires) {                                                                                        // 88
    var demoIdentityId = SandstormDb.getUserIdentityIds(accountUser)[0];                                            // 89
    Meteor.users.update({ _id: demoIdentityId }, { $unset: { expires: 1 },                                          // 90
      $set: { upgradedFromDemo: Date.now() } });                                                                    // 92
                                                                                                                    //
    // Mark the demo identity as nonlogin. It'd be nicer if the identity started out as nonlogin,                   //
    // but to get that to work we would need to adjust the account creation and first login logic.                  //
    Meteor.users.update({ _id: accountUser._id,                                                                     // 96
      "loginIdentities.id": demoIdentityId,                                                                         // 97
      "nonloginIdentities.id": { $not: { $eq: demoIdentityId } } }, { $pull: { loginIdentities: { id: demoIdentityId } },
      $push: { nonloginIdentities: { id: demoIdentityId } } });                                                     // 100
  }                                                                                                                 //
}                                                                                                                   //
                                                                                                                    //
Meteor.methods({                                                                                                    // 105
  loginWithIdentity: function (accountUserId) {                                                                     // 106
    // Logs into the account with ID `accountUserId`. Throws an exception if the current user is                    //
    // not an identity user listed in the account's `loginIdentities` field. This method is not                     //
    // intended to be called directly; client-side code should only invoke it through                               //
    // `Meteor.loginWithIdentity()`, which additionally maintains the standard Meteor client-side                   //
    // login state.                                                                                                 //
                                                                                                                    //
    check(accountUserId, String);                                                                                   // 113
                                                                                                                    //
    var identityUser = Meteor.user();                                                                               // 115
    if (!identityUser || !identityUser.profile) {                                                                   // 116
      throw new Meteor.Error(403, "Must be already logged in as an identity.");                                     // 117
    }                                                                                                               //
                                                                                                                    //
    var accountUser = Meteor.users.findOne(accountUserId);                                                          // 120
    if (!accountUser) {                                                                                             // 121
      throw new Meteor.Error(404, "No such user found: " + accountUserId);                                          // 122
    }                                                                                                               //
                                                                                                                    //
    var linkedIdentity = _.findWhere(accountUser.loginIdentities, { id: identityUser._id });                        // 125
                                                                                                                    //
    if (!linkedIdentity) {                                                                                          // 127
      throw new Meteor.Error(403, "Current identity is not a login identity for account " + accountUserId);         // 128
    }                                                                                                               //
                                                                                                                    //
    return Accounts._loginMethod(this, "loginWithIdentity", [accountUserId], "identity", function () {              // 132
      return { userId: accountUserId };                                                                             // 133
    });                                                                                                             //
  },                                                                                                                //
                                                                                                                    //
  createAccountForIdentity: function () {                                                                           // 136
    // Creates a new account for the currently-logged-in identity.                                                  //
                                                                                                                    //
    var user = Meteor.user();                                                                                       // 139
    if (!(user && user.profile)) {                                                                                  // 140
      throw new Meteor.Error(403, "Must be logged in as an identity in order to create an account.");               // 141
    }                                                                                                               //
                                                                                                                    //
    if (Meteor.users.findOne({                                                                                      // 144
      $or: [{ "loginIdentities.id": user._id }, { "nonloginIdentities.id": user._id }]                              // 145
    })) {                                                                                                           //
      throw new Meteor.Error(403, "Cannot create an account for an identity that's already " + "linked to another account.");
    }                                                                                                               //
                                                                                                                    //
    var newUser = {                                                                                                 // 154
      loginIdentities: [{ id: user._id }],                                                                          // 155
      nonloginIdentities: []                                                                                        // 156
    };                                                                                                              //
    if (user.services.dev) {                                                                                        // 158
      newUser.signupKey = "devAccounts";                                                                            // 159
      if (user.services.dev.isAdmin) {                                                                              // 160
        newUser.isAdmin = true;                                                                                     // 161
      }                                                                                                             //
                                                                                                                    //
      if (user.services.dev.hasCompletedSignup) {                                                                   // 164
        newUser.hasCompletedSignup = true;                                                                          // 165
      }                                                                                                             //
    } else if (user.expires) {                                                                                      //
      // Demo user.                                                                                                 //
      newUser.expires = user.expires;                                                                               // 169
      if (!!user.appDemoId) {                                                                                       // 170
        newUser.appDemoId = user.appDemoId;                                                                         // 171
      }                                                                                                             //
    }                                                                                                               //
                                                                                                                    //
    var options = {};                                                                                               // 175
                                                                                                                    //
    // This will throw an error if the identity has been added as a login identity to some                          //
    // other account while we were executing the body of this method.                                               //
    return Accounts.insertUserDoc(options, newUser);                                                                // 179
  },                                                                                                                //
                                                                                                                    //
  linkIdentityToAccount: function (token) {                                                                         // 182
    // Links the identity of the current user to the account that has `token` as a resume token.                    //
    // If the account is a demo account, makes the account durable and gives the identity login                     //
    // access to it.                                                                                                //
                                                                                                                    //
    check(token, String);                                                                                           // 187
                                                                                                                    //
    if (!this.userId) {                                                                                             // 189
      throw new Meteor.Error(403, "Cannot link to account if not logged in.");                                      // 190
    }                                                                                                               //
                                                                                                                    //
    var hashed = Accounts._hashLoginToken(token);                                                                   // 193
    var accountUser = Meteor.users.findOne({ "services.resume.loginTokens.hashedToken": hashed });                  // 194
                                                                                                                    //
    linkIdentityToAccountInternal(this.connection.sandstormDb, this.connection.sandstormBackend, this.userId, accountUser._id);
  },                                                                                                                //
                                                                                                                    //
  unlinkIdentity: function (accountUserId, identityId) {                                                            // 200
    // Unlinks the identity with ID `identityId` from the account with ID `accountUserId`.                          //
                                                                                                                    //
    check(identityId, String);                                                                                      // 203
    check(accountUserId, String);                                                                                   // 204
                                                                                                                    //
    if (!this.userId) {                                                                                             // 206
      throw new Meteor.Error(403, "Not logged in.");                                                                // 207
    }                                                                                                               //
                                                                                                                    //
    if (!this.connection.sandstormDb.userHasIdentity(this.userId, identityId)) {                                    // 210
      throw new Meteor.Error(403, "Current user does not own identity " + identityId);                              // 211
    }                                                                                                               //
                                                                                                                    //
    var identityUser = Meteor.users.findOne({ _id: identityId });                                                   // 214
    Meteor.users.update({ _id: accountUserId }, { $pull: { nonloginIdentities: { id: identityId },                  // 215
        loginIdentities: { id: identityId } } });                                                                   // 217
  },                                                                                                                //
                                                                                                                    //
  setIdentityAllowsLogin: function (identityId, allowLogin) {                                                       // 220
    // Sets whether the current account allows the identity with ID `identityId` to log in.                         //
                                                                                                                    //
    check(identityId, String);                                                                                      // 223
    check(allowLogin, Boolean);                                                                                     // 224
    if (!this.userId) {                                                                                             // 225
      throw new Meteor.Error(403, "Not logged in.");                                                                // 226
    }                                                                                                               //
                                                                                                                    //
    if (!this.connection.sandstormDb.userHasIdentity(this.userId, identityId)) {                                    // 229
      throw new Meteor.Error(403, "Current user does not own identity " + identityId);                              // 230
    }                                                                                                               //
                                                                                                                    //
    if (allowLogin) {                                                                                               // 233
      Meteor.users.update({ _id: this.userId,                                                                       // 234
        "nonloginIdentities.id": identityId,                                                                        // 235
        "loginIdentities.id": { $not: { $eq: identityId } } }, { $pull: { nonloginIdentities: { id: identityId } },
        $push: { loginIdentities: { id: identityId } } });                                                          // 238
    } else {                                                                                                        //
      Meteor.users.update({ _id: this.userId,                                                                       // 240
        "loginIdentities.id": identityId,                                                                           // 241
        "nonloginIdentities.id": { $not: { $eq: identityId } } }, { $pull: { loginIdentities: { id: identityId } },
        $push: { nonloginIdentities: { id: identityId } } });                                                       // 244
    }                                                                                                               //
  },                                                                                                                //
                                                                                                                    //
  logoutIdentitiesOfCurrentAccount: function () {                                                                   // 248
    // Logs out all identities that are allowed to log in to the current account.                                   //
    var user = Meteor.user();                                                                                       // 250
    if (user && user.loginIdentities) {                                                                             // 251
      user.loginIdentities.forEach(function (identity) {                                                            // 252
        Meteor.users.update({ _id: identity.id }, { $set: { "services.resume.loginTokens": [] } });                 // 253
      });                                                                                                           //
    }                                                                                                               //
  }                                                                                                                 //
});                                                                                                                 //
                                                                                                                    //
Accounts.linkIdentityToAccount = function (db, backend, identityId, accountId) {                                    // 259
  // Links the identity to the account. If the account is a demo account, makes it durable and                      //
  // gives the identity login access to it.                                                                         //
  check(db, SandstormDb);                                                                                           // 262
  check(backend, SandstormBackend);                                                                                 // 263
  check(identityId, String);                                                                                        // 264
  check(accountId, String);                                                                                         // 265
  linkIdentityToAccountInternal(db, backend, identityId, accountId);                                                // 266
};                                                                                                                  //
                                                                                                                    //
Meteor.publish("accountsOfIdentity", function (identityId) {                                                        // 269
  check(identityId, String);                                                                                        // 270
  if (!SandstormDb.ensureSubscriberHasIdentity(this, identityId)) return;                                           // 271
                                                                                                                    //
  // We maintain a map from identity IDs to live query handles that track profile changes.                          //
  var loginIdentities = {};                                                                                         // 274
                                                                                                                    //
  var _this = this;                                                                                                 // 276
  function addIdentitiesOfAccount(account) {                                                                        // 277
    account.loginIdentities.forEach(function (identity) {                                                           // 278
      if (!(identity.id in loginIdentities)) {                                                                      // 279
        var user = Meteor.users.findOne({ _id: identity.id });                                                      // 280
        if (user) {                                                                                                 // 281
          SandstormDb.fillInProfileDefaults(user);                                                                  // 282
          SandstormDb.fillInIntrinsicName(user);                                                                    // 283
          var filteredUser = _.pick(user, "_id", "profile");                                                        // 284
          filteredUser.loginAccountId = account._id;                                                                // 285
          filteredUser.sourceIdentityId = identityId;                                                               // 286
          _this.added("loginIdentitiesOfLinkedAccounts", user._id, filteredUser);                                   // 287
        }                                                                                                           //
                                                                                                                    //
        loginIdentities[identity.id] = Meteor.users.find({ _id: identity.id }, { fields: { profile: 1 } }).observeChanges({
          changed: function (id, fields) {                                                                          // 292
            _this.changed("loginIdentitiesOfLinkedAccounts", id, fields);                                           // 293
          }                                                                                                         //
        });                                                                                                         //
      }                                                                                                             //
    });                                                                                                             //
  }                                                                                                                 //
                                                                                                                    //
  var cursor = Meteor.users.find({                                                                                  // 300
    $or: [{ "loginIdentities.id": identityId }, { "nonloginIdentities.id": identityId }]                            // 301
  });                                                                                                               //
                                                                                                                    //
  var handle = cursor.observe({                                                                                     // 307
    added: function (account) {                                                                                     // 308
      addIdentitiesOfAccount(account);                                                                              // 309
    },                                                                                                              //
                                                                                                                    //
    changed: function (newAccount, oldAccount) {                                                                    // 312
      addIdentitiesOfAccount(newAccount);                                                                           // 313
    },                                                                                                              //
                                                                                                                    //
    removed: function (account) {                                                                                   // 316
      account.loginIdentities.forEach(function (identity) {                                                         // 317
        if (identity.id in loginIdentities) {                                                                       // 318
          self.removed("loginIdentitiesOfLinkedAccounts", identity.id);                                             // 319
          loginIdentities[identity.id].stop();                                                                      // 320
          delete loginIdentities[identity.id];                                                                      // 321
        }                                                                                                           //
      });                                                                                                           //
    }                                                                                                               //
  });                                                                                                               //
  this.ready();                                                                                                     // 326
                                                                                                                    //
  this.onStop(function () {                                                                                         // 328
    handle.stop();                                                                                                  // 329
    Object.keys(loginIdentities).forEach(function (identityId) {                                                    // 330
      loginIdentities[identityId].stop();                                                                           // 331
      delete loginIdentities[identityId];                                                                           // 332
    });                                                                                                             //
  });                                                                                                               //
});                                                                                                                 //
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}).call(this);


/* Exports */
if (typeof Package === 'undefined') Package = {};
Package['accounts-identity'] = {};

})();

//# sourceMappingURL=accounts-identity.js.map
