(function () {

/* Imports */
var Meteor = Package.meteor.Meteor;
var MeteorWrapperLdapjs = Package['typ:ldapjs'].MeteorWrapperLdapjs;
var Accounts = Package['accounts-base'].Accounts;
var AccountsServer = Package['accounts-base'].AccountsServer;
var check = Package.check.check;
var Match = Package.check.Match;
var _ = Package.underscore._;
var ECMAScript = Package.ecmascript.ECMAScript;
var babelHelpers = Package['babel-runtime'].babelHelpers;
var Symbol = Package['ecmascript-runtime'].Symbol;
var Map = Package['ecmascript-runtime'].Map;
var Set = Package['ecmascript-runtime'].Set;
var Promise = Package.promise.Promise;

/* Package-scope variables */
var Future, LDAP_DEFAULTS, LDAP;

(function(){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                             //
// packages/accounts-ldap/ldap_server.js                                                                       //
//                                                                                                             //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                               //
Future = Npm.require("fibers/future");                                                                         // 1
                                                                                                               //
// At a minimum, set up LDAP_DEFAULTS.url and .dn according to                                                 //
// your needs. url should appear as 'ldap://your.url.here'                                                     //
// dn should appear in normal ldap format of comma separated attribute=value                                   //
// e.g. 'uid=someuser,cn=users,dc=somevalue'                                                                   //
LDAP_DEFAULTS = {                                                                                              // 7
  url: false,                                                                                                  // 8
  port: "389",                                                                                                 // 9
  dn: false,                                                                                                   // 10
  searchDN: false,                                                                                             // 11
  searchCredentials: false,                                                                                    // 12
  createNewUser: true,                                                                                         // 13
  defaultDomain: false,                                                                                        // 14
  searchResultsProfileMap: false,                                                                              // 15
  base: null,                                                                                                  // 16
  search: "(objectclass=*)",                                                                                   // 17
  ldapsCertificate: false                                                                                      // 18
};                                                                                                             //
                                                                                                               //
/**                                                                                                            //
 @class LDAP                                                                                                   //
 @constructor                                                                                                  //
 */                                                                                                            //
var LDAP = function (options) {                                                                                // 25
  // Set options                                                                                               //
  this.options = _.clone(LDAP_DEFAULTS);                                                                       // 27
                                                                                                               //
  // Because NPM ldapjs module has some binary builds,                                                         //
  // We had to create a wraper package for it and build for                                                    //
  // certain architectures. The package typ:ldap-js exports                                                    //
  // 'MeteorWrapperLdapjs' which is a wrapper for the npm module                                               //
  this.ldapjs = MeteorWrapperLdapjs;                                                                           // 33
};                                                                                                             //
                                                                                                               //
/**                                                                                                            //
 * Attempt to bind (authenticate) ldap                                                                         //
 * and perform a dn search if specified                                                                        //
 *                                                                                                             //
 * @method ldapCheck                                                                                           //
 *                                                                                                             //
 * @param {Object} options  Object with username, ldapPass and overrides for LDAP_DEFAULTS object.             //
 * Additionally the searchBeforeBind parameter can be specified, which is used to search for the DN            //
 * if not provided.                                                                                            //
 */                                                                                                            //
LDAP.prototype.ldapCheck = function (db, options) {                                                            // 46
                                                                                                               //
  var _this = this;                                                                                            // 48
                                                                                                               //
  options = options || {};                                                                                     // 50
                                                                                                               //
  if (options.hasOwnProperty("username") && options.hasOwnProperty("ldapPass")) {                              // 52
    var _ret = (function () {                                                                                  //
      var explicitDnSelected = db.getLdapExplicitDnSelected();                                                 // 53
      if (explicitDnSelected) {                                                                                // 54
        _this.options.dn = db.getLdapDnPattern().replace("$USERNAME", options.username);                       // 55
      } else {                                                                                                 //
        _this.options.base = db.getLdapBase();                                                                 // 57
        _this.options.searchBeforeBind = {};                                                                   // 58
        _this.options.searchBeforeBind[db.getLdapSearchUsername()] = options.username;                         // 59
      }                                                                                                        //
                                                                                                               //
      var resolved = false;                                                                                    // 62
      var ldapAsyncFut = new Future();                                                                         // 63
                                                                                                               //
      // Create ldap client                                                                                    //
      var fullUrl = db.getLdapUrl();                                                                           // 66
      var client = null;                                                                                       // 67
                                                                                                               //
      var errorFunc = function (err) {                                                                         // 69
        if (err) {                                                                                             // 70
          if (resolved) return;                                                                                // 71
          resolved = true;                                                                                     // 72
          ldapAsyncFut["return"]({                                                                             // 73
            error: err                                                                                         // 74
          });                                                                                                  //
        }                                                                                                      //
      };                                                                                                       //
                                                                                                               //
      if (fullUrl.indexOf("ldaps://") === 0) {                                                                 // 79
        client = _this.ldapjs.createClient({                                                                   // 80
          url: fullUrl,                                                                                        // 81
          tlsOptions: {                                                                                        // 82
            ca: [_this.options.ldapsCertificate]                                                               // 83
          }                                                                                                    //
        }, errorFunc);                                                                                         //
      } else {                                                                                                 //
        client = _this.ldapjs.createClient({                                                                   // 87
          url: fullUrl                                                                                         // 88
        }, errorFunc);                                                                                         //
      }                                                                                                        //
                                                                                                               //
      client.on("error", errorFunc);                                                                           // 92
                                                                                                               //
      // Slide @xyz.whatever from username if it was passed in                                                 //
      // and replace it with the domain specified in defaults                                                  //
      var emailSliceIndex = options.username.indexOf("@");                                                     // 96
      var username = undefined;                                                                                // 97
      var domain = _this.options.defaultDomain;                                                                // 98
                                                                                                               //
      // If user appended email domain, strip it out                                                           //
      // And use the defaults.defaultDomain if set                                                             //
      if (emailSliceIndex !== -1) {                                                                            // 102
        username = options.username.substring(0, emailSliceIndex);                                             // 103
        domain = domain || options.username.substring(emailSliceIndex + 1, options.username.length);           // 104
      } else {                                                                                                 //
        username = options.username;                                                                           // 106
      }                                                                                                        //
                                                                                                               //
      // If DN is provided, use it to bind                                                                     //
      if (explicitDnSelected) {                                                                                // 110
        // Attempt to bind to ldap server with provided info                                                   //
        client.bind(_this.options.searchDN || _this.options.dn, _this.options.searchCredentials || options.ldapPass, function (err) {
          try {                                                                                                // 115
            if (err) {                                                                                         // 116
              // Bind failure, return error                                                                    //
              throw new Meteor.Error(err.code, err.message);                                                   // 118
            }                                                                                                  //
                                                                                                               //
            var handleSearchProfile = function (retObject, bindAfterSearch) {                                  // 121
              retObject.emptySearch = true;                                                                    // 122
                                                                                                               //
              // use dn if given, else use the base for the ldap search                                        //
              var searchBase = _this.options.dn || _this.options.base;                                         // 125
              var searchOptions = {                                                                            // 126
                scope: "sub",                                                                                  // 127
                sizeLimit: 1,                                                                                  // 128
                filter: _this.options.search                                                                   // 129
              };                                                                                               //
                                                                                                               //
              client.search(searchBase, searchOptions, function (err, res) {                                   // 132
                if (err) {                                                                                     // 133
                  if (resolved) return;                                                                        // 134
                  resolved = true;                                                                             // 135
                  ldapAsyncFut["return"]({                                                                     // 136
                    error: err                                                                                 // 137
                  });                                                                                          //
                  return;                                                                                      // 139
                }                                                                                              //
                                                                                                               //
                var found = false;                                                                             // 142
                res.on("searchEntry", function (entry) {                                                       // 143
                  found = true;                                                                                // 144
                  retObject.emptySearch = false;                                                               // 145
                  // Add entry results to return object                                                        //
                  retObject.searchResults = _.omit(entry.object, "userPassword");                              // 147
                                                                                                               //
                  if (bindAfterSearch) {                                                                       // 149
                    client.bind(retObject.searchResults.dn, options.ldapPass, function (err) {                 // 150
                      try {                                                                                    // 151
                        if (err) {                                                                             // 152
                          throw new Meteor.Error(err.code, err.message);                                       // 153
                        }                                                                                      //
                                                                                                               //
                        resolved = true;                                                                       // 156
                        ldapAsyncFut["return"](retObject);                                                     // 157
                      } catch (e) {                                                                            //
                        resolved = true;                                                                       // 159
                        ldapAsyncFut["return"]({                                                               // 160
                          error: e                                                                             // 161
                        });                                                                                    //
                      }                                                                                        //
                    });                                                                                        //
                  } else {                                                                                     //
                    resolved = true;                                                                           // 166
                    ldapAsyncFut["return"](retObject);                                                         // 167
                  }                                                                                            //
                });                                                                                            //
                                                                                                               //
                res.on("error", errorFunc);                                                                    // 171
                                                                                                               //
                res.on("end", function () {                                                                    // 173
                  if (!found) {                                                                                // 174
                    resolved = true;                                                                           // 175
                    ldapAsyncFut["return"]({                                                                   // 176
                      error: new Meteor.Error(500, "No user found")                                            // 177
                    });                                                                                        //
                  }                                                                                            //
                });                                                                                            //
              });                                                                                              //
            };                                                                                                 //
                                                                                                               //
            var retObject = {                                                                                  // 185
              username: username,                                                                              // 186
              searchResults: null,                                                                             // 187
              email: domain ? username + "@" + domain : false,                                                 // 188
              dn: _this.options.dn                                                                             // 189
            };                                                                                                 //
                                                                                                               //
            if (_this.options.searchDN) {                                                                      // 192
              handleSearchProfile(retObject, true);                                                            // 193
            } else {                                                                                           //
              handleSearchProfile(retObject, false);                                                           // 195
            }                                                                                                  //
          } catch (e) {                                                                                        //
            if (resolved) return;                                                                              // 198
            resolved = true;                                                                                   // 199
            ldapAsyncFut["return"]({                                                                           // 200
              error: e                                                                                         // 201
            });                                                                                                //
          }                                                                                                    //
        });                                                                                                    //
      }                                                                                                        //
      // DN not provided, search for DN and use result to bind                                                 //
      else {                                                                                                   //
          (function () {                                                                                       //
            // initialize result                                                                               //
            var retObject = {                                                                                  // 209
              username: username,                                                                              // 210
              email: domain ? username + "@" + domain : false,                                                 // 211
              emptySearch: true,                                                                               // 212
              searchResults: {}                                                                                // 213
            };                                                                                                 //
                                                                                                               //
            var filter = _this.options.search;                                                                 // 216
            Object.keys(_this.options.searchBeforeBind).forEach(function (searchKey) {                         // 217
              filter = "&" + filter + "(" + searchKey + "=" + _this.options.searchBeforeBind[searchKey] + ")";
            });                                                                                                //
                                                                                                               //
            var searchOptions = {                                                                              // 221
              scope: "sub",                                                                                    // 222
              sizeLimit: 1,                                                                                    // 223
              filter: filter                                                                                   // 224
            };                                                                                                 //
                                                                                                               //
            // perform LDAP search to determine DN                                                             //
            client.search(_this.options.base, searchOptions, function (err, res) {                             // 228
              if (err) {                                                                                       // 229
                if (resolved) return;                                                                          // 230
                resolved = true;                                                                               // 231
                ldapAsyncFut["return"]({                                                                       // 232
                  error: err                                                                                   // 233
                });                                                                                            //
                return;                                                                                        // 235
              }                                                                                                //
                                                                                                               //
              retObject.emptySearch = true;                                                                    // 238
              res.on("searchEntry", function (entry) {                                                         // 239
                retObject.dn = entry.objectName;                                                               // 240
                retObject.username = retObject.dn;                                                             // 241
                retObject.emptySearch = false;                                                                 // 242
                                                                                                               //
                retObject.searchResults = _.omit(entry.object, "userPassword");                                // 244
                                                                                                               //
                // use the determined DN to bind                                                               //
                client.bind(entry.objectName, options.ldapPass, function (err) {                               // 247
                  try {                                                                                        // 248
                    if (err) {                                                                                 // 249
                      throw new Meteor.Error(err.code, err.message);                                           // 250
                    } else {                                                                                   //
                      resolved = true;                                                                         // 252
                      ldapAsyncFut["return"](retObject);                                                       // 253
                    }                                                                                          //
                  } catch (e) {                                                                                //
                    if (resolved) return;                                                                      // 257
                    resolved = true;                                                                           // 258
                    ldapAsyncFut["return"]({                                                                   // 259
                      error: e                                                                                 // 260
                    });                                                                                        //
                  }                                                                                            //
                });                                                                                            //
              });                                                                                              //
                                                                                                               //
              res.on("error", errorFunc);                                                                      // 266
                                                                                                               //
              // If no dn is found, return as is.                                                              //
              res.on("end", function (result) {                                                                // 269
                if (retObject.dn === undefined) {                                                              // 270
                  resolved = true;                                                                             // 271
                  ldapAsyncFut["return"](retObject);                                                           // 272
                }                                                                                              //
              });                                                                                              //
            });                                                                                                //
          })();                                                                                                //
        }                                                                                                      //
                                                                                                               //
      return {                                                                                                 // 278
        v: ldapAsyncFut.wait()                                                                                 //
      };                                                                                                       //
    })();                                                                                                      //
                                                                                                               //
    if (typeof _ret === "object") return _ret.v;                                                               //
  } else {                                                                                                     //
    throw new Meteor.Error(403, "Missing LDAP Auth Parameter");                                                // 281
  }                                                                                                            //
};                                                                                                             //
                                                                                                               //
// Register login handler with Meteor                                                                          //
// Here we create a new LDAP instance with options passed from                                                 //
// Meteor.loginWithLDAP on client side                                                                         //
// @param {Object} loginRequest will consist of username, ldapPass, ldap, and ldapOptions                      //
Accounts.registerLoginHandler("ldap", function (loginRequest) {                                                // 290
  // If 'ldap' isn't set in loginRequest object,                                                               //
  // then this isn't the proper handler (return undefined)                                                     //
  if (!loginRequest.ldap) {                                                                                    // 293
    return undefined;                                                                                          // 294
  }                                                                                                            //
                                                                                                               //
  if (!Accounts.identityServices.ldap.isEnabled()) {                                                           // 297
    throw new Meteor.Error(403, "LDAP service is disabled.");                                                  // 298
  }                                                                                                            //
                                                                                                               //
  // Instantiate LDAP with options                                                                             //
  var userOptions = loginRequest.ldapOptions || {};                                                            // 302
  var ldapObj = new LDAP(userOptions);                                                                         // 303
                                                                                                               //
  // Call ldapCheck and get response                                                                           //
  var ldapResponse = ldapObj.ldapCheck(this.connection.sandstormDb, loginRequest);                             // 306
                                                                                                               //
  if (ldapResponse.error) {                                                                                    // 308
    return {                                                                                                   // 309
      userId: null,                                                                                            // 310
      error: ldapResponse.error                                                                                // 311
    };                                                                                                         //
  } else if (ldapResponse.emptySearch) {                                                                       //
    return {                                                                                                   // 314
      userId: null,                                                                                            // 315
      error: new Meteor.Error(403, "User not found in LDAP")                                                   // 316
    };                                                                                                         //
  } else {                                                                                                     //
    // Set initial userId and token vals                                                                       //
    return Accounts.updateOrCreateUserFromExternalService("ldap", { id: ldapResponse.dn, rawAttrs: ldapResponse.searchResults }, {});
  }                                                                                                            //
});                                                                                                            //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}).call(this);


/* Exports */
if (typeof Package === 'undefined') Package = {};
Package['accounts-ldap'] = {
  LDAP: LDAP,
  LDAP_DEFAULTS: LDAP_DEFAULTS
};

})();

//# sourceMappingURL=accounts-ldap.js.map
